@@ -33,20 +33,20 @@
  *      All the public types and function declarations for the core
  *	AOLserver.
  *
- *	$Header: /cvsroot/aolserver/aolserver/include/ns.h,v 1.55.2.19 2005/01/18 21:10:07 dossy Exp $
+ *	$Header: /cvsroot/aolserver/aolserver/include/ns.h,v 1.85 2006/06/26 00:27:54 jgdavidson Exp $
  */
 
 #ifndef NS_H
 #define NS_H
 
 #define NS_MAJOR_VERSION	4
-#define NS_MINOR_VERSION	0
-#define NS_RELEASE_SERIAL	10
+#define NS_MINOR_VERSION	5
+#define NS_RELEASE_SERIAL	0
 #define NS_VERSION_NUM          (NS_MAJOR_VERSION * 10000 \
                                 + NS_MINOR_VERSION * 100 \
                                 + NS_RELEASE_SERIAL)
-#define NS_VERSION		"4.0"
-#define NS_PATCH_LEVEL		"4.0.10"
+#define NS_VERSION		"4.5"
+#define NS_PATCH_LEVEL		"4.5.0"
 
 #define NS_ALPHA_RELEASE	0
 #define NS_BETA_RELEASE		1
@@ -74,13 +74,19 @@
  * Various constants.
  */
 
-#define NS_CONN_CLOSED		  1
-#define NS_CONN_SKIPHDRS	  2
-#define NS_CONN_SKIPBODY	  4
-#define NS_CONN_READHDRS	  8
-#define NS_CONN_SENTHDRS	 16
-#define NS_CONN_KEEPALIVE	 32
-#define NS_CONN_WRITE_ENCODED    64
+#define NS_CONN_CLOSED		  0x1
+#define NS_CONN_SKIPHDRS	  0x2
+#define NS_CONN_SKIPBODY	  0x4
+#define NS_CONN_READHDRS	  0x8
+#define NS_CONN_SENTHDRS	  0x10
+#define NS_CONN_KEEPALIVE	  0x20
+#define NS_CONN_WRITE_ENCODED     0x40
+#define NS_CONN_FILECONTENT       0x80
+#define NS_CONN_RUNNING           0x100
+#define NS_CONN_OVERFLOW	  0x200
+#define NS_CONN_TIMEOUT	       	  0x400
+#define NS_CONN_GZIP		  0x800
+#define NS_CONN_CHUNK		  0x1000
 
 #define NS_CONN_MAXCLS		 16
 
@@ -96,6 +102,7 @@
 #define NS_FILTER_POST_AUTH       2
 #define NS_FILTER_TRACE           4
 #define NS_FILTER_VOID_TRACE      8
+#define NS_FILTER_PRE_QUEUE	 16
 #define NS_REGISTER_SERVER_TRACE 16
 #define NS_OP_NOINHERIT		  2
 #define NS_OP_NODELETE		  4
@@ -106,18 +113,33 @@
 #define NS_SCHED_WEEKLY		  8
 #define NS_SCHED_PAUSED		 16 
 #define NS_SCHED_RUNNING	 32 
-#define NS_SOCK_READ		  1
-#define NS_SOCK_WRITE		  2
-#define NS_SOCK_EXCEPTION	  4
-#define NS_SOCK_EXIT		  8
-#define NS_SOCK_DROP             16
-#define NS_SOCK_CANCEL           32
-#define NS_SOCK_ANY		255
+
+#define NS_SOCK_READ	 	0x01 
+#define NS_SOCK_WRITE		0x02
+#define NS_SOCK_EXCEPTION	0x04
+#define NS_SOCK_EXIT		0x08
+#define NS_SOCK_DROP		0x10
+#define NS_SOCK_CANCEL		0x20
+#define NS_SOCK_TIMEOUT		0x40
+#define NS_SOCK_INIT		0x80
+#define NS_SOCK_ANY		(NS_SOCK_READ|NS_SOCK_WRITE|NS_SOCK_EXCEPTION)
+
 #define NS_ENCRYPT_BUFSIZE 	 16
 #define NS_DRIVER_ASYNC		  1	/* Use async read-ahead. */
 #define NS_DRIVER_SSL		  2	/* Use SSL port, protocol defaults. */
 #define NS_DRIVER_VERSION_1       1
 
+/*
+ * The following are valid Tcl traces.
+ */
+
+#define NS_TCL_TRACE_CREATE	   1
+#define NS_TCL_TRACE_DELETE	   2
+#define NS_TCL_TRACE_ALLOCATE	   4
+#define NS_TCL_TRACE_DEALLOCATE	   8
+#define NS_TCL_TRACE_GETCONN	  16
+#define NS_TCL_TRACE_FREECONN	  32
+
 #if defined(__alpha)
 typedef long			ns_int64;
 typedef unsigned long		ns_uint64;
@@ -238,6 +260,18 @@ NS_EXTERN int			kill(int pid, int sig);
 #define Ns_DStringTrunc		Tcl_DStringTrunc
 #define Ns_DStringSetLength	Tcl_DStringSetLength
 
+/*
+ * Typedefs of variables
+ */
+
+typedef struct _Ns_Cache	*Ns_Cache;
+typedef struct _Ns_Entry	*Ns_Entry;
+typedef Tcl_HashSearch 		 Ns_CacheSearch;
+typedef struct _Ns_Cls 		*Ns_Cls;
+typedef void 	      		*Ns_OpContext;
+typedef struct _Ns_TaskQueue 	*Ns_TaskQueue;
+typedef struct _Ns_Task 	*Ns_Task;
+
 /*
  * This is used for logging messages.
  */
@@ -276,6 +310,7 @@ typedef int   (Ns_TclInterpInitProc) (Tcl_Interp *interp, void *arg);
 typedef int   (Ns_TclTraceProc) (Tcl_Interp *interp, void *arg);
 typedef void  (Ns_TclDeferProc) (Tcl_Interp *interp, void *arg);
 typedef int   (Ns_SockProc) (SOCKET sock, void *arg, int why);
+typedef void  (Ns_TaskProc) (Ns_Task *task, SOCKET sock, void *arg, int why);
 typedef void  (Ns_SchedProc) (void *arg, int id);
 typedef int   (Ns_ServerInitProc) (char *server);
 typedef int   (Ns_ModuleInitProc) (char *server, char *module);
@@ -284,7 +319,9 @@ typedef int   (Ns_RequestAuthorizeProc) (char *server, char *method,
 typedef void  (Ns_AdpParserProc)(Ns_DString *outPtr, char *page);
 typedef int   (Ns_UserAuthorizeProc) (char *user, char *passwd);
 typedef int   (Ns_LogFlushProc) (char *msg, size_t len);
-typedef int   (Ns_LogProc) (Ns_DString *dsPtr, Ns_LogSeverity severity, char *fmt, va_list ap);
+typedef int   (Ns_LogProc) (Ns_DString *dsPtr, Ns_LogSeverity severity,
+			    char * fmt, va_list ap);
+typedef int   (Ns_GzipProc)(char *buf, int len, int level, Tcl_DString *dsPtr);
 
 /*
  * The field of a key-value data structure.
@@ -334,9 +371,24 @@ typedef struct Ns_Conn {
     char       *authUser;
     char       *authPasswd;
     int         contentLength;
-    int         flags;		/* Currently, only NS_CONN_CLOSED. */
+    int         flags;		/* Read-only: One or more NS_CONN_ bits which
+				 * specify the state of the connection. */
 } Ns_Conn;
 
+/* 
+ * The following structure maintains data from an
+ * updated form file.  File content can be accessed
+ * at the given offset and length from the bytes
+ * returned by Ns_ConnContent. 
+ */
+ 
+typedef struct Ns_ConnFile {
+    char   *name;
+    Ns_Set *headers;
+    off_t   offset;
+    off_t   length;
+} Ns_ConnFile;
+
 /*
  * The index data structure.  This is a linear array of values.
  */
@@ -445,17 +497,13 @@ typedef void  (Ns_TraceProc) (void *arg, Ns_Conn *conn);
 typedef int   (Ns_FilterProc) (void *arg, Ns_Conn *conn, int why);
 typedef int   (Ns_UrlToFileProc) (Ns_DString *dsPtr, char *server, char *url);
 typedef char *(Ns_LocationProc) (Ns_Conn *conn);
+typedef void  (Ns_QueueWaitProc) (Ns_Conn *conn, SOCKET sock, void *arg, int why);
 
 /*
- * Typedefs of variables
+ * init.c:
  */
 
-typedef struct _Ns_Cache	*Ns_Cache;
-typedef struct _Ns_Entry	*Ns_Entry;
-typedef Tcl_HashSearch 		 Ns_CacheSearch;
-
-typedef struct _Ns_Cls 		*Ns_Cls;
-typedef void 	      		*Ns_OpContext;
+NS_EXTERN void Ns_LibInit(void);
 
 /*
  * adpparse.c:
@@ -468,6 +516,7 @@ NS_EXTERN int Ns_AdpRegisterParser(char *extension, Ns_AdpParserProc *proc);
  */
 
 NS_EXTERN int Ns_AdpRequest(Ns_Conn *conn, char *file);
+NS_EXTERN int Ns_AdpRequestEx(Ns_Conn *conn, char *file, Ns_Time *ttlPtr);
 
 /*
  * auth.c:
@@ -492,7 +541,7 @@ NS_EXTERN Ns_Cache *Ns_CacheCreateSz(char *name, int keys, size_t maxSize,
 				  Ns_Callback *freeProc);
 NS_EXTERN void Ns_CacheDestroy(Ns_Cache *cache);
 NS_EXTERN Ns_Cache *Ns_CacheFind(char *name);
-NS_EXTERN void *Ns_CacheMalloc(Ns_Cache *cache, size_t len);
+NS_EXTERN void *Ns_CacheMalloc(Ns_Cache *cache, size_t len) _nsmalloc;
 NS_EXTERN void Ns_CacheFree(Ns_Cache *cache, void *bytes);
 NS_EXTERN Ns_Entry *Ns_CacheFindEntry(Ns_Cache *cache, char *key);
 NS_EXTERN Ns_Entry *Ns_CacheCreateEntry(Ns_Cache *cache, char *key, int *newPtr);
@@ -542,8 +591,8 @@ NS_EXTERN void Ns_ClsSet(Ns_Cls *clsPtr, Ns_Conn *conn, void *data);
  * compress.c:
  */
 
-NS_EXTERN int Ns_Compress(char *buf, int len, Tcl_DString *outPtr, int level);
-NS_EXTERN int Ns_CompressGzip(char *buf, int len, Tcl_DString *outPtr, int level);
+NS_EXTERN void Ns_SetGzipProc(Ns_GzipProc *procPtr);
+NS_EXTERN int Ns_Gzip(char *buf, int len, int level, Tcl_DString *dsPtr);
 
 /*
  * config.c:
@@ -565,23 +614,33 @@ NS_EXTERN void Ns_GetVersion(int *major, int *minor, int *patch, int *type);
 
 NS_EXTERN int Ns_ConnClose(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnInit(Ns_Conn *connPtr);
+NS_EXTERN int Ns_ConnId(Ns_Conn *connPtr);
 NS_EXTERN int Ns_ConnRead(Ns_Conn *conn, void *vbuf, int toread);
 NS_EXTERN int Ns_ConnWrite(Ns_Conn *conn, void *buf, int towrite);
+NS_EXTERN int Ns_ConnFlush(Ns_Conn *conn, char *buf, int len, int stream);
+NS_EXTERN int Ns_ConnFlushDirect(Ns_Conn *conn, char *buf, int len, int stream);
 NS_EXTERN int Ns_ConnContentFd(Ns_Conn *conn);
+NS_EXTERN int Ns_ConnContentOnDisk(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReadLine(Ns_Conn *conn, Ns_DString *dsPtr, int *nreadPtr);
 NS_EXTERN int Ns_WriteConn(Ns_Conn *conn, char *buf, int len);
 NS_EXTERN int Ns_WriteCharConn(Ns_Conn *conn, char *buf, int len);
 NS_EXTERN int Ns_ConnPuts(Ns_Conn *conn, char *string);
+NS_EXTERN int Ns_ConnSend(Ns_Conn *conn, struct iovec *bufs, int nbufs);
 NS_EXTERN int Ns_ConnSendDString(Ns_Conn *conn, Ns_DString *dsPtr);
 NS_EXTERN int Ns_ConnSendChannel(Ns_Conn *conn, Tcl_Channel chan, int nsend);
 NS_EXTERN int Ns_ConnSendFp(Ns_Conn *conn, FILE *fp, int nsend);
 NS_EXTERN int Ns_ConnSendFd(Ns_Conn *conn, int fd, int nsend);
+NS_EXTERN int Ns_ConnSendFdEx(Ns_Conn *conn, int fd, off_t off, int nsend);
 NS_EXTERN int Ns_ConnCopyToDString(Ns_Conn *conn, size_t ncopy,
 				   Ns_DString *dsPtr);
 NS_EXTERN int Ns_ConnCopyToChannel(Ns_Conn *conn, size_t ncopy, Tcl_Channel chan);
 NS_EXTERN int Ns_ConnCopyToFile(Ns_Conn *conn, size_t ncopy, FILE *fp);
 NS_EXTERN int Ns_ConnCopyToFd(Ns_Conn *conn, size_t ncopy, int fd);
 NS_EXTERN int Ns_ConnFlushContent(Ns_Conn *conn);
+NS_EXTERN void Ns_ConnSetType(Ns_Conn *conn, char *type);
+NS_EXTERN char *Ns_ConnGetType(Ns_Conn *conn);
+NS_EXTERN void Ns_ConnSetStatus(Ns_Conn *conn, int status);
+NS_EXTERN int Ns_ConnGetStatus(Ns_Conn *conn);
 NS_EXTERN void Ns_ConnSetEncoding(Ns_Conn *conn, Tcl_Encoding encoding);
 NS_EXTERN Tcl_Encoding Ns_ConnGetEncoding(Ns_Conn *conn);
 NS_EXTERN void Ns_ConnSetUrlEncoding(Ns_Conn *conn, Tcl_Encoding encoding);
@@ -590,8 +649,6 @@ NS_EXTERN int Ns_ConnModifiedSince(Ns_Conn *conn, time_t inTime);
 NS_EXTERN char *Ns_ConnGets(char *outBuffer, size_t inSize, Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReadHeaders(Ns_Conn *conn, Ns_Set *set, int *nreadPtr);
 NS_EXTERN int Ns_ParseHeader(Ns_Set *set, char *header, Ns_HeaderCaseDisposition disp);
-NS_EXTERN Ns_Set  *Ns_ConnGetQuery(Ns_Conn *conn);
-NS_EXTERN void Ns_ConnClearQuery(Ns_Conn *conn);
 NS_EXTERN int Ns_QueryToSet(char *query, Ns_Set *qset);
 NS_EXTERN Ns_Set *Ns_ConnHeaders(Ns_Conn *conn);
 NS_EXTERN Ns_Set *Ns_ConnOutputHeaders(Ns_Conn *conn);
@@ -612,8 +669,12 @@ NS_EXTERN int Ns_ConnPort(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnSock(Ns_Conn *conn);
 NS_EXTERN char *Ns_ConnDriverName(Ns_Conn *conn);
 NS_EXTERN void *Ns_ConnDriverContext(Ns_Conn *conn);
+NS_EXTERN int Ns_ConnGetKeepAliveFlag(Ns_Conn *conn);
+NS_EXTERN void Ns_ConnSetKeepAliveFlag(Ns_Conn *conn, int flag);
 NS_EXTERN int Ns_ConnGetWriteEncodedFlag(Ns_Conn *conn);
 NS_EXTERN void Ns_ConnSetWriteEncodedFlag(Ns_Conn *conn, int flag);
+NS_EXTERN int Ns_ConnGetGzipFlag(Ns_Conn *conn);
+NS_EXTERN void Ns_ConnSetGzipFlag(Ns_Conn *conn, int flag);
 NS_EXTERN void Ns_ConnSetUrlEncoding(Ns_Conn *conn, Tcl_Encoding encoding);
 
 /*
@@ -634,7 +695,10 @@ NS_EXTERN int Ns_GetAllAddrByHost(Ns_DString *dsPtr, char *host);
  * driver.c:
  */
 
-NS_EXTERN int Ns_DriverInit(char *server, char *module, Ns_DriverInitData *init);
+NS_EXTERN int Ns_DriverInit(char *server, char *module,
+			    Ns_DriverInitData *init);
+NS_EXTERN void Ns_QueueWait(Ns_Conn *conn, SOCKET sock, Ns_QueueWaitProc *proc,
+    	     void *arg, int when, Ns_Time *timePtr);
 
 /*
  * dstring.c:
@@ -643,7 +707,7 @@ NS_EXTERN int Ns_DriverInit(char *server, char *module, Ns_DriverInitData *init)
 NS_EXTERN char **Ns_DStringAppendArgv(Ns_DString *dsPtr);
 NS_EXTERN char *Ns_DStringVarAppend(Ns_DString *dsPtr, ...);
 NS_EXTERN char *Ns_DStringExport(Ns_DString *dsPtr);
-NS_EXTERN char *Ns_DStringPrintf(Ns_DString *dsPtr, char *fmt,...);
+NS_EXTERN char *Ns_DStringPrintf(Ns_DString *dsPtr, char *fmt,...) _nsprintflike(2,3);
 NS_EXTERN char *Ns_DStringVPrintf(Ns_DString *dsPtr, char *fmt, va_list ap);
 NS_EXTERN char *Ns_DStringAppendArg(Ns_DString *dsPtr, char *string);
 NS_EXTERN Ns_DString *Ns_DStringPop(void);
@@ -672,6 +736,7 @@ NS_EXTERN void Ns_SetUrlToFileProc(char *server, Ns_UrlToFileProc *procPtr);
 NS_EXTERN int Ns_UrlToFile(Ns_DString *dsPtr, char *server, char *url);
 NS_EXTERN int Ns_UrlIsFile(char *server, char *url);
 NS_EXTERN int Ns_UrlIsDir(char *server, char *url);
+NS_EXTERN int Ns_FastPathOp(void *arg, Ns_Conn *conn);
 
 /*
  * filter.c:
@@ -757,6 +822,22 @@ NS_EXTERN Ns_List *Ns_ListMapcar(Ns_List *lPtr, Ns_ElemValProc *valProc);
 
 NS_EXTERN void Ns_GenSeeds(unsigned long *seedsPtr, int nseeds);
 NS_EXTERN double Ns_DRand(void);
+
+/* 
+ * task.c: 
+ */
+
+NS_EXTERN Ns_TaskQueue *Ns_CreateTaskQueue(char *name);
+NS_EXTERN void Ns_DestroyTaskQueue(Ns_TaskQueue *queue);
+NS_EXTERN Ns_Task *Ns_TaskCreate(SOCKET sock, Ns_TaskProc *proc, void *arg);
+NS_EXTERN int  Ns_TaskEnqueue(Ns_Task *task, Ns_TaskQueue *queue);
+NS_EXTERN void  Ns_TaskRun(Ns_Task *task);
+NS_EXTERN void Ns_TaskCallback(Ns_Task *task, int when, Ns_Time *timeoutPtr);
+NS_EXTERN void Ns_TaskDone(Ns_Task *task);
+NS_EXTERN int  Ns_TaskCancel(Ns_Task *task);
+NS_EXTERN int  Ns_TaskWait(Ns_Task *task, Ns_Time *timeoutPtr);
+NS_EXTERN SOCKET Ns_TaskFree(Ns_Task *task);
+
 /*
  * tclobj.c:
  */
@@ -775,14 +856,11 @@ NS_EXTERN int Ns_TclDetachedThread(Tcl_Interp *interp, char *script);
  * tclxkeylist.c:
  */
 
-NS_EXTERN char *Tcl_DeleteKeyedListField (Tcl_Interp  *interp, CONST char *fieldName,
-        CONST char *keyedList);
-NS_EXTERN int Tcl_GetKeyedListField (Tcl_Interp  *interp, CONST char *fieldName,
-        CONST char *keyedList, char **fieldValuePtr);
-NS_EXTERN int Tcl_GetKeyedListKeys (Tcl_Interp  *interp, char CONST *subFieldName,
-        CONST char *keyedList, int *keysArgcPtr, char ***keysArgvPtr);
-NS_EXTERN char *Tcl_SetKeyedListField (Tcl_Interp  *interp, CONST char *fieldName,
-        CONST char *fieldvalue, CONST char *keyedList);
+NS_EXTERN int TclX_KeyedListDelete (Tcl_Interp *interp, Tcl_Obj *keylPtr, char *key);
+NS_EXTERN int TclX_KeyedListGet (Tcl_Interp *interp, Tcl_Obj *keylPtr, char *key, Tcl_Obj **valuePtrPtr);
+NS_EXTERN int TclX_KeyedListGetKeys (Tcl_Interp *interp, Tcl_Obj *keylPtr, char *key, Tcl_Obj **listObjPtrPtr);
+NS_EXTERN int TclX_KeyedListSet (Tcl_Interp *interp, Tcl_Obj *keylPtr, char *key, Tcl_Obj *valuePtr);
+
 
 /*
  * listen.c:
@@ -798,8 +876,8 @@ NS_EXTERN int Ns_SockPortBound(int port);
 
 NS_EXTERN char *Ns_InfoErrorLog(void);
 NS_EXTERN int   Ns_LogRoll(void);
-NS_EXTERN void  Ns_Log(Ns_LogSeverity severity, char *fmt, ...);
-NS_EXTERN void  Ns_Fatal(char *fmt, ...);
+NS_EXTERN void  Ns_Log(Ns_LogSeverity severity, char *fmt, ...) _nsprintflike(2,3);
+NS_EXTERN void  Ns_Fatal(char *fmt, ...) _nsprintflike(1,2);
 NS_EXTERN char *Ns_LogTime(char *timeBuf);
 NS_EXTERN char *Ns_LogTime2(char *timeBuf, int gmt);
 NS_EXTERN int   Ns_RollFile(char *file, int max);
@@ -850,7 +928,9 @@ NS_EXTERN char *Ns_GetMimeType(char *file);
 NS_EXTERN Tcl_Encoding Ns_GetEncoding(char *name);
 NS_EXTERN Tcl_Encoding Ns_GetFileEncoding(char *file);
 NS_EXTERN Tcl_Encoding Ns_GetTypeEncoding(char *type);
+NS_EXTERN char *Ns_FindCharset(char *type, int *lenPtr);
 NS_EXTERN Tcl_Encoding Ns_GetCharsetEncoding(char *charset);
+NS_EXTERN Tcl_Encoding Ns_GetCharsetEncodingEx(char *charset, int len);
 
 /*
  * modload.c:
@@ -889,11 +969,13 @@ NS_EXTERN int Ns_ConnRedirect(Ns_Conn *conn, char *url);
  */
 
 NS_EXTERN int Ns_PathIsAbsolute(char *path);
-NS_EXTERN char *Ns_NormalizePath(Ns_DString *dsPtr, char *path);
-NS_EXTERN char *Ns_MakePath(Ns_DString *dsPtr, ...);
-NS_EXTERN char *Ns_LibPath(Ns_DString *dsPtr, ...);
 NS_EXTERN char *Ns_HomePath(Ns_DString *dsPtr, ...);
-NS_EXTERN char *Ns_ModulePath(Ns_DString *dsPtr, char *server, char *module, ...);
+NS_EXTERN char *Ns_LibPath(Ns_DString *dsPtr, ...);
+NS_EXTERN char *Ns_BinPath(Ns_DString *dsPtr, ...);
+NS_EXTERN char *Ns_ModulePath(Ns_DString *dsPtr, char *server,
+			      char *module, ...);
+NS_EXTERN char *Ns_MakePath(Ns_DString *dsPtr, ...);
+NS_EXTERN char *Ns_NormalizePath(Ns_DString *dsPtr, char *path);
 
 /*
  * proc.c:
@@ -921,6 +1003,7 @@ NS_EXTERN void Ns_QuoteHtml(Ns_DString *pds, char *string);
 
 NS_EXTERN void Ns_FreeRequest(Ns_Request *request);
 NS_EXTERN Ns_Request *Ns_ParseRequest(char *line);
+NS_EXTERN Ns_Request *Ns_ParseRequestEx(char *line, Tcl_Encoding encoding);
 NS_EXTERN char *Ns_SkipUrl(Ns_Request *request, int n);
 NS_EXTERN void Ns_SetRequestUrl(Ns_Request *request, char *url);
 
@@ -940,7 +1023,7 @@ NS_EXTERN void Ns_ConnSetTypeHeader(Ns_Conn *conn, char *type);
 NS_EXTERN void Ns_ConnSetLengthHeader(Ns_Conn *conn, int length);
 NS_EXTERN void Ns_ConnSetLastModifiedHeader(Ns_Conn *conn, time_t *mtime);
 NS_EXTERN void Ns_ConnSetExpiresHeader(Ns_Conn *conn, char *expires);
-NS_EXTERN int Ns_ConnPrintfHeader(Ns_Conn *conn, char *fmt,...);
+NS_EXTERN int Ns_ConnPrintfHeader(Ns_Conn *conn, char *fmt,...) _nsprintflike(2,3);
 NS_EXTERN int Ns_ConnResetReturn(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReturnAdminNotice(Ns_Conn *conn, int status, char *title,
 				    char *notice);
@@ -961,6 +1044,7 @@ NS_EXTERN int Ns_ConnReturnNotFound(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReturnNotModified(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReturnNotImplemented(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReturnInternalError(Ns_Conn *conn);
+NS_EXTERN int Ns_ConnReturnServiceUnavailable(Ns_Conn *conn);
 NS_EXTERN int Ns_ConnReturnStatus(Ns_Conn *conn, int status);
 NS_EXTERN int Ns_ConnReturnOpenChannel(Ns_Conn *conn, int status, char *type,
 				    Tcl_Channel chan, int len);
@@ -968,6 +1052,8 @@ NS_EXTERN int Ns_ConnReturnOpenFile(Ns_Conn *conn, int status, char *type,
 				 FILE *fp, int len);
 NS_EXTERN int Ns_ConnReturnOpenFd(Ns_Conn *conn, int status, char *type, int fd,
 			       int len);
+NS_EXTERN int Ns_ConnReturnOpenFdEx(Ns_Conn *conn, int status, char *type,
+				    int fd, off_t off, int len);
 NS_EXTERN int Ns_ConnReturnFile(Ns_Conn *conn, int status, char *type,
 			     char *filename);
 
@@ -1040,6 +1126,7 @@ NS_EXTERN void Ns_SetPrint(Ns_Set *set);
 NS_EXTERN int Ns_SockRecv(SOCKET sock, void *vbuf, int nrecv, int timeout);
 NS_EXTERN int Ns_SockSend(SOCKET sock, void *vbuf, int nsend, int timeout);
 NS_EXTERN int Ns_SockWait(SOCKET sock, int what, int timeout);
+NS_EXTERN int Ns_SockWaitEx(SOCKET sock, int what, int ms);
 
 NS_EXTERN SOCKET Ns_BindSock(struct sockaddr_in *psa);
 NS_EXTERN SOCKET Ns_SockBind(struct sockaddr_in *psa);
@@ -1124,6 +1211,8 @@ NS_EXTERN Ns_Conn *Ns_TclGetConn(Tcl_Interp *interp);
 NS_EXTERN int Ns_TclRegisterAtCreate(Ns_TclTraceProc *proc, void *arg);
 NS_EXTERN int Ns_TclRegisterAtCleanup(Ns_TclTraceProc *proc, void *arg);
 NS_EXTERN int Ns_TclRegisterAtDelete(Ns_TclTraceProc *proc, void *arg);
+NS_EXTERN int Ns_TclRegisterTrace(char *server, Ns_TclTraceProc *proc,
+				  void *arg, int when);
 
 /*
  * tclop.c:
@@ -1209,15 +1298,16 @@ NS_EXTERN void *Ns_ServerSpecificDestroy(char *handle, int id, int flags);
 NS_EXTERN int Ns_CloseOnExec(int fd);
 NS_EXTERN int Ns_NoCloseOnExec(int fd);
 NS_EXTERN int Ns_DupHigh(int *fdPtr);
+NS_EXTERN int Ns_DevNull(void);
 NS_EXTERN int Ns_GetTemp(void);
 NS_EXTERN void Ns_ReleaseTemp(int fd);
+NS_EXTERN int ns_sockpair(SOCKET *socks);
+NS_EXTERN int ns_pipe(int *fds);
 
 /*
  * unix.c, win32.c:
  */
 
-NS_EXTERN int ns_sockpair(SOCKET *socks);
-NS_EXTERN int ns_pipe(int *fds);
 NS_EXTERN int Ns_GetUserHome(Ns_DString *pds, char *user);
 NS_EXTERN int Ns_GetGid(char *group);
 NS_EXTERN int Ns_GetUserGid(char *user);
@@ -1227,14 +1317,18 @@ NS_EXTERN int Ns_GetUid(char *user);
  * form.c:
  */
 
+NS_EXTERN Ns_Set  *Ns_ConnGetQuery(Ns_Conn *conn);
 NS_EXTERN void Ns_ConnClearQuery(Ns_Conn *conn);
+NS_EXTERN Ns_ConnFile *Ns_ConnGetFile(Ns_Conn *conn, char *file);
+NS_EXTERN Ns_ConnFile *Ns_ConnFirstFile(Ns_Conn *conn, Tcl_HashSearch *searchPtr);
+NS_EXTERN Ns_ConnFile *Ns_ConnNextFile(Tcl_HashSearch *searchPtr);
 
 /*
  * Compatibility macros.
  */
 
 #ifndef NS_NOCOMPAT
-#define DllExport           NS_EXPORT
+#define DllExport               NS_EXPORT
 #define Ns_Select	        select
 #define Ns_InfoHome()		Ns_InfoHomePath()
 #define Ns_InfoServer()		Ns_InfoServerName()
@@ -1255,6 +1349,7 @@ NS_EXTERN void Ns_ConnClearQuery(Ns_Conn *conn);
 #define Ns_ReturnForbidden(c)	Ns_ConnReturnForbidden(c)
 #define Ns_ReturnNotFound(c)	Ns_ConnReturnNotFound(c)
 #define Ns_ReturnInternalError(c) Ns_ConnReturnInternalError(c)
+#define Ns_ReturnServiceUnavailable(c) Ns_ConnReturnServiceUnavailable(c)
 #define Ns_ReturnNotImplemented(c) Ns_ConnReturnNotImplemented(c)
 #define Ns_ReturnStatus(c,s)	Ns_ConnReturnStatus(c,s)
 #define Ns_TypeHeader(c,t)      Ns_ConnSetTypeHeader(c,t)
