@@ -28,20 +28,18 @@
  */
 
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nscgi/nscgi.c,v 1.23.2.1 2004/06/23 18:54:28 dossy Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nscgi/nscgi.c,v 1.31 2006/04/13 19:05:57 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "ns.h"
 #include <sys/stat.h>
 #include <ctype.h>
 #include <stdlib.h>	/* environ */
 
-#define BUFSIZE	    4096
-#define NDSTRINGS   5
-#define DEFAULT_MAXINPUT    1024000
+#define BUFSIZE	    	4096
+#define NDSTRINGS   	5
 #define CGI_NPH	    	1
 #define CGI_GETHOST	2
-#define CGI_ECONTENT	4
-#define CGI_SYSENV	8
+#define CGI_SYSENV	4
 
 #ifdef _WIN32
 #define S_ISREG(m)	((m)&_S_IFREG)
@@ -65,7 +63,6 @@ typedef struct Mod {
     Ns_Set         *mergeEnv;
     struct Cgi     *firstCgiPtr;
     int		    flags;
-    int             maxInput;
     int             maxCgi;
     int     	    maxWait;
     int             activeCgi;
@@ -113,22 +110,13 @@ typedef struct Map {
     char     *path;
 } Map;
 
-/*
- * The following file descriptor is opened once on the first load and used
- * simply for duping as stdin in the child process.  This ensures the child
- * will get a proper EOF without having to allocate an empty temp file.
- */
- 
-static int devNull;
-
 static Ns_OpProc CgiRequest;
-static void     CgiRegister(Mod *modPtr, char *map);
+static int      CgiRegister(Mod *modPtr, char *map);
 static Ns_Callback CgiFreeMap;
 static Ns_DString *CgiDs(Cgi *cgiPtr);
 static int	CgiInit(Cgi *cgiPtr, Map *mapPtr, Ns_Conn *conn);
 static void	CgiFree(Cgi *cgiPtr);
 static int  	CgiExec(Cgi *cgiPtr, Ns_Conn *conn);
-static int	CgiSpool(Cgi *cgiPtr, Ns_Conn *conn);
 static int	CgiCopy(Cgi *cgiPtr, Ns_Conn *conn);
 static int	CgiRead(Cgi *cgiPtr);
 static int	CgiReadLine(Cgi *cgiPtr, Ns_DString *dsPtr);
@@ -136,13 +124,11 @@ static char    *NextWord(char *s);
 static void	SetAppend(Ns_Set *set, int index, char *sep, char *value);
 static void	SetUpdate(Ns_Set *set, char *key, char *value);
 
-int Ns_ModuleVersion = 1;	
-
 
 /*
  *----------------------------------------------------------------------
  *
- * Ns_ModuleInit --
+ * NsCgi_ModInit --
  *
  *	Create a new CGI module instance.  Note: This module can
  *	be loaded multiple times.
@@ -157,32 +143,13 @@ int Ns_ModuleVersion = 1;
  */
  
 int
-Ns_ModuleInit(char *server, char *module)
+NsCgi_ModInit(char *server, char *module)
 {
     char           *path, *key, *value, *section;
-    int             i;
+    int             i, ok;
     Ns_Set         *set;
     Ns_DString      ds;
     Mod		   *modPtr;
-    static int	    initialized;
-
-    /*
-     * On the first (and likely only) load, register
-     * the temp file cleanup routine and open devNull
-     * for requests without content data.
-     */
-
-    if (!initialized) {
-	devNull = open(DEVNULL, O_RDONLY);
-	if (devNull < 0) {
-	    Ns_Log(Error, "nscgi: open(%s) failed: %s",
-		   DEVNULL, strerror(errno));
-	    return NS_ERROR;
-	}
-	Ns_DupHigh(&devNull);
-	Ns_CloseOnExec(devNull);
-	initialized = 1;
-    }
 
     /*
      * Config basic options.
@@ -192,11 +159,6 @@ Ns_ModuleInit(char *server, char *module)
     modPtr = ns_calloc(1, sizeof(Mod));
     modPtr->module = module;
     modPtr->server = server;
-    Ns_MutexInit(&modPtr->lock);
-    Ns_MutexSetName2(&modPtr->lock, "nscgi", server);
-    if (!Ns_ConfigGetInt(path, "maxinput", &modPtr->maxInput)) {
-        modPtr->maxInput = DEFAULT_MAXINPUT;
-    }
     if (!Ns_ConfigGetInt(path, "limit", &modPtr->maxCgi)) {
         modPtr->maxCgi = 0;
     }
@@ -246,16 +208,22 @@ Ns_ModuleInit(char *server, char *module)
      * Register all requested mappings.
      */
 
+    ok = 0;
     set = Ns_ConfigGetSection(path);
     for (i = 0; set != NULL && i < Ns_SetSize(set); ++i) {
         key = Ns_SetKey(set, i);
         value = Ns_SetValue(set, i);
         if (STRIEQ(key, "map")) {
-            CgiRegister(modPtr, value);
+            ok += CgiRegister(modPtr, value);
         }
     }
     Ns_DStringFree(&ds);
-
+    if (!ok) {
+	ns_free(modPtr);
+    } else {
+    	Ns_MutexInit(&modPtr->lock);
+    	Ns_MutexSetName2(&modPtr->lock, "nscgi", server);
+    }
     return NS_OK;
 }
 
@@ -291,9 +259,6 @@ CgiRequest(void *arg, Ns_Conn *conn)
      * Check for input overflow and initialize the CGI context.
      */
 
-    if (modPtr->maxInput > 0 && conn->contentLength > modPtr->maxInput) {
-        return Ns_ReturnBadRequest(conn, "Exceeded maximum CGI input size");
-    }
     if (CgiInit(&cgi, mapPtr, conn) != NS_OK) {
 	return Ns_ReturnNotFound(conn);
     } else if (cgi.interp == NULL && access(cgi.exec, X_OK) != 0) {
@@ -314,16 +279,17 @@ CgiRequest(void *arg, Ns_Conn *conn)
     }
 
     /*
-     * Spool input to temp file if necessary.
+     * Get the content fd if necessary.
      */
 
-    if (conn->contentLength > 0 && CgiSpool(&cgi, conn) != NS_OK) {
-	if (cgi.flags & CGI_ECONTENT) {
-	    status = Ns_ConnReturnBadRequest(conn, "Insufficient Content");
-	} else {
+    if (conn->contentLength <= 0) {
+	cgi.ifd = Ns_DevNull();
+    } else {
+	cgi.ifd = Ns_ConnContentFd(conn);
+	if (cgi.ifd < 0) {
 	    status = Ns_ConnReturnInternalError(conn);
+	    goto done;
 	}
-	goto done;
     }
 
     /*
@@ -559,52 +525,6 @@ err:
     return NS_ERROR;
 }
 
-
-/*
- *----------------------------------------------------------------------
- *
- * CgiSpool --
- *
- *	Spool content to a temp file.
- *
- * Results:
- *	File descriptor of temp file or -1 on error.
- *
- * Side effects:
- *	May open a new temp file.
- *
- *----------------------------------------------------------------------
- */
-
-static int
-CgiSpool(Cgi *cgiPtr, Ns_Conn *conn)
-{
-    int     len, fd;
-    char   *content, *err;
-
-    err = NULL;
-    len = conn->contentLength;
-    content = Ns_ConnContent(conn);
-    fd = Ns_GetTemp();
-    if (fd < 0) {
-	Ns_Log(Error, "nscgi: could not allocate temp file.");
-    } else if (write(fd, content, (size_t)len) != len) {
-	err = "write";
-    } else if (lseek(fd, 0, SEEK_SET) != 0) {
-	err = "lseek";
-    }
-    if (err != NULL) {
-	Ns_Log(Error, "nscgi: temp file %s failed: %s", err, strerror(errno));
-	close(fd);
-	fd = -1;
-    }
-    if (fd < 0) {
-	return NS_ERROR;
-    }
-    cgiPtr->ifd = fd;
-    return NS_OK;
-}
-
 
 /*
  *----------------------------------------------------------------------
@@ -656,14 +576,6 @@ CgiFree(Cgi *cgiPtr)
     	close(cgiPtr->ofd);
     }
         
-    /*
-     * Release the temp file.
-     */
-
-    if (cgiPtr->ifd >= 0) {
-	Ns_ReleaseTemp(cgiPtr->ifd);
-    }
-     
     /*
      * Free the environment.
      */
@@ -712,6 +624,7 @@ static int
 CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
 {
     int i, index, opipe[2];
+    int https = 0;
     char *s, *e, *p, **envp;
     Ns_DString *dsPtr;
     Mod *modPtr = cgiPtr->modPtr;
@@ -771,6 +684,7 @@ CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
      */
 
     SetUpdate(cgiPtr->env, "SCRIPT_NAME", cgiPtr->name);
+    SetUpdate(cgiPtr->env, "SCRIPT_FILENAME", cgiPtr->path);
     if (cgiPtr->pathinfo != NULL && *cgiPtr->pathinfo != '\0') {
     	Ns_DString tmp;
 	
@@ -796,6 +710,8 @@ CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
     Ns_DStringVarAppend(dsPtr, Ns_InfoServer(), "/", Ns_InfoVersion(), NULL);
     SetUpdate(cgiPtr->env, "SERVER_SOFTWARE", dsPtr->string);
     Ns_DStringTrunc(dsPtr, 0);
+    SetUpdate(cgiPtr->env, "SERVER_ROOT", Ns_InfoHomePath());
+    SetUpdate(cgiPtr->env, "DOCUMENT_ROOT", Ns_PageRoot(modPtr->server));
     Ns_DStringPrintf(dsPtr, "HTTP/%2.1f", conn->request->version);
     SetUpdate(cgiPtr->env, "SERVER_PROTOCOL", dsPtr->string);
     Ns_DStringTrunc(dsPtr, 0);
@@ -811,6 +727,12 @@ CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
             Ns_Log(Warning, "nscgi: location does not contain '://'");
             s = NULL;
         } else {
+            if (strncmp(s, "https://", 8) == 0) {
+                SetUpdate(cgiPtr->env, "HTTPS", "on");
+                https = 1;
+            } else {
+                SetUpdate(cgiPtr->env, "HTTPS", "off");
+            }
             s = strchr(s, ':');         /* Get past the http */
             if (s != NULL) {
                 s += 3;                 /* Get past the // */
@@ -848,7 +770,11 @@ CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
         }
     }
     if (s == NULL) {
-        s = "80";
+        if (https) {
+            s = "443";
+        } else {
+            s = "80";
+        }
     }
     SetUpdate(cgiPtr->env, "SERVER_PORT", s);
     SetUpdate(cgiPtr->env, "AUTH_TYPE", "Basic");
@@ -956,9 +882,8 @@ CgiExec(Cgi *cgiPtr, Ns_Conn *conn)
      * Execute the CGI.
      */
      
-    cgiPtr->pid = Ns_ExecProcess(cgiPtr->exec, cgiPtr->dir,
-	cgiPtr->ifd < 0 ? devNull : cgiPtr->ifd,
-	opipe[1], dsPtr->string, cgiPtr->env);
+    cgiPtr->pid = Ns_ExecProcess(cgiPtr->exec, cgiPtr->dir, cgiPtr->ifd,
+				 opipe[1], dsPtr->string, cgiPtr->env);
     close(opipe[1]);
     if (cgiPtr->pid < 0) {
     	close(opipe[0]);
@@ -1102,7 +1027,7 @@ CgiCopy(Cgi *cgiPtr, Ns_Conn *conn)
                 ++value;
             }
             if (STRIEQ(ds.string, "status")) {
-                httpstatus = atoi(value);
+                httpstatus = (int) strtol(value, NULL, 10);
             } else if (STRIEQ(ds.string, "location")) {
                 httpstatus = 302;
                 if (*value == '/') {
@@ -1196,9 +1121,10 @@ NextWord(char *s)
  *----------------------------------------------------------------------
  */
 
-static void
+static int
 CgiRegister(Mod *modPtr, char *map)
 {
+    int		    ok = 0;
     char           *method;
     char           *url;
     char           *path;
@@ -1227,7 +1153,6 @@ CgiRegister(Mod *modPtr, char *map)
 	    goto done;
 	}
     }
-
     mapPtr = ns_malloc(sizeof(Map));
     mapPtr->modPtr = modPtr;
     mapPtr->url = ns_strdup(url);
@@ -1236,10 +1161,12 @@ CgiRegister(Mod *modPtr, char *map)
 	   method, url, path ? " -> " : "", path ? path : "");
     Ns_RegisterRequest(modPtr->server, method, url, 
 		       CgiRequest, CgiFreeMap, mapPtr, 0);
+    ok = 1;
 
 done:
     Ns_DStringFree(&ds1);
     Ns_DStringFree(&ds2);
+    return ok;
 }
 
 
