@@ -35,7 +35,7 @@
  *  	Tcl commands.
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nscp/nscp.c,v 1.22 2003/03/07 18:08:12 vasiljevic Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nscp/nscp.c,v 1.24 2005/08/08 15:32:34 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "ns.h"
 
@@ -92,13 +92,11 @@ static unsigned char dont_echo[]  = {TN_IAC, TN_DONT, TN_ECHO};
 static unsigned char will_echo[]  = {TN_IAC, TN_WILL, TN_ECHO};
 static unsigned char wont_echo[]  = {TN_IAC, TN_WONT, TN_ECHO};
 
-int Ns_ModuleVersion = 1;
-
 
 /*
  *----------------------------------------------------------------------
  *
- * Ns_ModuleInit --
+ * NsCp_ModInit --
  *
  *	Load the config parameters, setup the structures, and
  *	listen on the control port.
@@ -114,7 +112,7 @@ int Ns_ModuleVersion = 1;
  */
  
 int
-Ns_ModuleInit(char *server, char *module)
+NsCp_ModInit(char *server, char *module)
 {
     Mod *modPtr;
     char *path, *addr, *pass, *user, *key, *end;
