@@ -33,149 +33,160 @@
  *	ADP string and file eval.
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/adpeval.c,v 1.28 2003/10/07 15:08:36 mpagenva Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/adpeval.c,v 1.48 2006/06/27 16:34:49 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
+/*
+ * The following structure defines a cached ADP page result.  A cached
+ * object is created by executing the non-cached code and saving the
+ * resulting output which may include embedded non-cached components
+ * (see NsTclAdpIncludeObjCmd for details).
+ */
+
+typedef struct AdpCache {
+    int	      	   refcnt;	/* Current interps using cached results. */ 
+    Ns_Time	   expires;	/* Expiration time of cached results. */
+    AdpCode	   code;	/* ADP code for cached result. */
+} AdpCache;
+
 /*
  * The following structure defines a shared page in the ADP cache.  The
- * bytes for the filename as well as the block lengths array and page text 
- * referenced in the code structure are allocated with the Page struct
- * as one contiguous block of memory.
+ * size of the object is extended to include the filename bytes.
  */
 
 typedef struct Page {
-    NsServer	  *servPtr;
-    Tcl_HashEntry *hPtr;
-    time_t    	   mtime;
-    off_t     	   size;
-    int	      	   refcnt;
-    int		   evals;
-    char	  *file;
-    AdpCode	   code;
+    NsServer	  *servPtr;	/* Page server context (reg tags, etc.) */
+    Tcl_HashEntry *hPtr;	/* Entry in shared table of all pages. */
+    time_t    	   mtime;	/* Original modify time of file. */
+    off_t     	   size;	/* Original size of file. */
+    int		   flags;	/* Flags used on last compile, e.g., SAFE. */
+    int	      	   refcnt;	/* Refcnt of current interps using page. */
+    int		   evals;	/* Count of page evaluations. */
+    int		   locked;	/* Page locked for cache update. */
+    int		   cacheGen;	/* Cache generation id. */
+    AdpCache	  *cachePtr;	/* Cached output. */
+    AdpCode	   code;	/* ADP code blocks. */
+    char	   file[1];	/* First known filename for stats reporting. */
 } Page;
 
+/*
+ * The following structure holds per-interp script byte codes.  The
+ * size of the object is extended based on the number of script objects.
+ */
+
+typedef struct Objs {
+    int nobjs;			/* Number of scripts objects. */
+    Tcl_Obj *objs[1];		/* Scripts to be compiled and reused. */
+} Objs;
+
 /*
  * The following structure defines a per-interp page entry with
- * a pointer to the shared Page and private script Tcl_Obj's.
+ * a pointer to the shared Page and private Objs for cached and
+ * non-cached page results.
  */
 
 typedef struct InterpPage {
-    Page     *pagePtr;
-    Tcl_Obj  *objs[1];
+    Page     *pagePtr;		/* Pointer to shared page text. */
+    Objs     *objs;		/* Non-cache ADP code script. */
+    int	      cacheGen;		/* Cache generation id. */
+    Objs     *cacheObjs;	/* Cache results ADP code scripts. */
 } InterpPage;
 
-/*
- * The following structure maintains an ADP call frame.  PushFrame
- * is used to save the previous state of the per-thread NsInterp
- * structure in a Frame allocated on the stack and PopFrame restores
- * the previous state from the Frame.
- */
-
-typedef struct Frame {
-    int                objc;
-    Tcl_Obj          **objv;
-    char              *cwd;
-    Ns_DString         cwdBuf;
-    Tcl_DString	      *outputPtr;
-} Frame;
-
-
-/*
- * Constants to support construction of evaluation error messages
- */
-#define ADPEVAL_MAX_SCRIPT_TEXT 150
-#define ADPEVAL_ELIPSIS_LEN     3
-
-
 /*
  * Local functions defined in this file.
  */
 
-static Page *ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr);
-static void PushFrame(NsInterp *itPtr, Frame *framePtr, char *file, 
-		      int objc, Tcl_Obj *objv[], Tcl_DString *outputPtr);
-static void PopFrame(NsInterp *itPtr, Frame *framePtr);
-static void LogError(NsInterp *itPtr, int nscript);
-static int AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
-		Tcl_DString *outputPtr);
-static int AdpEval(NsInterp *itPtr, AdpCode *codePtr, Tcl_Obj *objs[]);
-static void ParseFree(AdpParse *parsePtr);
+static Page *ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr,
+		       int flags);
+static int AdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int flags,
+		   char *resvar);
+static int AdpExec(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *file,
+		   AdpCode *codePtr, Objs *objsPtr, Tcl_DString *outputPtr);
+static int AdpSource(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *file,
+		     Ns_Time *ttlPtr, int flags, Tcl_DString *outputPtr);
 static int AdpDebug(NsInterp *itPtr, char *ptr, int len, int nscript);
+static void DecrCache(AdpCache *cachePtr);
+static Objs *AllocObjs(int nobjs);
+static void FreeObjs(Objs *objsPtr);
+static void AdpTrace(NsInterp *itPtr, char *ptr, int len);
 static Ns_Callback FreeInterpPage;
 
 
 /*
  *----------------------------------------------------------------------
  *
- * NsAdpEval --
+ * NsAdpEval, NsAdpSource --
  *
- *	Evaluate an ADP string.
+ *	Evaluate an ADP string or file and return the output
+ *	as the interp result.
  *
  * Results:
  *	A standard Tcl result.
  *
  * Side effects:
- *	String is parsed and evaluated at current Tcl level in a
- *	new ADP call frame.
+ *	Variable named by resvar, if any, is updated with results of
+ *	Tcl interp before being replaced with ADP output.
  *
  *----------------------------------------------------------------------
  */
 
 int
-NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int safe, char *resvar)
+NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int flags, char *resvar)
 {
-    AdpParse	      parse;
-    Frame             frame;
+    return AdpEval(itPtr, objc, objv, flags, resvar);
+}
+
+int
+NsAdpSource(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int flags, char *resvar)
+{
+    return AdpEval(itPtr, objc, objv, (flags | ADP_EVAL_FILE), resvar);
+}
+
+static int
+AdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int flags, char *resvar)
+{
+    Tcl_Interp	     *interp = itPtr->interp;
+    AdpCode	      code;
     Tcl_DString       output;
+    Tcl_Obj	     *objPtr;
     int               result;
-    Tcl_Obj          *resPtr;
-    bool              lcl_resp = NS_FALSE;
+    char	     *obj0;
     
     /*
-     * Push a frame, execute the code, and then move any result to the
-     * interp from the local output buffer.
+     * If the ADP object is a file, simply source it. Otherwise, parse
+     * the script as a temporary ADP code object and execute it directly.
      */
-     
+
     Tcl_DStringInit(&output);
+    obj0 = Tcl_GetString(objv[0]);
+    if (flags & ADP_EVAL_FILE) {
+    	result = AdpSource(itPtr, objc, objv, obj0, NULL, flags, &output);
+    } else {
+    	NsAdpParse(&code, itPtr->servPtr, obj0, flags);
+    	result = AdpExec(itPtr, objc, objv, NULL, &code, NULL, &output);
+    	NsAdpFreeCode(&code);
+    }
 
     /*
-     * If we are not acting within a context which has set up the adp
-     * output buffer, then hook the responsePtr to our lcl buffer
-     * for the processing of this request.  It will be unhooked after.
+     * Set the interp result with the ADP output, saving the last interp
+     * result first if requested.
      */
-    if (itPtr->adp.responsePtr == NULL) {
-        itPtr->adp.responsePtr = &output;
-        lcl_resp = NS_TRUE;
-    }
-
-    PushFrame(itPtr, &frame, NULL, objc, objv, &output);
-    NsAdpParse(&parse, itPtr->servPtr, Tcl_GetString(objv[0]), safe);
-    result = AdpEval(itPtr, &parse.code, NULL);
-    PopFrame(itPtr, &frame);
-
-    if (lcl_resp) {
-        itPtr->adp.responsePtr = NULL;
-    }
 
     if (result == TCL_OK) {
-        /*
-         * If the caller has supplied a variable for the adp's result value,
-         * then save the interp's result there prior to overwritting it.
-         */
-        if (resvar != NULL) {
-            resPtr = Tcl_GetObjResult(itPtr->interp);
-            if (Tcl_SetVar2Ex(itPtr->interp, resvar, NULL, resPtr,
-                              TCL_LEAVE_ERR_MSG) == NULL) {
-                return TCL_ERROR;
-            }
-        }
-
-        Tcl_SetResult(itPtr->interp, output.string, TCL_VOLATILE);
+	if (resvar != NULL) {
+	    objPtr = Tcl_GetObjResult(interp);
+	    if (Tcl_SetVar2Ex(interp, resvar, NULL, objPtr, TCL_LEAVE_ERR_MSG)
+			      == NULL) {
+		result = TCL_ERROR;
+	    }
+	}
+	if (result == TCL_OK) {
+	    objPtr = Tcl_NewStringObj(output.string, output.length);
+	    Tcl_SetObjResult(interp, objPtr);
+	}
     }
-
     Tcl_DStringFree(&output);
-    ParseFree(&parse);
     return result;
 }
 
@@ -183,7 +194,7 @@ NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int safe, char *resvar)
 /*
  *----------------------------------------------------------------------
  *
- * NsAdpSource, NsAdpInclude --
+ * NsAdpInclude --
  *
  *	Evaluate an ADP file, utilizing per-thread byte-code pages.
  *
@@ -191,96 +202,139 @@ NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int safe, char *resvar)
  *	A standard Tcl result.
  *
  * Side effects:
- *	Output is either left in the ADP buffer (NsAdpInclude) or
- *	moved to the interp result (NsAdpSource).
+ *	Output is either left in current ADP buffer.
  *
  *----------------------------------------------------------------------
  */
 
 int
-NsAdpSource(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *resvar)
+NsAdpInclude(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *file,
+		Ns_Time *ttlPtr)
 {
-    Tcl_DString output;
-    int code;
-    Tcl_Obj     *resPtr;
-    bool         lcl_resp = NS_FALSE;
+    Ns_DString *outputPtr;
+    int flags = itPtr->adp.flags;
 
     /*
-     * Direct output to a local buffer.
+     * If an ADP execution is already active, use the current output
+     * buffer. Otherwise, use the top-level buffer in the ADP struct.
      */
 
-    Tcl_DStringInit(&output);
-
-    /*
-     * If we are not acting within a context which has set up the adp
-     * output buffer, then hook the responsePtr to our lcl buffer
-     * for the processing of this request.  It will be unhooked after.
-     */
-    if (itPtr->adp.responsePtr == NULL) {
-        itPtr->adp.responsePtr = &output;
-        lcl_resp = NS_TRUE;
+    if (itPtr->adp.framePtr != NULL) {
+	outputPtr = itPtr->adp.framePtr->outputPtr;
+    } else {
+	outputPtr = &itPtr->adp.output;
     }
-
-    code = AdpRun(itPtr, Tcl_GetString(objv[0]), objc, objv, &output);
-
-    if (lcl_resp) {
-        itPtr->adp.responsePtr = NULL;
-    }
-
-    if (code == TCL_OK) {
-        /*
-         * If the caller has supplied a variable for the adp's result value,
-         * then save the interp's result there prior to overwritting it.
-         */
-        if (resvar != NULL) {
-            resPtr = Tcl_GetObjResult(itPtr->interp);
-            if (Tcl_SetVar2Ex(itPtr->interp, resvar, NULL, resPtr,
-                              TCL_LEAVE_ERR_MSG) == NULL) {
-                return TCL_ERROR;
-            }
-        }
-
-	Tcl_DStringResult(itPtr->interp, &output);
-    }
-    Tcl_DStringFree(&output);
-    return code;
+    return AdpSource(itPtr, objc, objv, file, ttlPtr, flags, outputPtr);
 }
 
-int
-NsAdpInclude(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[])
-{
-    /*
-     * Direct output to the current ADP output buffer.
-     */
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsAdpInit, NsAdpFree --
+ *
+ *	Initialize or free the NsInterp ADP data structures.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *  	None.
+ *
+ *----------------------------------------------------------------------
+ */
 
-    if (itPtr->adp.outputPtr == NULL) {
-	Tcl_SetResult(itPtr->interp, "no connection", TCL_STATIC);
-	return TCL_ERROR;
-    }
-    return AdpRun(itPtr, file, objc, objv, itPtr->adp.outputPtr);
+void
+NsAdpInit(NsInterp *itPtr)
+{
+    Tcl_DStringInit(&itPtr->adp.output);
+    NsAdpReset(itPtr);
 }
 
+void
+NsAdpFree(NsInterp *itPtr)
+{
+    if (itPtr->adp.cache != NULL) {
+	Ns_CacheDestroy(itPtr->adp.cache);
+    }
+    Tcl_DStringFree(&itPtr->adp.output);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsAdpReset --
+ *
+ *	Reset the NsInterp ADP data structures for the next
+ *	execution request.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+NsAdpReset(NsInterp *itPtr)
+{
+    itPtr->adp.exception = ADP_OK;
+    itPtr->adp.debugLevel = 0;
+    itPtr->adp.debugInit = 0;
+    itPtr->adp.debugFile = NULL;
+    itPtr->adp.chan = NULL;
+    itPtr->adp.conn = NULL;
+    itPtr->adp.bufsize = itPtr->servPtr->adp.bufsize;
+    itPtr->adp.flags = itPtr->servPtr->adp.flags;
+    Tcl_DStringTrunc(&itPtr->adp.output, 0);
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * AdpSource --
+ *
+ *	Execute ADP code in a file with results returned in given
+ *	dstring.
+ *
+ * Results:
+ *	TCL_ERROR if the file could not be parsed, result of
+ *	AdpExec otherwise.
+ *
+ * Side effects:
+ *	Page text and ADP code results may be cached up to given time
+ *	limit, if any.
+ *
+ *----------------------------------------------------------------------
+ */
+
 static int
-AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
-       Tcl_DString *outputPtr)
+AdpSource(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *file,
+       Ns_Time *ttlPtr, int flags, Tcl_DString *outputPtr)
 {
     NsServer  *servPtr = itPtr->servPtr;
     Tcl_Interp *interp = itPtr->interp;
     Tcl_HashEntry *hPtr;
     struct stat st;
     Ns_DString tmp, path;
-    Frame frame;
     InterpPage *ipagePtr;
     Page *pagePtr, *oldPagePtr;
+    AdpCache *cachePtr;
+    AdpCode *codePtr;
+    Ns_Time now;
     Ns_Entry *ePtr;
-    int new, n;
+    Objs *objsPtr;
+    int new, cacheGen;
     char *p, *key;
     FileKey ukey;
-    int status;
+    int result;
 
     ipagePtr = NULL;
     pagePtr = NULL; 
-    status = TCL_ERROR;   /* assume error until accomplished success */
+    result = TCL_ERROR;   /* assume error until accomplished success */
     Ns_DStringInit(&tmp);
     Ns_DStringInit(&path);
     key = (char *) &ukey;
@@ -289,14 +343,11 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
      * Construct the full, normalized path to the ADP file.
      */
 
-    if (Ns_PathIsAbsolute(file)) {
-	Ns_NormalizePath(&path, file);
-    } else {
-	Ns_MakePath(&tmp, itPtr->adp.cwd, file, NULL);
-	Ns_NormalizePath(&path, tmp.string);
-	Ns_DStringTrunc(&tmp, 0);
+    if (!Ns_PathIsAbsolute(file)) {
+	file = Ns_MakePath(&tmp, itPtr->adp.cwd, file, NULL);
     }
-    file = path.string;
+    file = Ns_NormalizePath(&path, file);
+    Ns_DStringTrunc(&tmp, 0);
 
     /*
      * Check for TclPro debugging.
@@ -304,7 +355,7 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 
     if (itPtr->adp.debugLevel > 0) {
 	++itPtr->adp.debugLevel;
-    } else if (servPtr->adp.enabledebug != NS_FALSE &&
+    } else if ((flags & ADP_DEBUG) &&
 	itPtr->adp.debugFile != NULL &&
 	(p = strrchr(file, '/')) != NULL &&
 	Tcl_StringMatch(p+1, itPtr->adp.debugFile)) {
@@ -324,12 +375,12 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
     }
 
     if (itPtr->adp.cache == NULL) {
-	Ns_DStringPrintf(&tmp, "nsadp:%s:%p", itPtr->servPtr->server, itPtr);
+	Ns_DStringPrintf(&tmp, "nsadp:%p", itPtr);
 #ifdef _WIN32
-    itPtr->adp.cache = Ns_CacheCreateSz(tmp.string, TCL_STRING_KEYS,
+	itPtr->adp.cache = Ns_CacheCreateSz(tmp.string, TCL_STRING_KEYS,
 				itPtr->servPtr->adp.cachesize, FreeInterpPage);
 #else
-    itPtr->adp.cache = Ns_CacheCreateSz(tmp.string, FILE_KEYS,
+	itPtr->adp.cache = Ns_CacheCreateSz(tmp.string, FILE_KEYS,
 				itPtr->servPtr->adp.cachesize, FreeInterpPage);
 #endif
 	Ns_DStringTrunc(&tmp, 0);
@@ -350,7 +401,7 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 	 */
 	 
 #ifdef _WIN32
-    key = file;
+	key = file;
 #else
 	ukey.dev = st.st_dev;
 	ukey.ino = st.st_ino;
@@ -358,7 +409,9 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
         ePtr = Ns_CacheFindEntry(itPtr->adp.cache, key);
 	if (ePtr != NULL) {
     	    ipagePtr = Ns_CacheGetValue(ePtr);
-    	    if (ipagePtr->pagePtr->mtime != st.st_mtime || ipagePtr->pagePtr->size != st.st_size) {
+    	    if (ipagePtr->pagePtr->mtime != st.st_mtime
+			|| ipagePtr->pagePtr->size != st.st_size
+			|| ipagePtr->pagePtr->flags != flags) {
 		Ns_CacheFlushEntry(ePtr);
 		ipagePtr = NULL;
 	    }
@@ -375,7 +428,9 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 		Ns_CondWait(&servPtr->adp.pagecond, &servPtr->adp.pagelock);
 		hPtr = Tcl_CreateHashEntry(&servPtr->adp.pages, key, &new);
 	    }
-	    if (!new && (pagePtr->mtime != st.st_mtime || pagePtr->size != st.st_size)) {
+	    if (!new && (pagePtr->mtime != st.st_mtime
+			|| pagePtr->size != st.st_size
+			|| pagePtr->flags != flags)) {
 		/* NB: Clear entry to indicate read/parse in progress. */
 		Tcl_SetHashValue(hPtr, NULL);
 		pagePtr->hPtr = NULL;
@@ -383,23 +438,26 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 	    }
 	    if (new) {
 		Ns_MutexUnlock(&servPtr->adp.pagelock);
-		pagePtr = ParseFile(itPtr, file, &st);
+		pagePtr = ParseFile(itPtr, file, &st, flags);
 		Ns_MutexLock(&servPtr->adp.pagelock);
 		if (pagePtr == NULL) {
 		    Tcl_DeleteHashEntry(hPtr);
 		} else {
 #ifdef _WIN32
-            if (pagePtr->mtime != st.st_mtime || pagePtr->size != st.st_size) {
+            	    if (pagePtr->mtime != st.st_mtime
+				|| pagePtr->size != st.st_size)
 #else
-            if (ukey.dev != st.st_dev || ukey.ino != st.st_ino) {
+            	    if (ukey.dev != st.st_dev || ukey.ino != st.st_ino)
 #endif
+		    {
 			/* NB: File changed between stat above and ParseFile. */
 		    	Tcl_DeleteHashEntry(hPtr);
 #ifndef _WIN32
-                ukey.dev = st.st_dev;
+                	ukey.dev = st.st_dev;
 		    	ukey.ino = st.st_ino;
 #endif
-		    	hPtr = Tcl_CreateHashEntry(&servPtr->adp.pages, key, &new);
+		    	hPtr = Tcl_CreateHashEntry(&servPtr->adp.pages, key,
+						   &new);
 		    	if (!new) {
 			    oldPagePtr = Tcl_GetHashValue(hPtr);
 			    oldPagePtr->hPtr = NULL;
@@ -415,14 +473,17 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 	    }
 	    Ns_MutexUnlock(&servPtr->adp.pagelock);
 	    if (pagePtr != NULL) {
-		n = sizeof(Tcl_Obj *) * pagePtr->code.nscripts;
-	    	ipagePtr = ns_calloc(1, sizeof(InterpPage) + n);
+	    	ipagePtr = ns_malloc(sizeof(InterpPage));
 		ipagePtr->pagePtr = pagePtr;
+		ipagePtr->cacheGen = 0;
+		ipagePtr->objs = AllocObjs(pagePtr->code.nscripts);
+		ipagePtr->cacheObjs = NULL;
         	ePtr = Ns_CacheCreateEntry(itPtr->adp.cache, key, &new);
 		if (!new) {
 		    Ns_CacheUnsetValue(ePtr);
 		}
-        Ns_CacheSetValueSz(ePtr, ipagePtr, (size_t)ipagePtr->pagePtr->size);
+        	Ns_CacheSetValueSz(ePtr, ipagePtr,
+				   (size_t) ipagePtr->pagePtr->size);
 	    }
 	}
     }
@@ -432,11 +493,88 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
      */
          
     if (ipagePtr != NULL) {
-    	PushFrame(itPtr, &frame, file, objc, objv, outputPtr);
-	status = AdpEval(itPtr, &ipagePtr->pagePtr->code, ipagePtr->objs);
-    	PopFrame(itPtr, &frame);
+	pagePtr = ipagePtr->pagePtr;
+	if (ttlPtr == NULL || (flags & ADP_NOCACHE)) {
+	   cachePtr = NULL;
+	} else {
+	    Ns_MutexLock(&servPtr->adp.pagelock);
+
+	    /*
+	     * First, wait for an initial cache if already executing.
+	     */
+
+	    while ((cachePtr = pagePtr->cachePtr) == NULL && pagePtr->locked) {
+		Ns_CondWait(&servPtr->adp.pagecond, &servPtr->adp.pagelock);
+	    }
+
+	    /*
+	     * Next, if a cache exists and isn't locked, check expiration.
+	     */
+
+	    if (cachePtr != NULL && !pagePtr->locked) {
+		Ns_GetTime(&now);
+		if (Ns_DiffTime(&cachePtr->expires, &now, NULL) < 0) {
+		    pagePtr->locked = 1;
+		    cachePtr = NULL;
+		}
+	    }
+
+	    /*
+	     * Create the cached page if necessary.
+	     */
+
+	    if (cachePtr == NULL) {
+	    	Ns_MutexUnlock(&servPtr->adp.pagelock);
+		codePtr = &pagePtr->code;
+		++itPtr->adp.refresh;
+		result = AdpExec(itPtr, objc, objv, file, codePtr,
+				 ipagePtr->objs, &tmp);
+		--itPtr->adp.refresh;
+		if (result == TCL_OK) {
+		    cachePtr = ns_malloc(sizeof(AdpCache));
+		    NsAdpParse(&cachePtr->code, itPtr->servPtr, tmp.string,
+			       flags);
+		    Ns_GetTime(&cachePtr->expires);
+		    Ns_IncrTime(&cachePtr->expires, ttlPtr->sec, ttlPtr->usec);
+	    	    cachePtr->refcnt = 1;
+		}
+		Ns_DStringTrunc(&tmp, 0);
+	    	Ns_MutexLock(&servPtr->adp.pagelock);
+		if (cachePtr != NULL) {
+		    if (pagePtr->cachePtr != NULL) {
+		    	DecrCache(pagePtr->cachePtr);
+		    }
+		    ++pagePtr->cacheGen;
+		    pagePtr->cachePtr = cachePtr;
+		}
+		pagePtr->locked = 0;
+		Ns_CondBroadcast(&servPtr->adp.pagecond);
+	    }
+	    cacheGen = pagePtr->cacheGen;
+	    ++cachePtr->refcnt;
+	    Ns_MutexUnlock(&servPtr->adp.pagelock);
+	}
+	if (cachePtr == NULL) {
+	   codePtr = &pagePtr->code;
+	   objsPtr = ipagePtr->objs;
+	} else {
+	    codePtr = &cachePtr->code;
+	    if (ipagePtr->cacheObjs != NULL && cacheGen != ipagePtr->cacheGen) {
+		FreeObjs(ipagePtr->cacheObjs);
+		ipagePtr->cacheObjs = NULL;
+	    }
+	    if (ipagePtr->cacheObjs == NULL) {
+		ipagePtr->cacheObjs = AllocObjs(AdpCodeScripts(codePtr));
+		ipagePtr->cacheGen = cacheGen;
+	    }
+	    objsPtr = ipagePtr->cacheObjs;
+	}
+	result = AdpExec(itPtr, objc, objv, file, codePtr, objsPtr, outputPtr);
 	Ns_MutexLock(&servPtr->adp.pagelock);
 	++ipagePtr->pagePtr->evals;
+	if (cachePtr != NULL) {
+	    DecrCache(cachePtr);
+	}
 	Ns_MutexUnlock(&servPtr->adp.pagelock);
     }
     if (itPtr->adp.debugLevel > 0) {
@@ -446,7 +584,7 @@ AdpRun(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[],
 done:
     Ns_DStringFree(&path);
     Ns_DStringFree(&tmp);
-    return status;
+    return result;
 }
 
 
@@ -486,20 +624,19 @@ NsAdpDebug(NsInterp *itPtr, char *host, char *port, char *procs)
 	code = Tcl_EvalEx(interp, ds.string, ds.length, 0);
         Tcl_DStringFree(&ds);
 	if (code != TCL_OK) {
-	    Ns_TclLogError(interp);
+	    NsAdpLogError(itPtr);
 	    return TCL_ERROR;
 	}
 
 	/*
-	 * Link the ADP response buffer result to a global variable
+	 * Link the ADP output buffer result to a global variable
 	 * which can be monitored with a variable watch.
 	 */
 
-	if (itPtr->adp.responsePtr != NULL &&
-	    Tcl_LinkVar(interp, "ns_adp_output",
-			(char *) &itPtr->adp.responsePtr->string,
+	if (Tcl_LinkVar(interp, "ns_adp_output",
+			(char *) &itPtr->adp.output.string,
 		TCL_LINK_STRING | TCL_LINK_READ_ONLY) != TCL_OK) {
-	    Ns_TclLogError(interp);
+	    NsAdpLogError(itPtr);
 	}
 
 	itPtr->adp.debugInit = 1;
@@ -543,10 +680,11 @@ NsTclAdpStatsCmd(ClientData arg, Tcl_Interp *interp, int argc, char **argv)
 	pagePtr = Tcl_GetHashValue(hPtr);
 	keyPtr = (FileKey *) Tcl_GetHashKey(&servPtr->adp.pages, hPtr);
 	Tcl_AppendElement(interp, pagePtr->file);
-	sprintf(buf, "dev %ld ino %ld mtime %ld refcnt %d evals %d size %ld blocks %d scripts %d",
-		keyPtr->dev, keyPtr->ino, pagePtr->mtime, pagePtr->refcnt,
-		pagePtr->evals, pagePtr->size, pagePtr->code.nblocks,
-		pagePtr->code.nscripts);
+	sprintf(buf, "dev %ld ino %ld mtime %ld refcnt %d evals %d "
+		     "size %ld blocks %d scripts %d",
+		(long) keyPtr->dev, (long) keyPtr->ino, (long) pagePtr->mtime,
+		pagePtr->refcnt, pagePtr->evals, (long) pagePtr->size,
+		pagePtr->code.nblocks, pagePtr->code.nscripts);
 	Tcl_AppendElement(interp, buf);
 	hPtr = Tcl_NextHashEntry(&search);
     }
@@ -554,93 +692,6 @@ NsTclAdpStatsCmd(ClientData arg, Tcl_Interp *interp, int argc, char **argv)
     return TCL_OK;
 }
 
-
-/*
- *----------------------------------------------------------------------
- *
- * PushFrame --
- *
- *	Push an ADP call frame on the ADP stack.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	The given Frame is initialized, the current working directory
- *	is determined from the absolute filename (if not NULL), the
- *	previous state of the per-thread NsInterp structure is saved
- *	and then updated with the current call's arguments.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-PushFrame(NsInterp *itPtr, Frame *framePtr, char *file, int objc,
-	  Tcl_Obj *objv[], Tcl_DString *outputPtr)
-{
-    char    *slash;
-
-    /*
-     * Save current NsInterp state.
-     */
-
-    framePtr->cwd = itPtr->adp.cwd;
-    framePtr->objc = itPtr->adp.objc;
-    framePtr->objv = itPtr->adp.objv;
-    framePtr->outputPtr = itPtr->adp.outputPtr;
-    itPtr->adp.outputPtr = outputPtr;
-    itPtr->adp.objc = objc;
-    itPtr->adp.objv = objv;
-    ++itPtr->adp.depth;
-
-    /*
-     * If file is not NULL it indicates a call from
-     * NsAdpSource or NsAdpInclude.  If so, update the
-     * current working directory based on the
-     * absolute file pathname.
-     */
-
-    Ns_DStringInit(&framePtr->cwdBuf);
-    if (file != NULL) {
-	slash = strrchr(file, '/');
-    	Ns_DStringNAppend(&framePtr->cwdBuf, file, slash - file);
-    	itPtr->adp.cwd = framePtr->cwdBuf.string;
-    }
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * PopFrame --
- *
- *	Pop a previously pushed ADP call frame from the ADP stack.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	Previous state of the per-thread NsInterp structure is restored
- *	and the Frame is free'ed.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-PopFrame(NsInterp *itPtr, Frame *framePtr)
-{
-    /*
-     * Restore the previous frame.
-     */
-
-    itPtr->adp.objc = framePtr->objc;
-    itPtr->adp.objv = framePtr->objv;
-    itPtr->adp.cwd = framePtr->cwd;
-    itPtr->adp.outputPtr = framePtr->outputPtr;
-    --itPtr->adp.depth;
-    Ns_DStringFree(&framePtr->cwdBuf);
-}
-
 
 /*
  *----------------------------------------------------------------------
@@ -660,7 +711,7 @@ PopFrame(NsInterp *itPtr, Frame *framePtr)
  */
 
 static Page *
-ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr)
+ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr, int flags)
 {
     Tcl_Interp *interp = itPtr->interp;
     Tcl_Encoding encoding;
@@ -669,7 +720,6 @@ ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr)
     int fd, n, trys;
     size_t size;
     Page *pagePtr;
-    AdpParse parse;
     
     fd = open(file, O_RDONLY | O_BINARY);
     if (fd < 0) {
@@ -710,7 +760,7 @@ ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr)
 	     * File is not expected size, rewind and fstat/read again.
 	     */
 	
-	    if (lseek(fd, 0, SEEK_SET) != 0) {
+	    if (lseek(fd, (off_t) 0, SEEK_SET) != 0) {
 	    	Tcl_AppendResult(interp, "could not lseek \"", file,
 	    	    "\": ", Tcl_PosixError(interp), NULL);
 	        goto done;
@@ -731,24 +781,19 @@ ParseFile(NsInterp *itPtr, char *file, struct stat *stPtr)
 	    Tcl_ExternalToUtfDString(encoding, buf, n, &utf);
 	    page = utf.string;
 	}
-	NsAdpParse(&parse, itPtr->servPtr, page, 0);
-	Tcl_DStringFree(&utf);
-	n = parse.hdr.length + parse.text.length + strlen(file) + 1;
-	pagePtr = ns_malloc(sizeof(Page) + n);
+	pagePtr = ns_malloc(sizeof(Page) + strlen(file));
+	strcpy(pagePtr->file, file);
 	pagePtr->servPtr = itPtr->servPtr;
+	pagePtr->flags = flags;
 	pagePtr->refcnt = 0;
 	pagePtr->evals = 0;
+	pagePtr->locked = 0;
+	pagePtr->cacheGen = 0;
+	pagePtr->cachePtr = NULL;
 	pagePtr->mtime = stPtr->st_mtime;
 	pagePtr->size = stPtr->st_size;
-	pagePtr->code.nblocks = parse.code.nblocks;
-	pagePtr->code.nscripts = parse.code.nscripts;
-	pagePtr->code.len = (int *) (pagePtr + 1);
-	pagePtr->code.base = (char *) (pagePtr->code.len + parse.code.nblocks);
-	pagePtr->file = pagePtr->code.base + parse.text.length;
-	memcpy(pagePtr->code.len, parse.hdr.string, (size_t)parse.hdr.length);
-	memcpy(pagePtr->code.base, parse.text.string, (size_t)parse.text.length);
-	strcpy(pagePtr->file, file);
-	ParseFree(&parse);
+	NsAdpParse(&pagePtr->code, itPtr->servPtr, page, flags);
+	Tcl_DStringFree(&utf);
     }
 
 done:
@@ -761,7 +806,7 @@ done:
 /*
  *----------------------------------------------------------------------
  *
- * LogError --
+ * NsAdpLogError --
  *
  *	Log an ADP error, possibly invoking the log handling ADP
  *	file if configured.
@@ -775,43 +820,76 @@ done:
  *----------------------------------------------------------------------
  */
 
-
-static void
-LogError(NsInterp *itPtr, int nscript)
+void
+NsAdpLogError(NsInterp *itPtr)
 {
     Tcl_Interp *interp = itPtr->interp;
+    Ns_Conn *conn = itPtr->conn; 
     Ns_DString ds;
     Tcl_Obj *objv[2];
-    char *file;
-    char *script;
-    char buffer[ADPEVAL_MAX_SCRIPT_TEXT+ADPEVAL_ELIPSIS_LEN+1];
+    AdpFrame *framePtr;
+    char *adp, *inc, *dot, *err;
+    int i, len;
     
+    framePtr = itPtr->adp.framePtr;
     Ns_DStringInit(&ds);
-    Ns_DStringAppend(&ds, "\n    invoked from within chunk: ");
-    Ns_DStringPrintf(&ds, "%d", nscript);
-    Ns_DStringAppend(&ds, " of adp: ");
-    /*
-     * Need to limit the amount of text put into the error.
-     */
-    script = Tcl_GetString(itPtr->adp.objv[0]);
-    if (strlen(script) > 150) {
-        sprintf(buffer, "%.*s...", ADPEVAL_MAX_SCRIPT_TEXT, script);
-        script = buffer;
+    Ns_DStringPrintf(&ds, "\n    at line %d of ",
+		     framePtr->line + interp->errorLine);
+    inc = "";
+    while (framePtr != NULL) {
+	if (framePtr->file != NULL) {
+	    Ns_DStringPrintf(&ds, "%sadp file \"%s\"", inc, framePtr->file);
+	    if (framePtr->ident != NULL) {
+		Ns_DStringPrintf(&ds, " {%s}", Tcl_GetString(framePtr->ident));
+	    }
+	} else {
+	    adp = Tcl_GetStringFromObj(framePtr->objv[0], &len);
+	    dot = "";
+	    if (len > 150) {
+		len = 150;
+		dot = "...";
+	    }
+	    while ((adp[len] & 0xC0) == 0x80) {
+		/* NB: Avoid truncating multi-byte UTF-8 character. */
+		len--;
+		dot = "...";
+	    }
+	    Ns_DStringPrintf(&ds, "%sadp script:\n\"%.*s%s\"",
+			     inc, len, adp, dot);
+	}
+	framePtr = framePtr->prevPtr;
+	inc = "\n    included from ";
+    }
+    if (conn != NULL && (itPtr->adp.flags & ADP_DETAIL)) {
+	Ns_DStringPrintf(&ds, "\n    while processing connection #%d:\n%8s%s",
+			 Ns_ConnId(conn), "",
+			 conn->request->line);
+	for (i = 0; i < Ns_SetSize(conn->headers); ++i) {
+	    Ns_DStringPrintf(&ds, "\n        %s: %s",
+			     Ns_SetKey(conn->headers, i),
+			     Ns_SetValue(conn->headers, i));
+	}
     }
-    Ns_DStringAppend(&ds, script);
     Tcl_AddErrorInfo(interp, ds.string);
-    Ns_TclLogError(interp);
+    err = Ns_TclLogError(interp);
+    if (itPtr->adp.flags & ADP_DISPLAY) {
+	Ns_DStringTrunc(&ds, 0);
+	Ns_DStringAppend(&ds, "<br><pre>\n");
+	Ns_QuoteHtml(&ds, err);
+	Ns_DStringAppend(&ds, "\n<br></pre>\n");
+	NsAdpAppend(itPtr, ds.string, ds.length);
+    }
     Ns_DStringFree(&ds);
-    file = itPtr->servPtr->adp.errorpage;
-    if (file != NULL && itPtr->adp.errorLevel == 0) {
+    adp = itPtr->servPtr->adp.errorpage;
+    if (adp != NULL && itPtr->adp.errorLevel == 0) {
 	++itPtr->adp.errorLevel;
-	objv[0] = Tcl_NewStringObj(file, -1);
+	objv[0] = Tcl_NewStringObj(adp, -1);
 	Tcl_IncrRefCount(objv[0]);
 	objv[1] = Tcl_GetVar2Ex(interp, "errorInfo", NULL, TCL_GLOBAL_ONLY);
 	if (objv[1] == NULL) {
 	    objv[1] = Tcl_GetObjResult(interp);
 	}
-	(void) NsAdpInclude(itPtr, file, 2, objv);
+	(void) NsAdpInclude(itPtr, 2, objv, adp, NULL);
 	Tcl_DecrRefCount(objv[0]);
 	--itPtr->adp.errorLevel;
     }
@@ -821,13 +899,14 @@ LogError(NsInterp *itPtr, int nscript)
 /*
  *----------------------------------------------------------------------
  *
- * AdpEval --
+ * AdpExec --
  *
- *	Evaluate page code.
+ *	Execute ADP code.
  *
  * Results:
- *	A Tcl status code.
- *
+ *	TCL_OK unless there is an ADP error exception, stack overflow,
+ *	or script error when the ADP_STRICT option is set.
+ *	
  * Side Effects:
  *	Depends on page.
  *
@@ -835,57 +914,122 @@ LogError(NsInterp *itPtr, int nscript)
  */
 
 static int
-AdpEval(NsInterp *itPtr, AdpCode *codePtr, Tcl_Obj **objs)
+AdpExec(NsInterp *itPtr, int objc, Tcl_Obj *objv[], char *file,
+	AdpCode *codePtr, Objs *objsPtr, Tcl_DString *outputPtr)
 {
     Tcl_Interp *interp = itPtr->interp;
+    AdpFrame frame;
+    Ns_DString cwd;
     Tcl_Obj *objPtr;
-    int nscript, result, len, i;
-    char *ptr;
+    int nscript, nblocks, result, len, i;
+    char *ptr, *slash, *savecwd;
 
-    ptr = codePtr->base;
+    /*
+     * Check for stack overflow.
+     */
+
+    if (Ns_CheckStack() == NS_BREAK) {
+	Tcl_SetResult(interp, "adp stack overflow", TCL_STATIC);
+	return TCL_ERROR;
+    }
+
+    /*
+     * Setup the new call frame.
+     */
+
+    Ns_DStringInit(&cwd);
+    frame.file = file;
+    frame.objc = objc;
+    frame.objv = objv;
+    frame.outputPtr = outputPtr;
+    frame.ident = NULL;
+    savecwd = itPtr->adp.cwd;
+    if (file != NULL && (slash = strrchr(file, '/')) != NULL) {
+    	Ns_DStringNAppend(&cwd, file, slash - file);
+	itPtr->adp.cwd = cwd.string;
+    }
+    frame.prevPtr = itPtr->adp.framePtr;
+    itPtr->adp.framePtr = &frame;
+    itPtr->adp.depth++;
+
+    /*
+     * Execute the ADP by copying text blocks directly to the output
+     * stream and evaluating script blocks.
+     */
+
+    ptr = AdpCodeText(codePtr);
+    nblocks = AdpCodeBlocks(codePtr);
     nscript = 0;
     result = TCL_OK;
-    for (i = 0; itPtr->adp.exception == ADP_OK && i < codePtr->nblocks; ++i) {
-	len = codePtr->len[i];
+    for (i = 0; itPtr->adp.exception == ADP_OK && i < nblocks; ++i) {
+	frame.line = AdpCodeLine(codePtr, i);
+	len = AdpCodeLen(codePtr, i);
+	if (itPtr->adp.flags & ADP_TRACE) {
+	    AdpTrace(itPtr, ptr, len);
+	}
 	if (len > 0) {
-            /*
-             * outputPtr can have been cleared on previous iterations through
-             * this loop.  Need to check it before trying to use it.
-             */
-            if( itPtr->adp.outputPtr != NULL ) {
-                Ns_DStringNAppend(itPtr->adp.outputPtr, ptr, len);
-            }
+	    result = NsAdpAppend(itPtr, ptr, len);
 	} else {
 	    len = -len;
 	    if (itPtr->adp.debugLevel > 0) {
     	        result = AdpDebug(itPtr, ptr, len, nscript);
-	    } else if (objs == NULL) {
+	    } else if (objsPtr == NULL) {
 		result = Tcl_EvalEx(interp, ptr, len, 0);
 	    } else {
-		objPtr = objs[nscript];
-		if (objPtr != NULL) {
-    	            result = Tcl_EvalObjEx(interp, objPtr, 0);
-		} else {
+		objPtr = objsPtr->objs[nscript];
+		if (objPtr == NULL) {
 		    objPtr = Tcl_NewStringObj(ptr, len);
 		    Tcl_IncrRefCount(objPtr);
-    	            result = Tcl_EvalObjEx(interp, objPtr, 0);
-		    objs[nscript] = objPtr;
+		    objsPtr->objs[nscript] = objPtr;
 		}
-	    }
-	    if (result != TCL_OK && result != TCL_RETURN
-		    && itPtr->adp.exception == ADP_OK) {
-    	    	LogError(itPtr, nscript);
+    	        result = Tcl_EvalObjEx(interp, objPtr, 0);
 	    }
 	    ++nscript;
 	}
+
+	/*
+	 * Log an error message and optionally break from this ADP
+	 * call frame unless the error was generated to signal
+	 * and ADP exception.
+	 */
+	
+	if (result != TCL_OK && itPtr->adp.exception == ADP_OK) {
+	    if (!(itPtr->adp.flags & ADP_ERRLOGGED)) {
+	    	NsAdpLogError(itPtr);
+	    }
+	    if (itPtr->adp.flags & ADP_STRICT) {
+    	    	itPtr->adp.flags |= ADP_ERRLOGGED;
+		break;
+	    }
+	}
 	ptr += len;
-	NsAdpFlush(itPtr);
     }
-    if (itPtr->adp.exception == ADP_RETURN) {
+
+    /*
+     * Clear the return exception and reset result.
+     */
+
+    switch (itPtr->adp.exception) {
+    case ADP_OK:
+	break;
+    case ADP_RETURN:
 	itPtr->adp.exception = ADP_OK;
+	/* FALLTHROUGH */
+    default: 
 	result = TCL_OK;
+	break;
     }
-    NsAdpFlush(itPtr);
+
+    /*
+     * Restore the previous call frame.
+     */
+
+    itPtr->adp.framePtr = frame.prevPtr;
+    itPtr->adp.cwd = savecwd;
+    if (frame.ident != NULL) {
+	Tcl_DecrRefCount(frame.ident);
+    }
+    Ns_DStringFree(&cwd);
     return result;
 }
 
@@ -914,7 +1058,7 @@ AdpDebug(NsInterp *itPtr, char *ptr, int len, int nscript)
     int code, fd;
     Tcl_Interp *interp = itPtr->interp;
     int level = itPtr->adp.debugLevel;
-    char *file = Tcl_GetString(itPtr->adp.objv[0]);
+    char *file = Tcl_GetString(itPtr->adp.framePtr->objv[0]);
     char buf[10], debugfile[255];
     Ns_DString ds;
 
@@ -978,18 +1122,18 @@ FreeInterpPage(void *arg)
     InterpPage *ipagePtr = arg;
     Page *pagePtr = ipagePtr->pagePtr;
     NsServer *servPtr = pagePtr->servPtr;
-    int i;
 
-    for (i = 0; i < pagePtr->code.nscripts; ++i) {
-	if (ipagePtr->objs[i] != NULL) {
-	    Tcl_DecrRefCount(ipagePtr->objs[i]);
-	}
-    }
+    FreeObjs(ipagePtr->objs);
     Ns_MutexLock(&servPtr->adp.pagelock);
     if (--pagePtr->refcnt == 0) {
 	if (pagePtr->hPtr != NULL) {
 	    Tcl_DeleteHashEntry(pagePtr->hPtr);
 	}
+	if (pagePtr->cachePtr != NULL) {
+    	    FreeObjs(ipagePtr->cacheObjs);
+	    DecrCache(pagePtr->cachePtr);
+	}
+	NsAdpFreeCode(&pagePtr->code);
 	ns_free(pagePtr);
     }
     Ns_MutexUnlock(&servPtr->adp.pagelock);
@@ -1000,9 +1144,36 @@ FreeInterpPage(void *arg)
 /*
  *----------------------------------------------------------------------
  *
- * ParseFree --
+ * AllocObjs --
  *
- *  	Free buffers in an AdpParse structure.
+ *  	Allocate new page script objects.
+ *
+ * Results:
+ *	Pointer to new objects.
+ *
+ * Side Effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static Objs *
+AllocObjs(int nobjs)
+{
+    Objs *objsPtr;
+
+    objsPtr = ns_calloc(1, sizeof(Objs) + (nobjs * sizeof(Tcl_Obj *)));
+    objsPtr->nobjs = nobjs;
+    return objsPtr;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * FreeObjs --
+ *
+ *  	Free page objects, decrementing ref counts as needed.
  *
  * Results:
  *	None.
@@ -1014,8 +1185,75 @@ FreeInterpPage(void *arg)
  */
 
 static void
-ParseFree(AdpParse *parsePtr)
+FreeObjs(Objs *objsPtr)
 {
-    Tcl_DStringFree(&parsePtr->hdr);
-    Tcl_DStringFree(&parsePtr->text);
+    int i;
+
+    for (i = 0; i < objsPtr->nobjs; ++i) {
+	if (objsPtr->objs[i] != NULL) {
+	    Tcl_DecrRefCount(objsPtr->objs[i]);
+	}
+    }
+    ns_free(objsPtr);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * DecrCache --
+ *
+ *  	Decrement ref count of a cache entry, potentially freeing
+ *	the cache.
+ *
+ * Results:
+ *	None.
+ *
+ * Side Effects:
+ *	Will free cache on last reference count.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+DecrCache(AdpCache *cachePtr)
+{
+    if (--cachePtr->refcnt == 0) {
+	NsAdpFreeCode(&cachePtr->code);
+	ns_free(cachePtr);
+    }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * AdpTrace --
+ *
+ *	Trace execution of an ADP page.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Dumps tracing info, possibly truncated, via Ns_Log.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+AdpTrace(NsInterp *itPtr, char *ptr, int len)
+{
+    char type;
+
+    if (len >= 0) {
+	type = 'T';
+    } else {
+	type = 'S';
+	len = -len;
+    }
+    if (len > itPtr->servPtr->adp.tracesize) {
+	len = itPtr->servPtr->adp.tracesize;
+    }
+    Ns_Log(Notice, "adp[%d%c]: %.*s", itPtr->adp.depth, type, len, ptr);
 }
