@@ -34,10 +34,11 @@
  *      Manage the Ns_Conn structure
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/conn.c,v 1.38 2003/08/05 19:56:05 elizthom Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/conn.c,v 1.48 2006/04/13 19:06:14 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
+static void SetFlag(Ns_Conn *conn, int bit, int flag);
 static int GetChan(Tcl_Interp *interp, char *id, Tcl_Channel *chanPtr);
 static int GetIndices(Tcl_Interp *interp, Conn *connPtr, Tcl_Obj **objv,
 		      int *offPtr, int *lenPtr);
@@ -163,23 +164,105 @@ Ns_ConnContentLength(Ns_Conn *conn)
  *
  * Ns_ConnContent --
  *
- *	Return pointer to start of content.
+ *	Return pointer to start of content of length conn->contetLength.
+ *	Note the content is likely, but not guaranteed to be, null
+ *	terminated. Specifically, in the case of file mapped content
+ *	which ends on a page boundry, a terminating null may not be
+ *	present.  The content is safe to modify in place.
  *
  * Results:
- *	Start of content.
+ *	Start of content or NULL on mapping failure.
  *
  * Side effects:
- *	None. 
+ *	Content file will be mapped if currently only in an open file.
  *
  *----------------------------------------------------------------------
  */
 
 char *
 Ns_ConnContent(Ns_Conn *conn)
+{
+    return NsConnContent(conn, NULL, NULL);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Ns_ConnContentOnDisk --
+ *
+ *	Return 1 if the content has been copied to a temp file, either
+ *	because it was greater than maxinput, or because Ns_ConnContentFd
+ *	has been called.  Returns 0, otherwise.  This is useful in the case
+ *	the application wants to be as efficient as possible, and not cause
+ *	excess file creation or mmap()ing.
+ *
+ * Results:
+ *	0 if content only in RAM, 1 if in /tmp file.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+Ns_ConnContentOnDisk(Ns_Conn *conn)
 {
     Conn *connPtr = (Conn *) conn;
 
-    return connPtr->reqPtr->content;
+    return connPtr->tfd > -1;
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Ns_ConnContentFd --
+ *
+ *	Return open fd with request content.  The fd is owned by the
+ *	connection and should not be closed by the caller.
+ *
+ * Results:
+ *	Open temp file or -1 on new temp file failure.
+ *
+ * Side effects:
+ *	Content will be copied to a temp file if it is currently only
+ *	in memory.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+Ns_ConnContentFd(Ns_Conn *conn)
+{
+    Conn *connPtr = (Conn *) conn;
+    int len, fd;
+    char *err;
+
+    if (connPtr->tfd < 0) {
+	err = NULL;
+	fd = Ns_GetTemp();
+	if (fd < 0) {
+	    err = "Ns_GetTemp";
+	} else if ((len = connPtr->contentLength) > 0) {
+	    if (write(fd, connPtr->content, len) != len)  {
+		err = "write";
+	    } else if (lseek(fd, (off_t) 0, SEEK_SET) != 0) {
+		err = "lseek";
+	    }
+	    if (err) {
+		Ns_ReleaseTemp(fd);
+	    }
+	}
+	if (!err) {
+	    connPtr->tfd = fd;
+	} else {
+	    Ns_Log(Error, "conn[%d]: could not get fd: %s failed: %s",
+		connPtr->id, err, strerror(errno));
+	}
+    }
+    return connPtr->tfd;
 }
 
 
@@ -226,11 +309,24 @@ Ns_ConnServer(Ns_Conn *conn)
 
 int
 Ns_ConnResponseStatus(Ns_Conn *conn)
+{
+    return Ns_ConnGetStatus(conn);
+}
+
+int
+Ns_ConnGetStatus(Ns_Conn *conn)
 {
     Conn           *connPtr = (Conn *) conn;
 
-    return connPtr->responseStatus;
+    return connPtr->status;
+}
 
+void
+Ns_ConnSetStatus(Ns_Conn *conn, int status)
+{
+    Conn           *connPtr = (Conn *) conn;
+
+    connPtr->status = status;
 }
 
 
@@ -305,7 +401,7 @@ Ns_ConnPeer(Ns_Conn *conn)
 {
     Conn           *connPtr = (Conn *) conn;
 
-    return connPtr->reqPtr->peer;
+    return connPtr->peer;
 }
 
 
@@ -330,7 +426,7 @@ Ns_ConnPeerPort(Ns_Conn *conn)
 {
     Conn           *connPtr = (Conn *) conn;
 
-    return connPtr->reqPtr->port;
+    return connPtr->port;
 }
 
 
@@ -565,7 +661,7 @@ Ns_ConnStartTime(Ns_Conn *conn)
 {
     Conn *connPtr = (Conn *) conn;
 
-    return &connPtr->startTime;
+    return &connPtr->times.queue;
 }
 
 
@@ -620,7 +716,7 @@ Ns_ConnModifiedSince(Ns_Conn *conn, time_t since)
     Conn	   *connPtr = (Conn *) conn;
     char           *hdr;
 
-    if (connPtr->servPtr->opts.modsince) {
+    if (connPtr->servPtr->opts.flags & SERV_MODSINCE) {
         hdr = Ns_SetIGet(conn->headers, "If-Modified-Since");
         if (hdr != NULL && Ns_ParseHttpTime(hdr) >= since) {
 	    return NS_FALSE;
@@ -629,6 +725,66 @@ Ns_ConnModifiedSince(Ns_Conn *conn, time_t since)
     return NS_TRUE;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Ns_ConnGetType, Ns_ConnSetType --
+ *
+ *	Get (set) the response mime type.
+ *
+ * Results:
+ *	Pointer to current type.
+ *
+ * Side effects:
+ *	May update connection enconding.
+ *
+ *----------------------------------------------------------------------
+ */
+
+char *
+Ns_ConnGetType(Ns_Conn *conn)
+{
+    Conn *connPtr = (Conn *) conn;
+
+    return connPtr->type;
+}
+
+void
+Ns_ConnSetType(Ns_Conn *conn, char *type)
+{
+    Conn *connPtr = (Conn *) conn;
+    NsServer *servPtr = connPtr->servPtr;
+    Tcl_Encoding encoding;
+    Ns_DString ds;
+    char *charset;
+    int len;
+
+    /*
+     * If the output type is text, set the output encoding based on 
+     * the type charset.  If the charset is missing, use the server
+     * default.
+     */
+
+    Ns_DStringInit(&ds);
+    if (type != NULL && (strncmp(type, "text/", 5) == 0)) {
+	encoding = NULL;
+    	charset = Ns_FindCharset(type, &len);
+    	if (charset == NULL && (charset = servPtr->defcharset) != NULL) {
+	    Ns_DStringVarAppend(&ds, type, "; charset=", charset, NULL);
+	    type = ds.string;
+	    len = ds.length;
+	}
+    	if (charset != NULL) {
+	    encoding = Ns_GetCharsetEncodingEx(charset, len);
+	}
+	Ns_ConnSetEncoding(conn, encoding);
+    }
+    ns_free(connPtr->type);
+    connPtr->type = ns_strcopy(type);
+    Ns_DStringFree(&ds);
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -652,7 +808,7 @@ Ns_ConnGetEncoding(Ns_Conn *conn)
 {
     Conn *connPtr = (Conn *) conn;
 
-    return connPtr->encoding;
+    return connPtr->outputEncoding;
 }
 
 void
@@ -660,7 +816,7 @@ Ns_ConnSetEncoding(Ns_Conn *conn, Tcl_Encoding encoding)
 {
     Conn *connPtr = (Conn *) conn;
 
-    connPtr->encoding = encoding;
+    connPtr->outputEncoding = encoding;
 }
 
 
@@ -702,7 +858,7 @@ Ns_ConnSetUrlEncoding(Ns_Conn *conn, Tcl_Encoding encoding)
 /*
  *----------------------------------------------------------------------
  *
- * NsIsIdConn --
+ * NsTclCheckConnId --
  *
  *	Given an conn ID, could this be a conn ID?
  *
@@ -710,28 +866,35 @@ Ns_ConnSetUrlEncoding(Ns_Conn *conn, Tcl_Encoding encoding)
  *	Boolean. 
  *
  * Side effects:
- *	None. 
+ *	If interp is non-null, an error message will be left if
+ *	necessary.
  *
  *----------------------------------------------------------------------
  */
 
 int
-NsIsIdConn(char *connId)
+NsTclCheckConnId(Tcl_Interp *interp, Tcl_Obj *objPtr)
 {
-    if (connId == NULL || *connId != 'c') {
+    char *id = Tcl_GetString(objPtr);
+
+    if (id[0] != 'c') {
+	if (interp != NULL) {
+	    Tcl_AppendResult(interp, "invalid connid: ", id, NULL);
+	}
 	return NS_FALSE;
     }
     return NS_TRUE;
 }
 
 
-
 /*
  *----------------------------------------------------------------------
  *
  * Ns_ConnGetWriteEncodedFlag --
+ * Ns_ConnGetKeepAliveFlag --
+ * Ns_ConnGetGzipFlag --
  *
- *	Is the given connection set for encoded writes.
+ *	Get the current write encoding, keepalive, or gzip flag.
  *
  * Results:
  *	Boolean. 
@@ -745,17 +908,30 @@ NsIsIdConn(char *connId)
 int
 Ns_ConnGetWriteEncodedFlag(Ns_Conn *conn)
 {
-    return (conn->flags & NS_CONN_WRITE_ENCODED );
+    return (conn->flags & NS_CONN_WRITE_ENCODED);
 }
 
+int
+Ns_ConnGetKeepAliveFlag(Ns_Conn *conn)
+{
+    return (conn->flags & NS_CONN_KEEPALIVE);
+}
 
+int
+Ns_ConnGetGzipFlag(Ns_Conn *conn)
+{
+    return (conn->flags & NS_CONN_GZIP);
+}
 
+
 /*
  *----------------------------------------------------------------------
  *
  * Ns_ConnSetWriteEncodedFlag --
+ * Ns_ConnSetKeepAliveFlag --
+ * Ns_ConnSetGzipFlag --
  *
- *	Set the given connection encoded writes flag per parameter.
+ *	Set the current write encoding, keepalive, or gzip flag.
  *
  * Results:
  *	void. 
@@ -769,10 +945,28 @@ Ns_ConnGetWriteEncodedFlag(Ns_Conn *conn)
 void
 Ns_ConnSetWriteEncodedFlag(Ns_Conn *conn, int flag)
 {
-    if( flag ) {
-        conn->flags |= NS_CONN_WRITE_ENCODED;
+    SetFlag(conn, NS_CONN_WRITE_ENCODED, flag);
+}
+
+void
+Ns_ConnSetKeepAliveFlag(Ns_Conn *conn, int flag)
+{
+    SetFlag(conn, NS_CONN_KEEPALIVE, flag);
+}
+
+void
+Ns_ConnSetGzipFlag(Ns_Conn *conn, int flag)
+{
+    SetFlag(conn, NS_CONN_GZIP, flag);
+}
+
+static void
+SetFlag(Ns_Conn *conn, int bit, int flag)
+{
+    if (flag) {
+        conn->flags |= bit;
     } else {
-        conn->flags &= ~NS_CONN_WRITE_ENCODED;
+        conn->flags &= ~bit;
     }
 }
 
@@ -801,33 +995,34 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
     Conn         *connPtr;
     Ns_Set       *form;
     Ns_Request   *request;
-    Tcl_Encoding  encoding;
+    Tcl_Encoding  encoding, *encodingPtr;
     Tcl_Channel   chan;
     Tcl_Obj	 *result;
-    Tcl_HashEntry *hPtr;
     Tcl_HashSearch search;
-    FormFile	 *filePtr;
-    int		  idx, off, len;
-    int           write_encoded_flag;
+    Ns_ConnFile	 *filePtr;
+    int		  idx, off, len, flag, fd;
+    char	 *content;
 
     static CONST char *opts[] = {
 	 "authpassword", "authuser", "close", "content", "contentlength",
-	 "copy", "driver", "encoding", "files", "fileoffset",
-	 "filelength", "fileheaders", "flags", "form", "headers",
-	 "host", "id", "isconnected", "location", "method",
+	 "contentchannel", "copy", "driver", "encoding", "files",
+	 "fileoffset", "filelength", "fileheaders", "flags", "form",
+	 "headers", "host", "id", "isconnected", "location", "method",
 	 "outputheaders", "peeraddr", "peerport", "port", "protocol",
 	 "query", "request", "server", "sock", "start", "status",
-	 "url", "urlc", "urlencoding", "urlv", "version", "write_encoded", NULL
+	 "url", "urlc", "urlencoding", "urlv", "version",
+	 "write_encoded", NULL
     };
-    enum ISubCmdIdx {
+    enum {
 	 CAuthPasswordIdx, CAuthUserIdx, CCloseIdx, CContentIdx,
-	 CContentLengthIdx, CCopyIdx, CDriverIdx, CEncodingIdx,
-	 CFilesIdx, CFileOffIdx, CFileLenIdx, CFileHdrIdx, CFlagsIdx,
-	 CFormIdx, CHeadersIdx, CHostIdx, CIdIdx, CIsConnectedIdx,
-	 CLocationIdx, CMethodIdx, COutputHeadersIdx, CPeerAddrIdx,
-	 CPeerPortIdx, CPortIdx, CProtocolIdx, CQueryIdx, CRequestIdx,
-	 CServerIdx, CSockIdx, CStartIdx, CStatusIdx, CUrlIdx,
-	 CUrlcIdx, CUrlEncodingIdx, CUrlvIdx, CVersionIdx, CWriteEncodedIdx
+	 CContentLengthIdx, CContentChannelIdx, CCopyIdx, CDriverIdx,
+	 CEncodingIdx, CFilesIdx, CFileOffIdx, CFileLenIdx,
+	 CFileHdrIdx, CFlagsIdx, CFormIdx, CHeadersIdx, CHostIdx,
+	 CIdIdx, CIsConnectedIdx, CLocationIdx, CMethodIdx,
+	 COutputHeadersIdx, CPeerAddrIdx, CPeerPortIdx, CPortIdx,
+	 CProtocolIdx, CQueryIdx, CRequestIdx, CServerIdx, CSockIdx,
+	 CStartIdx, CStatusIdx, CUrlIdx, CUrlcIdx, CUrlEncodingIdx,
+	 CUrlvIdx, CVersionIdx, CWriteEncodedIdx
     } opt;
 
     if (objc < 2) {
@@ -838,24 +1033,21 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 			    (int *) &opt) != TCL_OK) {
 	return TCL_ERROR;
     }
-
     result  = Tcl_GetObjResult(interp);
-    connPtr = (Conn *) conn = itPtr->conn;
 
     /*
      * Only the "isconnected" option operates without a conn.
      */
 
     if (opt == CIsConnectedIdx) {
-	Tcl_SetBooleanObj(result, connPtr ? 1 : 0);
+	Tcl_SetBooleanObj(result, itPtr->conn ? 1 : 0);
 	return TCL_OK;
     }
-    if (connPtr == NULL) {
-	Tcl_SetResult(interp, "no current connection", TCL_STATIC);
+    if (NsTclGetConn(itPtr, &conn) != TCL_OK) {
         return TCL_ERROR;
     }
-
-    request = connPtr->request;
+    request = conn->request;
+    connPtr = (Conn *) conn;
     switch (opt) {
 
 	case CIsConnectedIdx:
@@ -863,7 +1055,7 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    break;
 		
 	case CUrlvIdx:
-	    if (objc == 2 || (objc == 3 && NsIsIdConn(Tcl_GetString(objv[2])))) {
+	    if (objc == 2 || (objc == 3 && NsTclCheckConnId(NULL, objv[2]))) {
 		for (idx = 0; idx < request->urlc; idx++) {
 		    Tcl_AppendElement(interp, request->urlv[idx]);
 		}
@@ -882,6 +1074,21 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    Tcl_SetResult(interp, connPtr->authPasswd, TCL_STATIC);
 	    break;
 
+	case CContentChannelIdx:
+	    fd = Ns_ConnContentFd(conn);
+	    if (fd >= 0 && (fd = dup(fd)) >= 0) {
+		/* NB: Dup the fd so the channel can be safely closed later. */
+		chan = Tcl_MakeFileChannel((ClientData) fd, TCL_READABLE);
+	 	if (chan == NULL) {
+		    close(fd);
+		} else {
+		    Tcl_RegisterChannel(interp, chan);
+		    Tcl_SetResult(interp, Tcl_GetChannelName(chan),
+				  TCL_VOLATILE);
+		}
+	    }
+	    break;
+
 	case CContentIdx:
 	    if (objc != 2 && objc != 4) {
 		Tcl_WrongNumArgs(interp, 2, objv, "?off len?");
@@ -903,43 +1110,24 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    break;
 
 	case CEncodingIdx:
-	    if (objc > 2) {
-		encoding = Ns_GetEncoding(Tcl_GetString(objv[2]));
-		if (encoding == NULL) {
-		    Tcl_AppendResult(interp, "no such encoding: ",
-			Tcl_GetString(objv[2]), NULL);
-		    return TCL_ERROR;
-		}
-		connPtr->encoding = encoding;
-	    }
-	    if (connPtr->encoding != NULL) {
-		Tcl_SetResult(interp, (char *) Tcl_GetEncodingName(connPtr->encoding),
-			      TCL_STATIC);
-	    }
-	    break;
-	
 	case CUrlEncodingIdx:
+	    if (opt == CEncodingIdx) {
+		encodingPtr = &connPtr->outputEncoding;
+	    } else {
+		encodingPtr = &connPtr->urlEncoding;
+	    }
 	    if (objc > 2) {
-		encoding = Ns_GetEncoding(Tcl_GetString(objv[2]));
-		if (encoding == NULL) {
+    		encoding = Ns_GetEncoding(Tcl_GetString(objv[2]));
+    		if (encoding == NULL) {
 		    Tcl_AppendResult(interp, "no such encoding: ",
-			Tcl_GetString(objv[2]), NULL);
+				     Tcl_GetString(objv[2]), NULL);
 		    return TCL_ERROR;
 		}
-                /*
-                 * Check to see if form data has already been parsed.
-                 * If so, and the urlEncoding is changing, then clear
-                 * the previous form data.
-                 */
-                if ((connPtr->urlEncoding != encoding) &&
-                    (itPtr->nsconn.flags & CONN_TCLFORM)) {
-                    Ns_ConnClearQuery(conn);
-                    itPtr->nsconn.flags ^= CONN_TCLFORM;
-                }
-		connPtr->urlEncoding = encoding;
+		*encodingPtr = encoding;
 	    }
-	    if (connPtr->urlEncoding != NULL) {
-		Tcl_SetResult(interp, (char *) Tcl_GetEncodingName(connPtr->urlEncoding),
+	    encoding = *encodingPtr;
+	    if (encoding != NULL) {
+    	    	Tcl_SetResult(interp, (char *) Tcl_GetEncodingName(encoding),
 			      TCL_STATIC);
 	    }
 	    break;
@@ -973,6 +1161,10 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    break;
 	
 	case CFormIdx:
+	     /* NB: Ignore any cached form if query is no longer valid. */
+	    if (!NsCheckQuery(conn)) {
+		itPtr->nsconn.flags &= ~CONN_TCLFORM;
+	    }
 	    if (itPtr->nsconn.flags & CONN_TCLFORM) {
 		Tcl_SetResult(interp, itPtr->nsconn.form, TCL_STATIC);
 	    } else {
@@ -992,10 +1184,10 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 		Tcl_WrongNumArgs(interp, 2, objv, NULL);
 		return TCL_ERROR;
 	    }
-	    hPtr = Tcl_FirstHashEntry(&connPtr->files, &search);
-	    while (hPtr != NULL) {
-		Tcl_AppendElement(interp, Tcl_GetHashKey(&connPtr->files, hPtr));
-		hPtr = Tcl_NextHashEntry(&search);
+	    filePtr = Ns_ConnFirstFile(conn, &search);
+	    while (filePtr != NULL) {
+		Tcl_AppendElement(interp, filePtr->name);
+		filePtr = Ns_ConnNextFile(&search);
 	    }
 	    break;
 
@@ -1003,21 +1195,21 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	case CFileLenIdx:
 	case CFileHdrIdx:
 	    if (objc != 3) {
-		Tcl_WrongNumArgs(interp, 2, objv, NULL);
+		Tcl_WrongNumArgs(interp, 2, objv, "file");
 		return TCL_ERROR;
 	    }
-	    hPtr = Tcl_FindHashEntry(&connPtr->files, Tcl_GetString(objv[2]));
-	    if (hPtr == NULL) {
-		Tcl_AppendResult(interp, "no such file: ", Tcl_GetString(objv[2]), NULL);
+	    filePtr = Ns_ConnGetFile(conn, Tcl_GetString(objv[2]));
+	    if (filePtr == NULL) {
+		Tcl_AppendResult(interp, "no such file: ",
+				 Tcl_GetString(objv[2]), NULL);
 		return TCL_ERROR;
 	    }
-	    filePtr = Tcl_GetHashValue(hPtr);
 	    if (opt == CFileOffIdx) {
-	    	Tcl_SetLongObj(result, (long) filePtr->off);
+	    	Tcl_SetLongObj(result, (long) filePtr->offset);
 	    } else if (opt == CFileLenIdx) {
-	    	Tcl_SetLongObj(result, (long) filePtr->len);
+	    	Tcl_SetLongObj(result, (long) filePtr->length);
 	    } else {
-		Ns_TclEnterSet(interp, filePtr->hdrs, NS_TCL_SET_STATIC);
+		Ns_TclEnterSet(interp, filePtr->headers, NS_TCL_SET_STATIC);
 	    }
 	    break;
 
@@ -1030,27 +1222,30 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 		GetChan(interp, Tcl_GetString(objv[4]), &chan) != TCL_OK) {
 		return TCL_ERROR;
 	    }
-	    if (Tcl_Write(chan, connPtr->reqPtr->content + off, len) != len) {
-		Tcl_AppendResult(interp, "could not write ", Tcl_GetString(objv[3]), " bytes to ",
-		    Tcl_GetString(objv[4]), ": ", Tcl_PosixError(interp), NULL);
-		return TCL_ERROR;
+	    if (len > 0) {
+	    	content = Ns_ConnContent(conn);
+	    	if (content == NULL) {
+		    Tcl_SetResult(interp, "could not get content", TCL_STATIC);
+		    return TCL_ERROR;
+		}
+	    	if (Tcl_Write(chan, content + off, len) != len) {
+		    Tcl_AppendResult(interp, "could not write ",
+			Tcl_GetString(objv[3]), " bytes to ",
+		    	Tcl_GetString(objv[4]), ": ",
+			Tcl_PosixError(interp), NULL);
+		    return TCL_ERROR;
+		}
 	    }
 	    break;
 
         case CWriteEncodedIdx:
 	    if (objc > 2) {
-		if (Tcl_GetIntFromObj(interp, objv[2], &write_encoded_flag) 
-                    != TCL_OK) {
-                    Tcl_AppendResult(interp, "Invalid write-encoded flag", NULL );
+		if (Tcl_GetBooleanFromObj(interp, objv[2], &flag) != TCL_OK) {
                     return TCL_ERROR;
                 }
-                Ns_ConnSetWriteEncodedFlag(conn, write_encoded_flag);
-           }
-           if (Ns_ConnGetWriteEncodedFlag(conn)) {
-               Tcl_AppendResult(interp, "1", NULL );
-           } else {
-               Tcl_AppendResult(interp, "0", NULL );
+                Ns_ConnSetWriteEncodedFlag(conn, flag);
            }
+	   Tcl_SetBooleanObj(result, Ns_ConnGetWriteEncodedFlag(conn));
            break;
 
 	case CRequestIdx:
@@ -1102,7 +1297,14 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    break;
 
 	case CStatusIdx:
-	    Tcl_SetIntObj(result, Ns_ConnResponseStatus(conn));
+	    if (objc > 2) {
+		int status;
+		if (Tcl_GetIntFromObj(interp, objv[2], &status) != TCL_OK) {
+		    return TCL_ERROR;
+		}
+		Ns_ConnSetStatus(conn, status);
+	    }
+	    Tcl_SetIntObj(result, Ns_ConnGetStatus(conn));
 	    break;
 
 	case CSockIdx:
@@ -1118,7 +1320,7 @@ NsTclConnObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    break;
 
 	case CStartIdx:
-	    Ns_TclSetTimeObj(result, &connPtr->startTime);
+	    Ns_TclSetTimeObj(result, &connPtr->times.queue);
 	    break;
 
 	case CCloseIdx:
@@ -1154,32 +1356,28 @@ int
 NsTclWriteContentObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 {
     NsInterp	*itPtr = arg;
+    Ns_Conn	*conn;
     Tcl_Channel  chan;
 
     if (objc != 2 && objc != 3) {
         Tcl_WrongNumArgs(interp, 1, objv, "?connid? channel");
         return TCL_ERROR;
     }
-    if (objc == 3 && !NsIsIdConn(Tcl_GetString(objv[1]))) {
-	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "bad connid: \"", 
-		Tcl_GetString(objv[1]), "\"", NULL);
+    if (objc == 3 && !NsTclCheckConnId(interp, objv[1])) {
 	return TCL_ERROR;
     }
-    if (itPtr->conn == NULL) {
-	Tcl_SetResult(interp, "no connection", TCL_STATIC);
+    if (NsTclGetConn(itPtr, &conn) != TCL_OK) {
         return TCL_ERROR;
     }
     if (GetChan(interp, Tcl_GetString(objv[objc-1]), &chan) != TCL_OK) {
 	return TCL_ERROR;
     }
     Tcl_Flush(chan);
-    if (Ns_ConnCopyToChannel(itPtr->conn, (size_t)itPtr->conn->contentLength, 
-                             chan) != NS_OK) {
+    if (Ns_ConnCopyToChannel(conn, (size_t) conn->contentLength, chan) != NS_OK) {
         Tcl_SetResult(interp, "could not copy content (likely client disconnect)",
 		TCL_STATIC);
         return TCL_ERROR;
     }
-    
     return TCL_OK;
 }
 
@@ -1205,75 +1403,49 @@ int
 NsTclStartContentObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
                         Tcl_Obj **objv)
 {
-    NsInterp     *itPtr = arg;
     Tcl_Encoding  encoding;
-    char         *opt;
-    int           status;
-    int           i;
+    Ns_Conn	 *conn;
+    char	 *opt;
+    static CONST char *flags[] = {
+	"-charset", "-type", NULL
+    };
+    enum {
+	FCharsetIdx, FTypeIdx
+    } flag;
 
-    status = TCL_OK;
+    if (objc != 1 && objc != 3) {
+	Tcl_WrongNumArgs(interp, 1, objv, "?-charset charset|-type type?");
+	return TCL_ERROR;
+    }
     encoding = NULL;
-
-    for (i = 1; i < objc && status == TCL_OK; i++) {
-	opt = Tcl_GetString(objv[i]);
-	if (STREQ(opt, "-charset")) {
-
-	    if (encoding != NULL) {
-		Tcl_AppendResult(interp, Tcl_GetString(objv[0]),
-		    ": charset may only be specified by one flag", NULL);
-		status = TCL_ERROR;
-	    }
-
-	    if (++i >= objc) {
-		Tcl_AppendResult(interp, Tcl_GetString(objv[0]),
-		    ": missing argument for -charset flag", NULL);
-		status = TCL_ERROR;
-	    }
-
-	    encoding = Ns_GetCharsetEncoding(Tcl_GetString(objv[i]));
-	    if (encoding == NULL) {
-		Tcl_AppendResult(interp, Tcl_GetString(objv[0]),
-		    ": could not find an encoding for charset ",
-		    Tcl_GetString(objv[i]), NULL);
-		status = TCL_ERROR;
-	    }
+    if (objc == 3) {
+	if (Tcl_GetIndexFromObj(interp, objv[1], flags, "flag", 0,
+		(int *) &flag) != TCL_OK) {
+	    return TCL_ERROR;
 	}
-
-	else if (STREQ(opt, "-type")) {
-
-	    if (encoding != NULL) {
-		Tcl_AppendResult(interp, Tcl_GetString(objv[0]),
-		    ": charset may only be specified by one flag", NULL);
-		status = TCL_ERROR;
-	    }
-
-	    if (++i >= objc) {
-		Tcl_AppendResult(interp, Tcl_GetString(objv[0]),
-		    ": missing argument for -type flag", NULL);
-		status = TCL_ERROR;
-	    }
-
-	    encoding = Ns_GetTypeEncoding(Tcl_GetString(objv[i]));
+	opt = Tcl_GetString(objv[2]);
+	switch (flag) {
+	case FCharsetIdx:
+	    encoding = Ns_GetCharsetEncoding(opt);
+	    break;
+	case FTypeIdx:
+	    encoding  = Ns_GetTypeEncoding(opt);
+	    break;
 	}
-
-	else {
-	    Tcl_AppendResult(interp, "usage: ", Tcl_GetString(objv[0]),
-		" ?-charset charsetname? ?-type content-type?", NULL);
-	    status = TCL_ERROR;
+	if (encoding == NULL) {
+	    Tcl_AppendResult(interp, "no encoding for ",
+		(flags[flag])+1, " \"", opt, "\"", NULL);
+	    return TCL_ERROR;
 	}
     }
-
-    if (status != TCL_OK) {
-	return status;
-    }
-
-    Ns_ConnSetWriteEncodedFlag( itPtr->conn, NS_TRUE );
-    Ns_ConnSetEncoding( itPtr->conn, encoding );
-
-    return status;
+    if (NsTclGetConn((NsInterp *) arg, &conn) != TCL_OK) {
+	return TCL_ERROR;
+    } 
+    Ns_ConnSetWriteEncodedFlag(conn, NS_TRUE);
+    Ns_ConnSetEncoding(conn, encoding);
+    return TCL_OK;
 }
 
-
 
 /*
  *----------------------------------------------------------------------
@@ -1338,11 +1510,11 @@ GetIndices(Tcl_Interp *interp, Conn *connPtr, Tcl_Obj **objv, int *offPtr, int *
 	Tcl_GetIntFromObj(interp, objv[1], &len) != TCL_OK) {
 	return TCL_ERROR;
     }
-    if (off < 0 || off > connPtr->reqPtr->length) {
+    if (off < 0 || off > connPtr->contentLength) {
 	Tcl_AppendResult(interp, "invalid offset: ", Tcl_GetString(objv[0]), NULL);
 	return TCL_ERROR;
     }
-    if (len < 0 || len > (connPtr->reqPtr->length - off)) {
+    if (len < 0 || len > (connPtr->contentLength - off)) {
 	Tcl_AppendResult(interp, "invalid length: ", Tcl_GetString(objv[1]), NULL);
 	return TCL_ERROR;
     }
