@@ -34,9 +34,12 @@
 #define _GNU_SOURCE
 #endif
 
+#define NSD_EXPORTS
 #include "ns.h"
+
 #ifndef _WIN32
 #include <pthread.h>
+#include <sys/mman.h>
 #endif
 #include <assert.h>
 
@@ -67,7 +70,8 @@
 #else
   #define POLLIN 1
   #define POLLOUT 2
-  #define POLLPRI 3
+  #define POLLPRI 4
+  #define POLLHUP 8
   struct pollfd {
     int fd;
     short events;
@@ -103,14 +107,17 @@
 #define NS_SIGHUP   SIGHUP
 #endif
 
+#define _MAX(x,y) ((x) > (y) ? (x) : (y))
+#define _MIN(x,y) ((x) > (y) ? (y) : (x))
+
 /*
  * constants
  */
 
 #define NSD_NAME             "AOLserver"
 #define NSD_VERSION	     NS_PATCH_LEVEL
-#define NSD_LABEL            "aolserver4_0"
-#define NSD_TAG              "$Name: aolserver_v40_r10 $"
+#define NSD_LABEL            "aolserver4_5"
+#define NSD_TAG              "$Name: aolserver_v45_r0 $"
 #define NS_CONFIG_PARAMETERS "ns/parameters"
 #define NS_CONFIG_THREADS    "ns/threads"
 
@@ -119,13 +126,6 @@
 #define ADP_ABORT    2
 #define ADP_RETURN   4
 
-#define LOG_ROLL	1
-#define LOG_EXPAND	2
-#define LOG_DEBUG	4
-#define LOG_DEV		8
-#define LOG_NONOTICE	16
-#define LOG_USEC	32
-
 /*
  * The following is the default text/html content type
  * sent to the browsers.  The charset is also used for
@@ -154,21 +154,7 @@ struct _nsconf {
     char	    address[16];
     int             shutdowntimeout;
     int             backlog;
-
-    /*
-     * The following table holds the configured virtual servers.  The
-     * dstring maintains a Tcl list of the names.
-     */
-
-    Tcl_HashTable   servertable;
-    Tcl_DString     servers;
-
-    /*
-     * The following table holds config section sets from
-     * the config file.
-     */
-
-    Tcl_HashTable   sections;
+    int             debug;
 
     /*
      * The following struct maintains server state.
@@ -180,16 +166,17 @@ struct _nsconf {
     	int		started;
     	int		stopping;
     } state;
-    
-    struct {
-	char *file;
-	int  flags;
-	int  maxlevel;
-	int  maxback;
-	int  maxbuffer;
-	int  flushint;
-    } log;
 
+    /*
+     * The following struct is the maximum HTTP major/minor version
+     * supported.
+     */
+
+    struct {
+	unsigned int major;
+	unsigned int minor;
+    } http;
+    
     struct {
 	int maxelapsed;
     } sched;
@@ -200,28 +187,12 @@ struct _nsconf {
     } exec;
 #endif
 
-    struct {
-	bool enabled;
-	int timeout;
-	int maxkeep;
-	int npending;
-	int allmethods;
-    } keepalive;
-
     struct {
 	char *sharedlibrary;
 	char *version;
         bool lockoninit;
     } tcl;
 
-    struct {
-        char         *outputCharset;
-        Tcl_Encoding  outputEncoding;
-        bool          hackContentTypeP;
-        char         *urlCharset;
-        Tcl_Encoding  urlEncoding;
-    } encoding;
-
 };
 
 extern struct _nsconf nsconf;
@@ -238,70 +209,65 @@ typedef struct FileKey {
 
 #define FILE_KEYS (sizeof(FileKey)/sizeof(int))
 
+/*
+ * The following structure maintains an ADP call frame.
+ */
+
+typedef struct AdpFrame {
+    struct AdpFrame   *prevPtr;
+    int		       line;
+    int                objc;
+    Tcl_Obj	      *ident;
+    Tcl_Obj          **objv;
+    char	      *savecwd;
+    char	      *file;
+    Ns_DString         cwdbuf;
+    Tcl_DString	      *outputPtr;
+} AdpFrame;
+
 /*
  * The following structure defines blocks of ADP.  The
  * len pointer is an array of ints with positive values
  * indicating text to copy and negative values indicating
  * scripts to evaluate.  The text and script chars are
  * packed together without null char separators starting
- * at base.  The len and base data are either stored
- * in an AdpParse structure or copied at the end of
- * a cached Page structure.
+ * at base.  The len data is stored at the end of the
+ * text dstring when parsing is complete.
  */
 
 typedef struct AdpCode {
     int		nblocks;
     int		nscripts;
-    char       *base;
     int	       *len;
+    int	       *line;
+    Tcl_DString text;
 } AdpCode;
 
-/*
- * The following structure is used to accumulate the 
- * results of parsing an ADP string.
- */
-
-typedef struct AdpParse {
-    AdpCode	code;
-    Tcl_DString hdr;
-    Tcl_DString text;
-} AdpParse;
+#define AdpCodeLen(cp,i)	((cp)->len[(i)])
+#define AdpCodeLine(cp,i)	((cp)->line[(i)])
+#define AdpCodeText(cp)		((cp)->text.string)
+#define AdpCodeBlocks(cp)	((cp)->nblocks)
+#define AdpCodeScripts(cp)	((cp)->nscripts)
 
 /*
- * The following structure defines the entire request
- * including HTTP request line, headers, and content.
+ * Various ADP option bits.
  */
 
-typedef struct Request {
-    struct Request *nextPtr;	/* Next on free list. */
-    Ns_Request	   *request;	/* Parsed request line. */
-    Ns_Set	   *headers;	/* Input headers. */
-    char	    peer[16];	/* Client peer address. */
-    int		    port;	/* Client peer port. */
-
-    /*
-     * The following pointers are used to access the
-     * buffer contents after the read-ahead is complete.
-     */
-
-    char	   *next;	/* Next read offset. */
-    char	   *content;	/* Start of content. */
-    int		    length;	/* Length of content. */
-    int		    avail;	/* Bytes avail in buffer. */
-    int		    leadblanks;	/* # of leading blank lines read */
-
-    /*
-     * The following offsets are used to manage the 
-     * buffer read-ahead process.
-     */
-
-    int		    woff;	/* Next write buffer offset. */
-    int		    roff;	/* Next read buffer offset. */
-    int		    coff;	/* Content buffer offset. */
-
-    Tcl_DString	    buffer;	/* Request and content buffer. */
-
-} Request;
+#define ADP_SAFE	0x01	/* Use Tcl_SafeEval for ADP. */
+#define ADP_SINGLE	0x02	/* Combine blocks into a single script. */
+#define ADP_DEBUG	0x04	/* Enable debugging. */
+#define ADP_EXPIRE	0x08	/* Send Expires: now header on output. */
+#define ADP_NOCACHE	0x10	/* Disable caching. */
+#define ADP_TRACE	0x20	/* Trace execution. */
+#define ADP_GZIP	0x80	/* Enable gzip compression. */
+#define ADP_DETAIL	0x100	/* Log connection details on error. */
+#define ADP_STRICT	0x200	/* Strict error handling. */
+#define ADP_DISPLAY	0x400	/* Display error messages in output stream. */
+#define ADP_TRIM	0x800	/* Display error messages in output stream. */
+#define ADP_FLUSHED	0x1000	/* Some output has been sent. */
+#define ADP_ERRLOGGED	0x2000	/* Error message has already been logged. */
+#define ADP_AUTOABORT	0x4000	/* Raise abort on flush error. */
+#define ADP_EVAL_FILE	0x8000	/* Object to evaluate is a file. */
 
 /*
  * The following structure maitains data for each instance of
@@ -319,7 +285,7 @@ typedef struct Driver {
     void	*arg;		    /* Driver callback data. */
     char	*server;	    /* Virtual server name. */
     char	*module;	    /* Driver module. */
-    char        *name;		    /* Driver name. */
+    char        *name;		    /* Driver name, e.g., "nssock". */
     char        *location;	    /* Location, e.g, "http://foo:9090" */
     char        *address;	    /* Address in location. */
     int     	 sendwait;	    /* send() I/O timeout. */
@@ -334,19 +300,60 @@ typedef struct Driver {
 
     struct Driver *nextPtr;	    /* Next in list of drivers. */
     struct NsServer *servPtr;	    /* Driver virtual server. */
+    char	*fullname;	    /* Full name, i.e., server/module. */
+    int          flags;             /* Driver state flags. */
+    Ns_Thread	 thread;	    /* Thread id to join on shutdown. */
+    Ns_Mutex	 lock;		    /* Lock to protect lists below. */
+    Ns_Cond 	 cond;		    /* Cond to signal reader threads,
+				     * driver query, startup, and shutdown. */
+    int     	 trigger[2];	    /* Wakeup trigger pipe. */
+
     Ns_DriverProc *proc;	    /* Driver callback. */
     int		 opts;		    /* Driver options. */
+
     int     	 closewait;	    /* Graceful close timeout. */
     int     	 keepwait;	    /* Keepalive timeout. */
-    SOCKET		 sock;		    /* Listening socket. */
-    int		 pidx;		    /* poll() index. */
     char        *bindaddr;	    /* Numerical listen address. */
     int          port;		    /* Port in location. */
     int		 backlog;	    /* listen() backlog. */
+    
+    int          maxline;           /* Maximum request line length to read. */
+    int          maxheader;         /* Maximum total header length to read. */
+    int		 maxinput;	    /* Maximum request bytes to read. */
 
-    int          maxinput;          /* Maximum request bytes to read. */
-    unsigned int loggingFlags;      /* Logging control flags */
+    struct Sock *freeSockPtr;       /* Sock free list. */
+    int     	 maxsock;	    /* Maximum open Sock's. */
+    int     	 nactive;	    /* Number of active Sock's. */
+    unsigned int nextid;	    /* Next sock unique id. */
 
+    Ns_Thread   *readers;	    /* Array of reader Ns_Thread's. */
+    int		 maxreaders;	    /* Max reader threads. */
+    int		 nreaders;	    /* Current num reader threads. */
+    int		 idlereaders;	    /* Idle reader threads. */
+
+    struct Sock *readSockPtr;       /* Sock's waiting for reader threads. */
+    struct Sock *runSockPtr;        /* Sock's returning from reader threads. */
+    struct Sock *closeSockPtr;      /* Sock's returning from conn threads. */
+
+    struct Conn *firstConnPtr;      /* First Conn waiting to run. */
+    struct Conn *lastConnPtr;       /* Last Conn waiting to run. */
+    struct Conn *freeConnPtr;       /* Conn's returning from conn threads. */
+
+    struct QueWait *freeQueWaitPtr;
+
+    Tcl_DString *queryPtr;	    /* Buffer to copy driver query data. */
+
+    struct {
+        unsigned int spins;
+        unsigned int accepts;
+        unsigned int reads;
+        unsigned int writes;
+        unsigned int queued;
+        unsigned int timeout;
+        unsigned int overflow;
+        unsigned int dropped;
+    } stats;
+    
 } Driver;
 
 /*
@@ -357,7 +364,6 @@ typedef struct Driver {
  */
 
 typedef struct Sock {
-
     /*
      * Visible in Ns_Sock.
      */
@@ -371,26 +377,34 @@ typedef struct Sock {
      */
 
     struct Sock *nextPtr;
-    struct NsServer *servPtr;
-    char *location;
+    struct Conn *connPtr;
     struct sockaddr_in sa;
-    int		 keep;
+    unsigned int id;
+    int		 state;
     int		 pidx;		    /* poll() index. */
+    Ns_Time      acceptTime;
     Ns_Time	 timeout;
-    Request	*reqPtr;
-
+    unsigned int nreads;
+    unsigned int nwrites;
 } Sock;
 
 /*
- * The following structure maintains data from an
- * updated form file.
+ * The following structure defines a per-request limits.
  */
 
-typedef struct FormFile {
-    Ns_Set *hdrs;
-    off_t   off;
-    off_t   len;
-} FormFile;
+typedef struct Limits {
+    Ns_Mutex        lock;
+    char           *name;
+    unsigned int    maxrun;
+    unsigned int    maxwait;
+    unsigned int    nrunning;
+    unsigned int    nwaiting;
+    unsigned int    ndropped;
+    unsigned int    noverflow;
+    unsigned int    ntimeout;
+    size_t	    maxupload;
+    int             timeout;
+} Limits;
 
 /*
  * The following structure maintains state for a connection
@@ -398,7 +412,6 @@ typedef struct FormFile {
  */
 
 typedef struct Conn {
-
     /*
      * Visible in an Ns_Conn:
      */
@@ -415,9 +428,24 @@ typedef struct Conn {
      * Visible only in a Conn:
      */
     
-    struct Conn *prevPtr;
     struct Conn *nextPtr;
+    struct Conn *prevPtr;
     struct Sock *sockPtr;
+    struct Limits *limitsPtr;
+
+    /*
+     * Client http major/minor version number.
+     */
+
+    unsigned int major;
+    unsigned int minor;
+
+    /*
+     * Start and end of request line for later parsing.
+     */
+
+    char *rstart;
+    char *rend;
 
     /*
      * The following are copied from sockPtr so they're valid
@@ -426,53 +454,98 @@ typedef struct Conn {
 
     char *server;
     char *location;
-    struct Request *reqPtr;
     struct NsServer *servPtr;
     struct Driver *drvPtr;
 
-    int          id;
+    unsigned int id;
     char	 idstr[16];
-    Ns_Time	 startTime;
-    Tcl_Interp  *interp;
-    Tcl_Encoding encoding;
+    struct {
+        Ns_Time  accept;
+        Ns_Time  read;
+        Ns_Time  ready;
+        Ns_Time  queue;
+        Ns_Time  run;
+        Ns_Time  close;
+        Ns_Time  done;
+    } times;
+    struct NsInterp *itPtr;
+    char	*type;
+    Tcl_Encoding outputEncoding;
     Tcl_Encoding urlEncoding;
+    Tcl_Encoding queryEncoding;
     int          nContentSent;
-    int          responseStatus;
+    int          status;
     int          responseLength;
     int          recursionCount;
     Ns_Set      *query;
     Tcl_HashTable files;
-    Tcl_DString	 queued;
-    void	*cls[NS_CONN_MAXCLS];
+
+    /*
+     * The following are copied from Sock.
+     */
+     
+    char	    peer[16];	/* Client peer address. */
+    int		    port;	/* Client peer port. */
+
+    /*
+     * The following array maintains conn local storage.
+     */
+     
+    void	   *cls[NS_CONN_MAXCLS];
+    struct QueWait *queWaitPtr;
+
+    /*
+     * The following pointers are used to access the
+     * buffer contents after the read-ahead is complete.
+     */
+
+    char	   *next;	/* Next read offset. */
+    size_t          avail;	/* Bytes avail in buffer. */
+    char	   *content;	/* Start of content. */
+    int             tfd;        /* Temp fd for file-based content. */
+    void	   *map;	/* Mmap'ed content, if any. */
+    void	   *maparg;	/* Argument for NsUnMap. */
+
+    /*
+     * The following offsets are used to manage the 
+     * buffer read-ahead process.
+     *
+     * NB: The ibuf and obuf dstrings must be the last elements of
+     * the conn as all elements before are zero'ed during conn cleanup.
+     *
+     */
+
+    int		    roff;	/* Next read buffer offset. */
+    Tcl_DString	    ibuf;	/* Request and content input buffer. */
+    Tcl_DString	    obuf;	/* Output buffer for queued headers. */
+
 } Conn;
 
 /*
  * The following structure maintains a connection thread pool.
  */
 
-typedef struct ConnPool {
-    char 	    *pool;
-    struct ConnPool *nextPtr;
-    struct NsServer *servPtr;
+typedef struct Pool {
+    Ns_Mutex        lock;
+    Ns_Cond         cond;
+    char           *name;
+    int             shutdown;
 
     /*
      * The following struct maintains the active and waiting connection
-     * queues, the free conn list, the next conn id, and the number
-     * of waiting connects.
+     * queues, the next conn id, and the number of waiting connects.
      */
 
     struct {
-    	Conn	    	   *freePtr;
 	struct {
 	    int     	    num;
-    	    Conn    	   *firstPtr;
-    	    Conn    	   *lastPtr;
-	} wait;
-	struct {
-    	    Conn    	   *firstPtr;
-	    Conn    	   *lastPtr;
+            struct Conn    *firstPtr;
+            struct Conn    *lastPtr;
+        } wait;
+        struct {
+            struct Conn    *firstPtr;
+            struct Conn    *lastPtr;
 	} active;
-	Ns_Cond     	    cond;
     } queue;
 
     /*
@@ -481,7 +554,8 @@ typedef struct ConnPool {
      * current number of threads remains within that range with individual
      * threads waiting no more than the timeout for a connection to
      * arrive.  The number of idle threads is maintained for the benefit of
-     * the ns_server command. 
+     * the ns_server command.  Threads will handle up to maxconns before
+     * exit (default is the "connsperthread" virtual server config).
      */
 
     struct {
@@ -491,9 +565,26 @@ typedef struct ConnPool {
     	int 	    	    current;
 	int 	    	    idle;
 	int 	    	    timeout;
+	int		    maxconns;
+    	unsigned int	    queued;
     } threads;
 
-} ConnPool;
+} Pool;
+
+#define SERV_AOLPRESS		0x0001	/* AOLpress support. */
+#define SERV_CHUNKED		0x0002	/* Output can be chunked. */
+#define SERV_MODSINCE		0x0004	/* Check if-modified-since. */
+#define SERV_NOTICEDETAIL	0x0008	/* Add detail to notice messages. */
+#define SERV_GZIP		0x0010	/* Enable GZIP compression. */
+
+/*
+ * The following struct maintains nsv's, shared string variables.
+ */
+
+typedef struct Nsv {
+    struct Bucket  	   *buckets;
+    int 	    	    nbuckets;
+} Nsv;
 
 /*
  * The following structure is allocated for each virtual server.
@@ -504,52 +595,47 @@ typedef struct NsServer {
     Ns_LocationProc 	   *locationProc;
 
     /*
-     * The following struct maintains the connection pool(s).
+     * Default charset for text/ types.
      */
 
-    struct {
-	Ns_Mutex    	    lock;
-	int		    nextconnid;
-	bool		    shutdown;
-    	ConnPool	   *firstPtr;
-    	ConnPool	   *defaultPtr;
-	Ns_Thread   	    joinThread;
-    } pools;
-    
+    char		  *defcharset;
+
+    /*
+     * The following encoding is used for decoding request URL's.
+     * This is a server-wide config as the request must be read
+     * and parsed before the URL can be determined to identify
+     * a possible alternate encoding.
+     */
+
+    Tcl_Encoding           urlEncoding;
+
+    /*
+     * The following encoding is used for decoding input
+     * query strings and forms unless a more specific encoding
+     * is set via ns_register_encoding.
+     */
+
+    Tcl_Encoding           inputEncoding;
+
     /* 
      * The following struct maintains various server options.
      */
 
     struct {
-	bool	    	    aolpress;
-	bool	    	    flushcontent;
-	bool	    	    modsince;
-	bool 	    	    noticedetail;
+   	int		    flags;
+	size_t 	    	    gzipmin;
+	int		    gziplevel;
     	char 	    	   *realm;
 	Ns_HeaderCaseDisposition hdrcase;
     } opts;
     
-    /*
-     * Encoding defaults for the server
-     */
-    struct {
-        char         *outputCharset;
-        Tcl_Encoding  outputEncoding;
-        bool          hackContentTypeP;
-        char         *urlCharset;
-        Tcl_Encoding  urlEncoding;
-    } encoding;
-
     /* 
      * The following struct maintains conn-related limits.
      */
 
     struct {
-    	int 	    	    maxheaders;
-	int 	    	    maxline;
-	int 	    	    maxpost;
-	int	    	    sendfdmin;
     	int 	    	    errorminsize;
+	int		    connsperthread;
     } limits;
 
     struct {
@@ -589,43 +675,65 @@ typedef struct NsServer {
      * The following struct maintains the core Tcl config.
      */
 
-    struct {         
-	char	           *library;
-	struct Trace       *traces[4];
+    struct {
+	/*
+	 * The following is the bootstrap script, normally bin/init.tcl.
+	 */
+
 	char	    	   *initfile;
-	Ns_RWLock	    lock;
-	char		   *script;
+
+	/*
+	 * The following support the loop control facilities.
+	 */
+
+	Tcl_HashTable loops;
+	Ns_Mutex      llock;
+	Ns_Cond	      lcond;
+
+	/*
+	 * The following support traces and one-time inits.
+	 */
+
+	Ns_RWLock	    tlock;	/* Lock for trace list. */
+	struct TclTrace    *firstTracePtr;
+	struct TclTrace    *lastTracePtr;
+	Ns_Cs		    olock;	/* Lock for one-time inits. */
+	Tcl_HashTable	    once;	/* Table of one-time inits. */
+	Ns_Mutex	    plock;	/* Lock for package table. */
+	Tcl_HashTable	    packages;	/* Table of server packages. */
+
+	/*
+	 * The following support the legacy module directories config.
+	 */
+
+	char	           *library;	/* Legacy library. */
+	char		   *script;	/* Legacy init script. */
 	int		    length;
 	int		    epoch;
-	Tcl_Obj		   *modules;
+	Ns_RWLock	    slock;	/* Lock for init script. */
+	Tcl_DString	    modules;	/* List of server modules. */
     } tcl;
-    
+
     /*
      * The following struct maintains ADP config,
      * registered tags, and read-only page text.
      */
 
     struct {
+	int		    flags;
+	int		    tracesize;
 	char	    	   *errorpage;
 	char	    	   *startpage;
-	bool	    	    enableexpire;
-	bool	    	    enabledebug;
 	char	    	   *debuginit;
-	char	    	   *defaultparser;
-	size_t 	    	    cachesize;
+	size_t		    bufsize;
+	size_t		    cachesize;
 	Ns_Cond	    	    pagecond;
 	Ns_Mutex	    pagelock;
-	Tcl_HashTable	    pages;
+	Tcl_HashTable       pages;
 	Ns_RWLock	    taglock;
-	Tcl_HashTable	    tags;
-
-        struct {
-            bool            enable;
-            int             level;
-            int             minsize;
-        } compress;
+	Tcl_HashTable       tags;
     } adp;
-    
+
     /*
      * The following struct maintains the Ns_Set's
      * entered into Tcl with NS_TCL_SET_SHARED.
@@ -637,14 +745,10 @@ typedef struct NsServer {
     } sets;    
     
     /*
-     * The following struct maintains the arrays
-     * for the nsv commands.
+     * The following maintains per-server nsv's.
      */
 
-    struct {
-	struct Bucket  	   *buckets;
-	int 	    	    nbuckets;
-    } nsv;
+    Nsv nsv;
     
     /*
      * The following struct maintains the vars and
@@ -687,12 +791,11 @@ typedef struct NsServer {
  */
 
 typedef struct NsInterp {
-
-    struct NsInterp	  *nextPtr;
-    Tcl_Interp		  *interp;
-    NsServer  	    	  *servPtr;
-    int		   	   delete;
-    int			   epoch;
+    struct NsInterp	  *nextPtr;	/* Next interps for given server. */
+    Tcl_Interp		  *interp;	/* Pointer to cooresponding interp. */
+    NsServer  	    	  *servPtr;	/* Pointer to interp server. */
+    int		   	   delete;	/* Delete interp on next deallocate. */
+    int			   epoch;	/* Epoch of legacy config. */
 
     /*
      * The following pointer maintains the first in
@@ -722,7 +825,7 @@ typedef struct NsInterp {
 
     Ns_Conn		  *conn;
 
-    struct {
+    struct nsconn {
 	int		   flags;
 	char	   	   form[16];
 	char	   	   hdrs[16];
@@ -734,22 +837,22 @@ typedef struct NsInterp {
      * context including the private pages cache.
      */
 
-    struct {
-	bool               stream;
-        bool               compress;
-	int	           exception;
-	int                depth;
-	int                objc;
-	Tcl_Obj		 **objv;
-	char              *cwd;
+    struct adp {
+	int		   flags;
+	int		   exception;
+	int		   refresh;
+	size_t		   bufsize;
 	int                errorLevel;
 	int                debugLevel;
 	int                debugInit;
 	char              *debugFile;
 	Ns_Cache	  *cache;
-	Tcl_DString	  *outputPtr;
-	Tcl_DString	  *responsePtr;
-	Tcl_DString	  *typePtr;
+	int                depth;
+	char		  *cwd;
+	struct AdpFrame	  *framePtr;
+	Ns_Conn		  *conn;
+	Tcl_Channel	   chan;
+	Tcl_DString	   output;
     } adp;
     
     /*
@@ -781,43 +884,48 @@ typedef struct NsInterp {
 extern void NsInitBinder(void);
 extern void NsInitCache(void);
 extern void NsInitConf(void);
+extern void NsInitConfig(void);
 extern void NsInitEncodings(void);
+extern void NsInitFd(void);
 extern void NsInitListen(void);
 extern void NsInitLog(void);
-extern void NsInitInfo(void);
 extern void NsInitMimeTypes(void);
 extern void NsInitModLoad(void);
+extern void NsInitNsv(void);
 extern void NsInitProcInfo(void);
 extern void NsInitQueue(void);
+extern void NsInitLimits(void);
+extern void NsInitPools(void);
 extern void NsInitDrivers(void);
 extern void NsInitSched(void);
+extern void NsInitServers(void);
 extern void NsInitTcl(void);
+extern void NsInitTclCache(void);
 extern void NsInitUrlSpace(void);
 extern void NsInitRequests(void);
-
-extern int NsQueueConn(Sock *sockPtr, Ns_Time *nowPtr);
-extern void NsMapPool(ConnPool *poolPtr, char *map);
-extern int NsSockSend(Sock *sockPtr, struct iovec *bufs, int nbufs);
+extern char *NsFindVersion(char *request, unsigned int *majorPtr,
+			   unsigned int *minorPtr);
+extern void NsQueueConn(Conn *connPtr);
+extern int NsCheckQuery(Ns_Conn *conn);
+extern void NsAppendConn(Tcl_DString *bufPtr, Conn *connPtr, char *state);
+extern void NsAppendRequest(Tcl_DString *dsPtr, Ns_Request *request);
+extern int  NsConnSend(Ns_Conn *conn, struct iovec *bufs, int nbufs);
 extern void NsSockClose(Sock *sockPtr, int keep);
-
-extern Request *NsGetRequest(Sock *sockPtr);
-extern void NsFreeRequest(Request *reqPtr);
-
+extern int  NsPoll(struct pollfd *pfds, int nfds, Ns_Time *timeoutPtr);
+extern void NsFreeConn(Conn *connPtr);
 extern NsServer *NsGetServer(char *server);
+extern char *NsGetServers(void);
 extern NsServer *NsGetInitServer(void);
-extern NsInterp *NsGetInterp(Tcl_Interp *interp);
-
-extern Ns_OpProc NsFastGet;
+extern NsInterp *NsGetInterpData(Tcl_Interp *interp);
+extern void NsFreeConnInterp(Conn *connPtr);
 extern Ns_OpProc NsAdpProc;
 
 extern Ns_Cache *NsFastpathCache(char *server, int size);
-
-extern void NsFreeAdp(NsInterp *itPtr);
-extern void NsFreeAtClose(NsInterp *itPtr);
-
-extern void NsRunAtClose(Tcl_Interp *interp);
-
-extern int NsUrlToFile(Ns_DString *dsPtr, NsServer *servPtr, char *url);
+extern void NsAdpInit(NsInterp *itPtr);
+extern void NsAdpReset(NsInterp *itPtr);
+extern void NsAdpFree(NsInterp *itPtr);
+extern void NsTclRunAtClose(NsInterp *itPtr);
+extern int  NsUrlToFile(Ns_DString *dsPtr, NsServer *servPtr, char *url);
 
 /*
  * External callback functions.
@@ -840,37 +948,44 @@ extern void NsGetCallbacks(Tcl_DString *dsPtr);
 extern void NsGetSockCallbacks(Tcl_DString *dsPtr);
 extern void NsGetScheduled(Tcl_DString *dsPtr);
 
+extern char *NsConnContent(Ns_Conn *conn, char **nextPtr, int *availPtr);
+extern void NsConnSeek(Ns_Conn *conn, int count);
+extern void *NsMap(int fd, off_t start, size_t len, int writeable, void **argPtr);
+extern void NsUnMap(void *addr, void *arg);
+
 #ifdef _WIN32
-extern int NsConnectService(void);
-extern int NsInstallService(char *service);
-extern int NsRemoveService(char *service);
+extern int  NsConnectService(void);
+extern int  NsInstallService(char *service);
+extern int  NsRemoveService(char *service);
 #endif
 
 extern void NsCreatePidFile(char *service);
 extern void NsRemovePidFile(char *service);
 
 extern void NsLogOpen(void);
+extern void NsLogConf(void);
 extern void NsTclInitObjs(void);
 extern void NsUpdateMimeTypes(void);
 extern void NsUpdateEncodings(void);
-extern void NsUpdateUrlEncode(void);
 extern void NsRunPreStartupProcs(void);
-extern void NsStartServers(void);
 extern void NsBlockSignals(int debug);
 extern void NsHandleSignals(void);
 extern void NsStopDrivers(void);
 extern void NsPreBind(char *bindargs, char *bindfile);
+extern SOCKET NsSockGetBound(struct sockaddr_in *saPtr);
 extern void NsClosePreBound(void);
 extern void NsInitServer(char *server, Ns_ServerInitProc *initProc);
 extern char *NsConfigRead(char *file);
 extern void NsConfigEval(char *config, int argc, char **argv, int optind);
 extern void NsConfUpdate(void);
-extern void NsEnableDNSCache(int timeout, int maxentries);
-extern void NsStopServers(Ns_Time *toPtr);
-extern void NsStartServer(NsServer *servPtr);
-extern void NsStopServer(NsServer *servPtr);
-extern void NsWaitServer(NsServer *servPtr, Ns_Time *toPtr);
-extern void NsStartDrivers(void);
+extern void NsEnableDNSCache(void);
+extern void NsStartPools(void);
+extern void NsStopPools(Ns_Time *timeoutPtr);
+extern int NsTclGetPool(Tcl_Interp *interp, char *pool, Pool **poolPtrPtr);
+extern Tcl_ObjCmdProc NsTclListPoolsObjCmd;
+extern void NsCreateConnThread(Pool *poolPtr);
+extern void NsJoinConnThreads(void);
+extern int  NsStartDrivers(void);
 extern void NsWaitDriversShutdown(Ns_Time *toPtr);
 extern void NsStartSchedShutdown(void);
 extern void NsWaitSchedShutdown(Ns_Time *toPtr);
@@ -878,21 +993,30 @@ extern void NsStartSockShutdown(void);
 extern void NsWaitSockShutdown(Ns_Time *toPtr);
 extern void NsStartShutdownProcs(void);
 extern void NsWaitShutdownProcs(Ns_Time *toPtr);
+extern void NsWaitDriversShutdown(Ns_Time *toPtr);
+extern void NsStartQueueShutdown(void);
+extern void NsWaitQueueShutdown(Ns_Time *toPtr);
 
 extern void NsStartJobsShutdown(void);
 extern void NsWaitJobsShutdown(Ns_Time *toPtr);
 
 extern void NsTclInitServer(char *server);
+extern int NsTclGetServer(NsInterp *itPtr, char **serverPtr);
+extern int NsTclGetConn(NsInterp *itPtr, Ns_Conn **connPtr);
 extern void NsLoadModules(char *server);
 extern struct Bucket *NsTclCreateBuckets(char *server, int nbuckets);
-
 extern void NsClsCleanup(Conn *connPtr);
 extern void NsTclAddCmds(Tcl_Interp *interp, NsInterp *itPtr);
-extern void NsTclAddServerCmds(Tcl_Interp *interp, NsInterp *itPtr);
-
 extern void NsRestoreSignals(void);
 extern void NsSendSignal(int sig);
 
+/*
+ * Conn routines.
+ */
+
+extern Limits *NsGetRequestLimits(char *server, char *method, char *url);
+extern Pool *NsGetConnPool(Conn *connPtr);
+
 /*
  * ADP routines.
  */
@@ -900,27 +1024,31 @@ extern void NsSendSignal(int sig);
 extern Ns_Cache *NsAdpCache(char *server, int size);
 extern void NsAdpSetMimeType(NsInterp *itPtr, char *type);
 extern void NsAdpSetCharSet(NsInterp *itPtr, char *charset);
-extern void NsAdpFlush(NsInterp *itPtr);
-extern void NsAdpStream(NsInterp *itPtr);
-extern void NsAdpCompress(NsInterp *itPtr, int compress);
+extern int NsAdpGetBuf(NsInterp *itPtr, Tcl_DString **dsPtrPtr);
+extern int NsAdpAppend(NsInterp *itPtr, char *buf, int len);
+extern int NsAdpFlush(NsInterp *itPtr, int stream);
 extern int NsAdpDebug(NsInterp *itPtr, char *host, char *port, char *procs);
-extern int NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int safe,
+extern int NsAdpEval(NsInterp *itPtr, int objc, Tcl_Obj *objv[], int flags,
                      char *resvar);
 extern int NsAdpSource(NsInterp *itPtr, int objc, Tcl_Obj *objv[],
-                       char *resvar);
-extern int NsAdpInclude(NsInterp *itPtr, char *file, int objc, Tcl_Obj *objv[]);
-extern void NsAdpParse(AdpParse *parsePtr, NsServer *servPtr, char *utf, int safe);
+                       int flags, char *resvar);
+extern int NsAdpInclude(NsInterp *itPtr, int objc, Tcl_Obj *objv[],
+			char *file, Ns_Time *ttlPtr);
+extern void NsAdpParse(AdpCode *codePtr, NsServer *servPtr, char *utf,
+		       int flags);
+extern void NsAdpFreeCode(AdpCode *codePtr);
+extern void NsAdpLogError(NsInterp *itPtr);
 
 /*
  * Tcl support routines.
  */
 
-extern Ns_TclInterpInitProc NsTclCreateCmds;
-extern char 	  *NsTclConnId(Ns_Conn *conn);
-extern int 	   NsIsIdConn(char *inID);
-extern void	   NsTclInitQueueType(void);
-extern void	   NsTclInitAddrType(void);
-extern void	   NsTclInitTimeType(void);
+extern int  NsTclCheckConnId(Tcl_Interp *interp, Tcl_Obj *objPtr);
+extern void NsTclInitQueueType(void);
+extern void NsTclInitAddrType(void);
+extern void NsTclInitCacheType(void);
+extern void NsTclInitKeylistType(void);
+extern void NsTclInitTimeType(void);
 
 /*
  * Callback routines.
@@ -935,11 +1063,14 @@ extern void NsRunStartupProcs(void);
 extern void NsRunAtReadyProcs(void);
 extern void NsRunAtExitProcs(void);
 
-
 /*
  * Utility functions.
  */
 
+extern bool  NsParamBool(char *key, bool def);
+extern int   NsParamInt(char *key, int def);
+extern char *NsParamString(char *key, char *def);
+
 extern int  NsCloseAllFiles(int errFd);
 #ifndef _WIN32
 extern int  Ns_ConnRunRequest(Ns_Conn *conn);
@@ -949,10 +1080,8 @@ extern int  Ns_TclGetOpenFd(Tcl_Interp *, char *, int write, int *fp);
 #endif
 extern void NsStopSockCallbacks(void);
 extern void NsStopScheduledProcs(void);
-extern void NsGetBuf(char **bufPtr, int *sizePtr);
-extern Tcl_Encoding NsGetTypeEncodingWithDef(char *type, int *used_default);
-extern void NsComputeEncodingFromType(char *type, Tcl_Encoding *enc,
-                                      int *new_type, Tcl_DString *type_ds);
+extern Tcl_Encoding NsGetInputEncoding(Conn *connPtr);
+extern Tcl_Encoding NsGetOutputEncoding(Conn *connPtr);
 
 /*
  * Proxy support
