@@ -34,7 +34,7 @@
  *	and service threads.
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/queue.c,v 1.23.2.1 2004/08/25 19:52:35 dossy Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/queue.c,v 1.38 2006/04/13 19:06:24 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
@@ -44,29 +44,28 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/queue
  * info for running threads.
  */
 
-typedef struct {
-    ConnPool *poolPtr;
+typedef struct ConnData {
+    struct ConnData *nextPtr;
+    Pool *poolPtr;
     Conn *connPtr;
-} Arg;
+    Ns_Thread thread;
+} ConnData;
 
 /*
  * Local functions defined in this file
  */
 
 static void ConnRun(Conn *connPtr);	/* Connection run routine. */
-static void ParseAuth(Conn *connPtr, char *auth);
-static void CreateConnThread(ConnPool *poolPtr);
-static void JoinConnThread(Ns_Thread *threadPtr);
-static void AppendConn(Tcl_DString *dsPtr, Conn *connPtr, char *state);
-static void AppendConnList(Tcl_DString *dsPtr, Conn *firstPtr,
-    	char *state);
+static void AppendConnList(Tcl_DString *dsPtr, Conn *firstPtr, char *state);
 
 /*
  * Static variables defined in this file.
  */
 
-static Ns_Tls argtls;
-static int poolid;
+static Ns_Tls        ctdtls;
+static Ns_Mutex	     connlock;
+static Ns_Mutex	     joinlock;
+static ConnData     *joinPtr;
 
 
 /*
@@ -88,8 +87,9 @@ static int poolid;
 void
 NsInitQueue(void)
 {
-    Ns_TlsAlloc(&argtls, NULL);
-    poolid = Ns_UrlSpecificAlloc();
+    Ns_TlsAlloc(&ctdtls, NULL);
+    Ns_MutexSetName(&connlock, "ns:connlock");
+    Ns_MutexSetName(&joinlock, "ns:joinlock");
 }
 
 
@@ -135,44 +135,10 @@ Ns_QueueConn(void *drv, void *arg)
 Ns_Conn *
 Ns_GetConn(void)
 {
-    Arg *argPtr;
+    ConnData *dataPtr;
 
-    argPtr = Ns_TlsGet(&argtls);
-    return (argPtr ? ((Ns_Conn *) argPtr->connPtr) : NULL);
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * NsMapPool --
- *
- *	Map a method/URL to the given pool.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	Requests for given URL's will be serviced by given pool.
- *
- *----------------------------------------------------------------------
- */
-
-void
-NsMapPool(ConnPool *poolPtr, char *map)
-{
-    char *server = poolPtr->servPtr->server;
-    char **mv;
-    int mc;
-
-    if (Tcl_SplitList(NULL, map, &mc, (CONST char***)&mv) == TCL_OK) {
-	if (mc == 2) {
-	    Ns_UrlSpecificSet(server, mv[0], mv[1], poolid, poolPtr, 0, NULL);
-	    Ns_Log(Notice, "pool[%s]: mapped %s %s -> %s", server, mv[0], mv[1], 
-               poolPtr->pool);
-	}
-	ckfree((char *) mv);
-    }
+    dataPtr = Ns_TlsGet(&ctdtls);
+    return (dataPtr ? ((Ns_Conn *) dataPtr->connPtr) : NULL);
 }
 
 
@@ -184,77 +150,47 @@ NsMapPool(ConnPool *poolPtr, char *map)
  *	Append a connection to the run queue.
  *
  * Results:
- *	1 if queued, 0 otherwise.
+ *	None.
  *
  * Side effects:
- *	Conneciton will run shortly.
+ *	Connection will run shortly.
  *
  *----------------------------------------------------------------------
  */
 
-int
-NsQueueConn(Sock *sockPtr, Ns_Time *nowPtr)
+void
+NsQueueConn(Conn *connPtr)
 {
-    NsServer *servPtr = sockPtr->servPtr;
-    ConnPool *poolPtr = NULL;
-    Conn *connPtr = NULL;
+    Pool *poolPtr = NsGetConnPool(connPtr);
     int create = 0;
 
     /*
-     * Select server connection pool.
+     * Queue connection.
      */
 
-    if (sockPtr->reqPtr != NULL) {
-	poolPtr = Ns_UrlSpecificGet(servPtr->server,
-			sockPtr->reqPtr->request->method,
-			sockPtr->reqPtr->request->url, poolid);
-    }
-    if (poolPtr == NULL) {
-	poolPtr = servPtr->pools.defaultPtr;
-    }
-
-   /*
-    * Queue connection if a free Conn is available.
-    */
-
-    Ns_MutexLock(&servPtr->pools.lock);
-    if (!servPtr->pools.shutdown) {
-	connPtr = poolPtr->queue.freePtr;
-	if (connPtr != NULL) {
-	    poolPtr->queue.freePtr = connPtr->nextPtr;
-	    connPtr->startTime = *nowPtr;
-	    connPtr->id = servPtr->pools.nextconnid++;
-	    connPtr->sockPtr = sockPtr;
-	    connPtr->drvPtr = sockPtr->drvPtr;
-	    connPtr->servPtr = servPtr;
-	    connPtr->server = servPtr->server;
-	    connPtr->location = sockPtr->location;
-	    if (poolPtr->queue.wait.firstPtr == NULL) {
-		poolPtr->queue.wait.firstPtr = connPtr;
-	    } else {
-		poolPtr->queue.wait.lastPtr->nextPtr = connPtr;
-	    }
-	    poolPtr->queue.wait.lastPtr = connPtr;
-	    connPtr->nextPtr = NULL;
-	    if (poolPtr->threads.idle == 0
-		    && poolPtr->threads.current < poolPtr->threads.max) {
-		++poolPtr->threads.idle;
-		++poolPtr->threads.current;
-		create = 1;
-	    }
-	    ++poolPtr->queue.wait.num;
-	}
-    }
-    Ns_MutexUnlock(&servPtr->pools.lock);
-    if (connPtr == NULL) {
-	return 0;
-    }
-    if (create) {
-    	CreateConnThread(poolPtr);
+    connPtr->flags |= NS_CONN_RUNNING;
+    Ns_MutexLock(&poolPtr->lock);
+    ++poolPtr->threads.queued;
+    if (poolPtr->queue.wait.firstPtr == NULL) {
+        poolPtr->queue.wait.firstPtr = connPtr;
     } else {
-	Ns_CondSignal(&poolPtr->queue.cond);
+        poolPtr->queue.wait.lastPtr->nextPtr = connPtr;
+    }
+    poolPtr->queue.wait.lastPtr = connPtr;
+    connPtr->nextPtr = NULL;
+    if (poolPtr->threads.idle == 0
+            && poolPtr->threads.current < poolPtr->threads.max) {
+        ++poolPtr->threads.idle;
+        ++poolPtr->threads.current;
+        create = 1;
+    }
+    ++poolPtr->queue.wait.num;
+    Ns_MutexUnlock(&poolPtr->lock);
+    if (create) {
+        NsCreateConnThread(poolPtr);
+    } else {
+        Ns_CondSignal(&poolPtr->cond);
     }
-    return 1;
 }
 
 
@@ -279,9 +215,7 @@ int
 NsTclServerObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
 		  Tcl_Obj **objv)
 {
-    NsInterp *itPtr = arg;
-    NsServer *servPtr = itPtr->servPtr;
-    ConnPool *poolPtr = servPtr->pools.defaultPtr;
+    Pool *poolPtr;
     char buf[100], *pool;
     Tcl_DString ds;
     static CONST char *opts[] = {
@@ -291,7 +225,7 @@ NsTclServerObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
     enum {
 	 SActiveIdx, SAllIdx, SConnectionsIdx, SKeepaliveIdx, SPoolsIdx,
 	 SQueuedIdx, SThreadsIdx, SWaitingIdx,
-    } opt;
+    } _nsmayalias opt;
 
     if (objc != 2 && objc != 3) {
 	Tcl_WrongNumArgs(interp, 1, objv, "option ?pool?");
@@ -301,27 +235,21 @@ NsTclServerObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
 			    (int *) &opt) != TCL_OK) {
 	return TCL_ERROR;
     }
+    if (opt == SPoolsIdx) {
+	return NsTclListPoolsObjCmd(arg, interp, objc, objv);
+    }
     if (objc == 2) {
-	poolPtr = servPtr->pools.defaultPtr;
+        pool = "default";
     } else {
 	pool = Tcl_GetString(objv[2]);
-	poolPtr = servPtr->pools.firstPtr;
-	while (poolPtr != NULL && !STREQ(poolPtr->pool, pool)) {
-	    poolPtr = poolPtr->nextPtr;
-	}
-	if (poolPtr == NULL) {
-	    Tcl_AppendResult(interp, "no such pool: ", pool, NULL);
-	    return TCL_ERROR;
-	}
     }
-    Ns_MutexLock(&servPtr->pools.lock);
+    if (NsTclGetPool(interp, pool, &poolPtr) != TCL_OK) {
+	return TCL_ERROR;
+    }
+    Ns_MutexLock(&poolPtr->lock);
     switch (opt) {
     case SPoolsIdx:
-	poolPtr = servPtr->pools.firstPtr;
-	while (poolPtr != NULL) {
-	    Tcl_AppendElement(interp, poolPtr->pool);
-	    poolPtr = poolPtr->nextPtr;
-	}
+	/* NB: Silence compiler. */
 	break;
 	  
     case SWaitingIdx:
@@ -329,11 +257,11 @@ NsTclServerObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
 	break;
 
     case SKeepaliveIdx:
-        Tcl_SetObjResult(interp, Tcl_NewIntObj(nsconf.keepalive.npending));
+        Tcl_SetObjResult(interp, Tcl_NewIntObj(0/*nsconf.keepalive.npending*/));
 	break;
 
     case SConnectionsIdx:
-        Tcl_SetObjResult(interp, Tcl_NewIntObj((int) servPtr->pools.nextconnid));
+        Tcl_SetObjResult(interp, Tcl_NewIntObj((int) poolPtr->threads.nextid));
 	break;
 
     case SThreadsIdx:
@@ -361,106 +289,10 @@ NsTclServerObjCmd(ClientData arg, Tcl_Interp *interp, int objc,
 	}
         Tcl_DStringResult(interp, &ds);
     }
-    Ns_MutexUnlock(&servPtr->pools.lock);
+    Ns_MutexUnlock(&poolPtr->lock);
     return TCL_OK;
 }
 
-
-/*
- *----------------------------------------------------------------------
- *
- * NsStartServer --
- *
- *	Start the core connection thread interface.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	Minimum connection threads may be created.
- *
- *----------------------------------------------------------------------
- */
-
-void
-NsStartServer(NsServer *servPtr)
-{
-    ConnPool *poolPtr;
-    int n;
-
-    poolPtr = servPtr->pools.firstPtr;
-    while (poolPtr != NULL) {
-    	poolPtr->threads.current = poolPtr->threads.idle = poolPtr->threads.min;
-    	for (n = 0; n < poolPtr->threads.min; ++n) {
-    	    CreateConnThread(poolPtr);
-	}
-	poolPtr = poolPtr->nextPtr;
-    }
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * NsStopServer --
- *
- *	Signal and wait for connection threads to exit.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	None.
- *
- *----------------------------------------------------------------------
- */
-
-void
-NsStopServer(NsServer *servPtr)
-{
-    ConnPool *poolPtr;
-
-    Ns_Log(Notice, "serv: stopping server: %s", servPtr->server);
-    Ns_MutexLock(&servPtr->pools.lock);
-    servPtr->pools.shutdown = 1;
-    Ns_MutexUnlock(&servPtr->pools.lock);
-    poolPtr = servPtr->pools.firstPtr;
-    while (poolPtr != NULL) {
-    	Ns_CondBroadcast(&poolPtr->queue.cond);
-	poolPtr = poolPtr->nextPtr;
-    }
-}
-
-void
-NsWaitServer(NsServer *servPtr, Ns_Time *toPtr)
-{
-    ConnPool *poolPtr;
-    Ns_Thread joinThread;
-    int status;
-    
-    status = NS_OK;
-    poolPtr = servPtr->pools.firstPtr;
-    Ns_MutexLock(&servPtr->pools.lock);
-    while (poolPtr != NULL && status == NS_OK) {
-    	while (status == NS_OK &&
-	       (poolPtr->queue.wait.firstPtr != NULL || poolPtr->threads.current > 0)) {
-	    status = Ns_CondTimedWait(&poolPtr->queue.cond, &servPtr->pools.lock, toPtr);
-    	}
-	poolPtr = poolPtr->nextPtr;
-    }
-    joinThread = servPtr->pools.joinThread;
-    servPtr->pools.joinThread = NULL;
-    Ns_MutexUnlock(&servPtr->pools.lock);
-    if (status != NS_OK) {
-	Ns_Log(Warning, "serv: timeout waiting for connection thread exit");
-    } else {
-	if (joinThread != NULL) {
-	    JoinConnThread(&joinThread);
-	}
-	Ns_Log(Notice, "serv: connection threads stopped");
-    }
-}
-
 
 /*
  *----------------------------------------------------------------------
@@ -473,7 +305,7 @@ NsWaitServer(NsServer *servPtr, Ns_Time *toPtr)
  *	None.
  *
  * Side effects:
- *	See AppendConn.
+ *	See NsAppendConn.
  *
  *----------------------------------------------------------------------
  */
@@ -481,17 +313,15 @@ NsWaitServer(NsServer *servPtr, Ns_Time *toPtr)
 void
 NsConnArgProc(Tcl_DString *dsPtr, void *arg)
 {
-    Arg *argPtr = arg;
+    ConnData *dataPtr = arg;
     
-    /*
-     * A race condition here causes problems occasionally.
-     */
-
-    if (arg != NULL) {
-    	AppendConn(dsPtr, argPtr->connPtr, "running");
+    Ns_MutexLock(&connlock);
+    if (dataPtr->connPtr != NULL) {
+        NsAppendConn(dsPtr, dataPtr->connPtr, "running");
     } else {
     	Tcl_DStringAppendElement(dsPtr, "");
     }
+    Ns_MutexUnlock(&connlock);
 }
 
 
@@ -514,55 +344,32 @@ NsConnArgProc(Tcl_DString *dsPtr, void *arg)
 void
 NsConnThread(void *arg)
 {
-    Arg    	    *argPtr = arg;
-    ConnPool	    *poolPtr = argPtr->poolPtr;
-    NsServer	    *servPtr = poolPtr->servPtr;
+    ConnData  	    *dataPtr = arg;
+    Pool            *poolPtr = dataPtr->poolPtr;
     Conn            *connPtr;
     Ns_Time          wait, *timePtr;
-    unsigned int     id;
-    Ns_DString	     ds;
-    int              status, cpt, ncons;
-    char            *p, *path;
-    Ns_Thread	     joinThread;
+    char             name[100];
+    int              status, ncons;
+    char            *msg;
     
     /*
      * Set the conn thread name.
      */
 
-    Ns_TlsSet(&argtls, argPtr);
-    Ns_MutexLock(&servPtr->pools.lock);
-    id = poolPtr->threads.nextid++;
-    Ns_MutexUnlock(&servPtr->pools.lock);
-    Ns_DStringInit(&ds);
-    Ns_DStringVarAppend(&ds, "-conn:", servPtr->server, NULL);
-    if (poolPtr->pool != NULL) {
-    	Ns_DStringVarAppend(&ds, ":", poolPtr->pool, NULL);
-    }
-    Ns_DStringPrintf(&ds, ":%d", id);
-    Ns_ThreadSetName(ds.string);
-    Ns_DStringFree(&ds);
-
-    /*
-     * See how many connections this thread should run.
-     * Setting this parameter to > 0 will cause the
-     * thread to graceously exit, after processing that
-     * many requests, thus initiating kind-of Tcl-level
-     * garbage collection.
-     */
-
-    path = Ns_ConfigGetPath(servPtr->server, NULL, NULL);
-    if (!Ns_ConfigGetInt(path, "connsperthread", &cpt)) {
-        cpt = 0; /* == unlimited # of connections */
-    }
-
-    ncons = cpt;
-
+    Ns_TlsSet(&ctdtls, dataPtr);
+    Ns_MutexLock(&poolPtr->lock);
+    sprintf(name, "-conn:%d-", poolPtr->threads.nextid++);
+    Ns_MutexUnlock(&poolPtr->lock);
+    Ns_ThreadSetName(name);
+    ncons = poolPtr->threads.maxconns;
+    msg = "exceeded max connections per thread";
+    
     /*
      * Start handling connections.
      */
 
-    Ns_MutexLock(&servPtr->pools.lock);
-    while (cpt == 0 || ncons > 0) {
+    Ns_MutexLock(&poolPtr->lock);
+    while (poolPtr->threads.maxconns <= 0 || ncons-- > 0) {
 
 	/*
 	 * Wait for a connection to arrive, exiting if one doesn't
@@ -578,17 +385,18 @@ NsConnThread(void *arg)
 	}
 
 	status = NS_OK;
-    	while (!servPtr->pools.shutdown
+        while (!poolPtr->shutdown
 		&& status == NS_OK
-		&& poolPtr->queue.wait.firstPtr == NULL) {
-	    status = Ns_CondTimedWait(&poolPtr->queue.cond, &servPtr->pools.lock, timePtr);
+                && poolPtr->queue.wait.firstPtr == NULL) {
+            status = Ns_CondTimedWait(&poolPtr->cond, &poolPtr->lock, timePtr);
 	}
 	if (poolPtr->queue.wait.firstPtr == NULL) {
+	    msg = "timeout waiting for connection";
 	    break;
 	}
 
 	/*
-	 * Pull the first connection of the waiting list.
+	 * Pull the first connection off the waiting list.
 	 */
 
     	connPtr = poolPtr->queue.wait.firstPtr;
@@ -607,21 +415,28 @@ NsConnThread(void *arg)
 	}
 	poolPtr->threads.idle--;
 	poolPtr->queue.wait.num--;
-	argPtr->connPtr = connPtr;
-    	Ns_MutexUnlock(&servPtr->pools.lock);
+        Ns_MutexUnlock(&poolPtr->lock);
 
 	/*
 	 * Run the connection.
 	 */
 
-	ConnRun(connPtr);
+	Ns_MutexLock(&connlock);
+        dataPtr->connPtr = connPtr;
+	Ns_MutexUnlock(&connlock);
+
+	Ns_GetTime(&connPtr->times.run);
+        ConnRun(connPtr);
+
+	Ns_MutexLock(&connlock);
+        dataPtr->connPtr = NULL;
+	Ns_MutexUnlock(&connlock);
 
 	/*
 	 * Remove from the active list and push on the free list.
 	 */
 
-	Ns_MutexLock(&servPtr->pools.lock);
-	argPtr->connPtr = NULL;
+        Ns_MutexLock(&poolPtr->lock);
 	if (connPtr->prevPtr != NULL) {
 	    connPtr->prevPtr->nextPtr = connPtr->nextPtr;
 	} else {
@@ -633,39 +448,36 @@ NsConnThread(void *arg)
 	    poolPtr->queue.active.lastPtr = connPtr->prevPtr;
 	}
 	poolPtr->threads.idle++;
-	connPtr->prevPtr = NULL;
-	connPtr->nextPtr = poolPtr->queue.freePtr;
-	poolPtr->queue.freePtr = connPtr;
-	if (connPtr->nextPtr == NULL) {
-	    /*
-	     * If this thread just free'd up the busy server,
-	     * run the ready procs to signal other subsystems.
-	     */
-	    Ns_MutexUnlock(&servPtr->pools.lock);
-	    NsRunAtReadyProcs();
-	    Ns_MutexLock(&servPtr->pools.lock);
-	}
-	if (cpt && --ncons <= 0) {
-	    break; /* Served given # of connections in this thread */
-    	}
+        Ns_MutexUnlock(&poolPtr->lock);
+        NsFreeConn(connPtr);
+        Ns_MutexLock(&poolPtr->lock);
+    }
+
+    /*
+     * Append this thread to list of threads to reap.
+     */
+
+    Ns_MutexLock(&joinlock);
+    dataPtr->nextPtr = joinPtr;
+    joinPtr = dataPtr;
+    Ns_MutexUnlock(&joinlock);
+
+    /*
+     * Mark this thread as no longer active.
+     */
+
+    if (poolPtr->shutdown) {
+	msg = "shutdown pending";
     }
     poolPtr->threads.idle--;
     poolPtr->threads.current--;
     if (poolPtr->threads.current == 0) {
-    	Ns_CondBroadcast(&poolPtr->queue.cond);
+        Ns_CondBroadcast(&poolPtr->cond);
     }
-    if (servPtr->pools.shutdown) {
-	p = "shutdown pending";
-    } else {
-	p = "no waiting connections";
-    }
-    joinThread = servPtr->pools.joinThread;
-    Ns_ThreadSelf(&servPtr->pools.joinThread);
-    Ns_MutexUnlock(&servPtr->pools.lock);
-    if (joinThread != NULL) {
-	JoinConnThread(&joinThread);
-    }
-    Ns_ThreadExit(argPtr);
+    Ns_MutexUnlock(&poolPtr->lock);
+
+    Ns_Log(Notice, "exiting: %s", msg);
+    Ns_ThreadExit(dataPtr);
 }
 
 
@@ -691,38 +503,22 @@ ConnRun(Conn *connPtr)
 {
     Ns_Conn 	  *conn = (Ns_Conn *) connPtr;
     NsServer	  *servPtr = connPtr->servPtr;
-    Tcl_HashEntry *hPtr;
-    Tcl_HashSearch search;
-    FormFile	  *filePtr;
     int            i, status;
-    char	  *auth;
+    Tcl_Encoding    encoding = NULL;
 	
     /*
-     * Re-initialize and run the connection.
+     * Initialize the connection encodings and headers. 
      */
-
-    connPtr->reqPtr = NsGetRequest(connPtr->sockPtr);
-    if (connPtr->reqPtr == NULL) {
-        Ns_ConnClose(conn);
-	return;
-    }
-    connPtr->contentLength = connPtr->reqPtr->length;
-    connPtr->headers = connPtr->reqPtr->headers;
-    connPtr->request = connPtr->reqPtr->request;
-    connPtr->flags = 0;
-    connPtr->nContentSent = 0;
-    connPtr->responseStatus = 0;
-    connPtr->responseLength = 0;
-    connPtr->recursionCount = 0;
-    connPtr->encoding = servPtr->encoding.outputEncoding;
-    connPtr->urlEncoding = servPtr->encoding.urlEncoding;
-    Tcl_DStringInit(&connPtr->queued);
-    Tcl_InitHashTable(&connPtr->files, TCL_STRING_KEYS);
-    sprintf(connPtr->idstr, "cns%d", connPtr->id);
+    
     connPtr->outputheaders = Ns_SetCreate(NULL);
-    if (connPtr->request->version < 1.0) {
-	conn->flags |= NS_CONN_SKIPHDRS;
+    encoding = NsGetInputEncoding(connPtr);
+    if (encoding == NULL) {
+    	encoding = NsGetOutputEncoding(connPtr);
+	if (encoding == NULL) {
+	    encoding = connPtr->servPtr->urlEncoding;
+	}
     }
+    Ns_ConnSetUrlEncoding(conn, encoding);
     if (servPtr->opts.hdrcase != Preserve) {
 	for (i = 0; i < Ns_SetSize(connPtr->headers); ++i) {
     	    if (servPtr->opts.hdrcase == ToLower) {
@@ -732,13 +528,6 @@ ConnRun(Conn *connPtr)
 	    }
 	}
     }
-    auth = Ns_SetIGet(connPtr->headers, "authorization");
-    if (auth != NULL) {
-	ParseAuth(connPtr, auth);
-    }
-    if (conn->request->method && STREQ(conn->request->method, "HEAD")) {
-	conn->flags |= NS_CONN_SKIPBODY;
-    }
 
     /*
      * Run the request.
@@ -751,7 +540,7 @@ ConnRun(Conn *connPtr)
 	if (status == NS_OK) {
 	    status = Ns_AuthorizeRequest(servPtr->server,
 			connPtr->request->method, connPtr->request->url, 
-			connPtr->authUser, connPtr->authPasswd, connPtr->reqPtr->peer);
+			connPtr->authUser, connPtr->authPasswd, connPtr->peer);
 	    switch (status) {
 	    case NS_OK:
 		status = NsRunFilters(conn, NS_FILTER_POST_AUTH);
@@ -793,93 +582,30 @@ ConnRun(Conn *connPtr)
     }
 
     /*
-     * Perform various garbage collection tasks.  Note
-     * the order is significant:  The driver freeProc could
-     * possibly use Tcl and Tcl deallocate callbacks
-     * could possibly access header and/or request data.
+     * Cleanup the connections, calling any registered cleanup traces
+     * followed by free the connection interp if it was used.
      */
 
     NsRunCleanups(conn);
-    NsClsCleanup(connPtr);
-    Ns_FreeConnInterp(conn);
-    if (connPtr->authUser != NULL) {
-	ns_free(connPtr->authUser);
-	connPtr->authUser = connPtr->authPasswd = NULL;
+    NsFreeConnInterp(connPtr);
+    if (connPtr->type != NULL) {
+        Ns_ConnSetType(conn, NULL);
     }
     if (connPtr->query != NULL) {
-	Ns_SetFree(connPtr->query);
-	connPtr->query = NULL;
+        Ns_ConnClearQuery(conn);
     }
-    hPtr = Tcl_FirstHashEntry(&connPtr->files, &search);
-    while (hPtr != NULL) {
-	filePtr = Tcl_GetHashValue(hPtr);
-	Ns_SetFree(filePtr->hdrs);
-	ns_free(filePtr);
-	hPtr = Tcl_NextHashEntry(&search);
+    if (connPtr->outputheaders != NULL) {
+    	Ns_SetFree(connPtr->outputheaders);
+	connPtr->outputheaders = NULL;
     }
-    Tcl_DeleteHashTable(&connPtr->files);
-    Tcl_DStringFree(&connPtr->queued);
-    Ns_SetFree(connPtr->outputheaders);
-    connPtr->outputheaders = NULL;
-    NsFreeRequest(connPtr->reqPtr);
-    connPtr->reqPtr = NULL;
+    Ns_DStringTrunc(&connPtr->obuf, 0);
 }
 
 
 /*
  *----------------------------------------------------------------------
  *
- * ParseAuth --
- *
- *	Parse an HTTP authorization string.
- *
- * Results:
- *	None.
- *
- * Side effects:
- *	May set the authPasswd and authUser connection pointers.
- *
- *----------------------------------------------------------------------
- */
-
-static void
-ParseAuth(Conn *connPtr, char *auth)
-{
-    register char *p, *q;
-    int            n;
-    char    	   save;
-    
-    p = auth;
-    while (*p != '\0' && !isspace(UCHAR(*p))) {
-        ++p;
-    }
-    if (*p != '\0') {
-    	save = *p;
-	*p = '\0';
-        if (STRIEQ(auth, "Basic")) {
-    	    q = p + 1;
-            while (*q != '\0' && isspace(UCHAR(*q))) {
-                ++q;
-            }
-	    n = strlen(q) + 3;
-	    connPtr->authUser = ns_malloc((size_t) n);
-            n = Ns_HtuuDecode(q, (unsigned char *) connPtr->authUser, n);
-            connPtr->authUser[n] = '\0';
-            q = strchr(connPtr->authUser, ':');
-            if (q != NULL) {
-                *q++ = '\0';
-                connPtr->authPasswd = q;
-            }
-        }
-	*p = save;
-    }
-}
-
-
-/*
- *----------------------------------------------------------------------
- *
- * CreateConnThread --
+ * NsCreateConnThread --
  *
  *	Create a connection thread.
  *
@@ -892,25 +618,34 @@ ParseAuth(Conn *connPtr, char *auth)
  *----------------------------------------------------------------------
  */
 
-static void
-CreateConnThread(ConnPool *poolPtr)
+void
+NsCreateConnThread(Pool *poolPtr)
 {
-    Ns_Thread thread;
-    Arg *argPtr;
+    ConnData *dataPtr;
 
-    argPtr = ns_malloc(sizeof(Arg));
-    argPtr->poolPtr = poolPtr;
-    argPtr->connPtr = NULL;
-    Ns_ThreadCreate(NsConnThread, argPtr, 0, &thread);
+    /*
+     * Reap any dead threads.
+     */
+
+    NsJoinConnThreads();
+
+    /*
+     * Create a new connection thread.
+     */
+
+    dataPtr = ns_malloc(sizeof(ConnData));
+    dataPtr->poolPtr = poolPtr;
+    dataPtr->connPtr = NULL;
+    Ns_ThreadCreate(NsConnThread, dataPtr, 0, &dataPtr->thread);
 }
-
+ 
 
 /*
  *----------------------------------------------------------------------
  *
- * JoinConnThread --
+ * NsJoinConnThreads --
  *
- *	Join a connection thread, freeing the threads connPtrPtr.
+ *	Join any connection threads which have exited.
  *
  * Results:
  *	None.
@@ -921,20 +656,28 @@ CreateConnThread(ConnPool *poolPtr)
  *----------------------------------------------------------------------
  */
 
-static void
-JoinConnThread(Ns_Thread *threadPtr)
+void
+NsJoinConnThreads(void)
 {
-    Arg *argPtr;
+    ConnData *firstPtr;
+    void *arg;
 
-    Ns_ThreadJoin(threadPtr, (void **) &argPtr);
-    ns_free(argPtr);
+    Ns_MutexLock(&joinlock);
+    firstPtr = joinPtr;
+    joinPtr = NULL;
+    Ns_MutexUnlock(&joinlock);
+    while (firstPtr != NULL) {
+	Ns_ThreadJoin(&firstPtr->thread, &arg);
+	firstPtr = firstPtr->nextPtr;
+	ns_free(arg);
+    }
 }
 
 
 /*
  *----------------------------------------------------------------------
  *
- * AppendConn --
+ * NsAppendConn --
  *
  *	Append connection data to a dstring.
  *
@@ -947,44 +690,20 @@ JoinConnThread(Ns_Thread *threadPtr)
  *----------------------------------------------------------------------
  */
 
-static void
-AppendConn(Tcl_DString *dsPtr, Conn *connPtr, char *state)
+void
+NsAppendConn(Tcl_DString *dsPtr, Conn *connPtr, char *state)
 {
-    char buf[100];
-    char  *p;
     Ns_Time now, diff;
 
+    Ns_GetTime(&now);
+    Ns_DiffTime(&now, &connPtr->times.queue, &diff);
     Tcl_DStringStartSublist(dsPtr);
-
-    /*
-     * An annoying race condition can be lethal here.
-     */
-    if ( connPtr != NULL ) {
-	sprintf(buf, "%d", connPtr->id);
-	Tcl_DStringAppendElement(dsPtr, buf);
-	Tcl_DStringAppendElement(dsPtr, Ns_ConnPeer((Ns_Conn *) connPtr));
-	Tcl_DStringAppendElement(dsPtr, state);
-	
-	/*
-	 * Carefully copy the bytes to avoid chasing a pointer
-	 * which may be changing in the connection thread.  This
-	 * is not entirely safe but acceptible for a seldom-used
-	 * admin command.
-	 */
-	
-	p = (connPtr->request && connPtr->request->method) ?
-	    connPtr->request->method : "?";
-	Tcl_DStringAppendElement(dsPtr, strncpy(buf, p, sizeof(buf)));
-	p = (connPtr->request && connPtr->request->url) ?
-	    connPtr->request->url : "?";
-	Tcl_DStringAppendElement(dsPtr, strncpy(buf, p, sizeof(buf)));
-	Ns_GetTime(&now);
-	Ns_DiffTime(&now, &connPtr->startTime, &diff);
-	sprintf(buf, "%ld.%ld", diff.sec, diff.usec);
-	Tcl_DStringAppendElement(dsPtr, buf);
-	sprintf(buf, "%d", connPtr->nContentSent);
-	Tcl_DStringAppendElement(dsPtr, buf);
-    }
+    Ns_DStringPrintf(dsPtr, "%d", connPtr->id);
+    Tcl_DStringAppendElement(dsPtr, Ns_ConnPeer((Ns_Conn *) connPtr));
+    Tcl_DStringAppendElement(dsPtr, state);
+    NsAppendRequest(dsPtr, connPtr->request);
+    Ns_DStringPrintf(dsPtr, " %ld.%ld %d",
+		     diff.sec, diff.usec, connPtr->nContentSent);
     Tcl_DStringEndSublist(dsPtr);
 }
 
@@ -1009,7 +728,7 @@ static void
 AppendConnList(Tcl_DString *dsPtr, Conn *firstPtr, char *state)
 {
     while (firstPtr != NULL) {
-	AppendConn(dsPtr, firstPtr, state);
+	NsAppendConn(dsPtr, firstPtr, state);
 	firstPtr = firstPtr->nextPtr;
     }
 }
