@@ -35,7 +35,7 @@
  *	Pool memory is used as an optimization.
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/request.c,v 1.7 2003/03/07 18:08:33 vasiljevic Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/request.c,v 1.12 2005/08/10 13:24:41 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
@@ -44,8 +44,10 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/reque
 /*
  * Local functions defined in this file.
  */
-static void SetUrl(Ns_Request * request, char *url);
+
+static void SetUrl(Ns_Request * request, char *url, Tcl_Encoding encoding);
 static void FreeUrl(Ns_Request * request);
+static Ns_Mutex reqlock;
 
 
 /*
@@ -82,9 +84,10 @@ Ns_FreeRequest(Ns_Request * request)
 /*
  *----------------------------------------------------------------------
  *
- * Ns_ParseRequest --
+ * Ns_ParseRequest, Ns_ParseRequestEx --
  *
  *	Parse a request from a browser into an Ns_Request structure. 
+ *	Utilize the given encoding, if present.
  *
  * Results:
  *	A new Ns_Request. 
@@ -97,11 +100,19 @@ Ns_FreeRequest(Ns_Request * request)
 
 Ns_Request *
 Ns_ParseRequest(char *line)
+{
+    return Ns_ParseRequestEx(line, NULL);
+}
+
+Ns_Request *
+Ns_ParseRequestEx(char *line, Tcl_Encoding encoding)
 {
     char       *url;
     char       *p;
     Ns_DString  ds;
     Ns_Request *request;
+    unsigned int major, minor;
+    int i;
 
     request = ns_calloc(1, sizeof(Ns_Request));
     Ns_DStringInit(&ds);
@@ -151,25 +162,14 @@ Ns_ParseRequest(char *line)
      */
     
     request->version = 0.0;
-    p = url + strlen(url);
-    while (p-- > url) {
-        if (!isdigit(UCHAR(*p)) && *p != '.') {
-            break;
-        }
-    }
-    p -= (sizeof(HTTP) - 2);
-    if (p >= url) {
-        if (strncmp(p, HTTP, sizeof(HTTP) - 1) == 0) {
-
-            /*
-             * If atof fails, version will be set to 0 and the server
-             * will treat the connection as if it had no HTTP/n.n keyword.
-             */
-	    
-            *p = '\0';
-            p += sizeof(HTTP) - 1;
-            request->version = atof(p);
-        }
+    p = NsFindVersion(url, &major, &minor);
+    if (p != NULL) {
+	*p = '\0';
+	i = 10;
+	while ((minor / i) > 0) {
+	    i *= 10;
+	}
+	request->version = major + (double) minor / i;
     }
     url = Ns_StrTrim(url);
     if (*url == '\0') {
@@ -212,13 +212,16 @@ Ns_ParseRequest(char *line)
                 p = strchr(h, ':');
                 if (p != NULL) {
                     *p++ = '\0';
-                    request->port = atoi(p);
+		    if (Tcl_GetInt(NULL, p, &i) != TCL_OK) {
+			goto done;
+		    }
+		    request->port = (unsigned short) i;
                 }
                 request->host = ns_strdup(h);
             }
         }
     }
-    SetUrl(request, url);
+    SetUrl(request, url, encoding);
 
 done:
     if (request->url == NULL) {
@@ -229,6 +232,49 @@ done:
     return request;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsFindVersion --
+ *
+ *	Find the HTTP/x.y version string in a request line.
+ *
+ * Results:
+ *	Start of version string or NULL if not found.
+ *
+ * Side effects:
+ *	Given majorPtr and minorPtr will be updated.
+ *
+ *----------------------------------------------------------------------
+ */
+
+char *
+NsFindVersion(char *request, unsigned int *majorPtr, unsigned int *minorPtr)
+{
+    char *next, *version;
+    int major, minor;
+
+    version = NULL;
+    while ((next = strstr(request, " HTTP/")) != NULL) {
+	version = next;
+	request += 6;
+    }
+    if (version != NULL && sscanf(version+6, "%u.%u", &major, &minor) != 2) {
+	version = NULL;
+    }
+    if (version != NULL) {
+	++version;
+	if (majorPtr != NULL) {
+	    *majorPtr = major;
+	}
+	if (minorPtr != NULL) {
+	    *minorPtr = minor;
+	}
+    }
+    return version;
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -283,10 +329,12 @@ Ns_SetRequestUrl(Ns_Request * request, char *url)
 {
     Ns_DString      ds;
 
+    Ns_MutexLock(&reqlock);
     FreeUrl(request);
     Ns_DStringInit(&ds);
     Ns_DStringAppend(&ds, url);
-    SetUrl(request, ds.string);
+    SetUrl(request, ds.string, NULL);
+    Ns_MutexUnlock(&reqlock);
     Ns_DStringFree(&ds);
 }
 
@@ -339,8 +387,9 @@ FreeUrl(Ns_Request * request)
  */
 
 static void
-SetUrl(Ns_Request * request, char *url)
+SetUrl(Ns_Request *request, char *url, Tcl_Encoding encoding)
 {
+    Ns_Conn *conn;
     Ns_DString  ds1, ds2;
     char       *p;
 
@@ -361,10 +410,14 @@ SetUrl(Ns_Request * request, char *url)
     }
 
     /*
-     * Decode and normalize the URL.
+     * Decode and normalize the URL.  If the encoding isn't specified,
+     * use the encoding of the current connection, if any.
      */
 
-    p = Ns_DecodeUrlCharset(&ds1, url, NULL);
+    if (encoding == NULL && (conn = Ns_GetConn()) != NULL) {
+	encoding = Ns_ConnGetUrlEncoding(conn);
+    }
+    p = Ns_DecodeUrlWithEncoding(&ds1, url, encoding);
     if (p == NULL) {
         p = url;
     }
@@ -491,3 +544,33 @@ Ns_ParseHeader(Ns_Set *set, char *line, Ns_HeaderCaseDisposition disp)
     }
     return NS_OK;
 }
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsAppendRequest --
+ *
+ *	Utility function to append string of given request.
+
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	None
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+NsAppendRequest(Tcl_DString *dsPtr, Ns_Request *request)
+{
+    if (request == NULL) {
+	Tcl_DStringAppend(dsPtr, " ? ?", -1);
+    } else {
+    	Ns_MutexLock(&reqlock);
+	Tcl_DStringAppendElement(dsPtr, request->method);
+	Tcl_DStringAppendElement(dsPtr, request->url);
+    	Ns_MutexUnlock(&reqlock);
+    }
+}
