@@ -34,25 +34,21 @@
  *	Functions that return data to a browser. 
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/return.c,v 1.33.2.6 2004/07/19 02:47:10 dossy Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/return.c,v 1.49 2006/04/19 17:49:01 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
-#define MAX_RECURSION 3       /* Max return direct recursion limit. */
-
-#define HTTP11_HDR_TE "Transfer-Encoding"
-#define HTTP11_TE_CHUNKED "chunked"
-
-
 /*
  * Local functions defined in this file
  */
 
 static int ReturnRedirect(Ns_Conn *conn, int status, int *resultPtr);
 static int ReturnOpen(Ns_Conn *conn, int status, char *type, Tcl_Channel chan,
-		      FILE *fp, int fd, int len);
-static int ReturnCharData(Ns_Conn *conn, int status, char *data, int len,
-                          char *type, int sendRaw);
+		      FILE *fp, int fd, off_t off, int len);
+static int ReturnData(Ns_Conn *conn, int status, char *data, int len,
+                          char *type, int direct);
+static int HdrEq(Ns_Set *hdrs, char *key, char *value);
+static int CheckKeep(Ns_Conn *conn, int status);
 
 /*
  * This structure connections HTTP response codes to their descriptions.
@@ -115,7 +111,7 @@ static struct {
  * Static variables defined in this file.
  */
 
-static int             nreasons = (sizeof(reasons) / sizeof(reasons[0]));
+static int nreasons = (sizeof(reasons) / sizeof(reasons[0]));
 
 
 /*
@@ -158,29 +154,6 @@ Ns_RegisterReturn(int status, char *url)
     }
 }
 
-static int
-IsSetupForChunkedEncoding(Ns_Conn *conn) 
-{
-    int headerCount = 0;
-    int i;
-    Ns_Set* outHeaders;
-
-    if (conn == NULL)
-        return 0;
-
-    outHeaders = Ns_ConnOutputHeaders(conn);
-    headerCount = Ns_SetSize (outHeaders);
-
-    if (outHeaders && headerCount) {
-        for (i = 0 ; i < headerCount ; i++) {
-           if ( !strcasecmp(Ns_SetKey (outHeaders, i), HTTP11_HDR_TE) &&
-	        !strcasecmp(Ns_SetValue (outHeaders, i), HTTP11_TE_CHUNKED) )
-	       return 1;
-           }
-    }
-    return 0;
-}
-
 
 /*
  *----------------------------------------------------------------------
@@ -202,34 +175,28 @@ IsSetupForChunkedEncoding(Ns_Conn *conn)
 void
 Ns_ConnConstructHeaders(Ns_Conn *conn, Ns_DString *dsPtr)
 {
-    int   i, length;
+    Conn *connPtr = (Conn *) conn;
+    int   i, status;
     char *reason;
-    char  buf[100];
     char *value, *keep;
-    char *key, *lengthHdr;
-    Conn *connPtr;
-    int  doChunkEncoding = 0;
+    char *key;
 
     /*
      * Construct the HTTP response status line.
      */
 
-    connPtr = (Conn *) conn;
-    sprintf(buf, "%d", connPtr->responseStatus);
+    status = Ns_ConnGetStatus(conn);
     reason = "Unknown Reason";
     for (i = 0; i < nreasons; i++) {
-	if (reasons[i].status == connPtr->responseStatus) {
+	if (reasons[i].status == status) {
 	    reason = reasons[i].reason;
 	    break;
 	}
     }
-
-    doChunkEncoding = IsSetupForChunkedEncoding(conn);
-
-    if (!doChunkEncoding)
-        Ns_DStringVarAppend(dsPtr, "HTTP/1.0 ", buf, " ", reason, "\r\n", NULL);
-    else
-        Ns_DStringVarAppend(dsPtr, "HTTP/1.1 ", buf, " ", reason, "\r\n", NULL);
+    Ns_DStringPrintf(dsPtr, "HTTP/%u.%u %d %s\r\n", 
+		     _MIN((connPtr->major), nsconf.http.major),
+		     _MIN((connPtr->minor), nsconf.http.minor),
+		     status, reason);
 
     /*
      * Output any headers.
@@ -237,42 +204,23 @@ Ns_ConnConstructHeaders(Ns_Conn *conn, Ns_DString *dsPtr)
 
     if (conn->outputheaders != NULL) {
 	/*
-	 * Update the response length value directly from the
-	 * header to be sent, i.e., don't trust programmers
-	 * correctly called Ns_ConnSetLengthHeader().
+	 * Set keep-alive if the driver and connection support it.
 	 */
 
-	length = connPtr->responseLength;
-	lengthHdr = Ns_SetIGet(conn->outputheaders, "content-length");
-	if (lengthHdr != NULL) {
-	    connPtr->responseLength = atoi(lengthHdr);
+	if (!Ns_ConnGetKeepAliveFlag(conn) && CheckKeep(conn, status)) {
+	    Ns_ConnSetKeepAliveFlag(conn, NS_TRUE);
 	}
-	
-	/*
-	 * Output a connection keep-alive header only on
-	 * any HTTP status 200 response which included
-	 * a valid and correctly set content-length header.
-	 */
-
-	if (nsconf.keepalive.enabled &&
-	    connPtr->headers != NULL &&
-	    connPtr->request != NULL &&
-	    (( (connPtr->responseStatus >= 200 && connPtr->responseStatus < 300) &&
-	    (lengthHdr != NULL &&
-	    connPtr->responseLength == length) || doChunkEncoding) ||
-	    (connPtr->responseStatus == 304 || connPtr->responseStatus == 201 || connPtr->responseStatus == 207) ) &&
-	    (nsconf.keepalive.allmethods == NS_TRUE ||
-	    STREQ(connPtr->request->method, "GET")) &&
-	    (key = Ns_SetIGet(conn->headers, "connection")) != NULL &&
-	    STRIEQ(key, "keep-alive")) {
-	    conn->flags |= NS_CONN_KEEPALIVE;
+	if (Ns_ConnGetKeepAliveFlag(conn)) {
 	    keep = "keep-alive";
 	} else {
 	    keep = "close";
 	}
-
 	Ns_ConnCondSetHeaders(conn, "Connection", keep);
 
+	/*
+	 * Output all headers.
+	 */
+
 	for (i = 0; i < Ns_SetSize(conn->outputheaders); i++) {
 	    key = Ns_SetKey(conn->outputheaders, i);
 	    value = Ns_SetValue(conn->outputheaders, i);
@@ -310,10 +258,10 @@ Ns_ConnQueueHeaders(Ns_Conn *conn, int status)
     Conn *connPtr = (Conn *) conn;
 
     if (!(conn->flags & NS_CONN_SENTHDRS)) {
-    	connPtr->responseStatus = status;
+	Ns_ConnSetStatus(conn, status);
     	if (!(conn->flags & NS_CONN_SKIPHDRS)) {
-	    Ns_ConnConstructHeaders(conn, &connPtr->queued);
-	    connPtr->nContentSent -= connPtr->queued.length;
+	    Ns_ConnConstructHeaders(conn, &connPtr->obuf);
+	    connPtr->nContentSent -= connPtr->obuf.length;
     	}
     	conn->flags |= NS_CONN_SENTHDRS;
     }
@@ -436,9 +384,10 @@ Ns_ConnReplaceHeaders(Ns_Conn *conn, Ns_Set *newheaders)
  */
 
 void
-Ns_ConnSetRequiredHeaders(Ns_Conn *conn, char *type, int length)
+Ns_ConnSetRequiredHeaders(Ns_Conn *conn, char *newtype, int length)
 {
     Conn *connPtr = (Conn *) conn;
+    char *type;
     Ns_DString ds;
 
     /*
@@ -455,7 +404,7 @@ Ns_ConnSetRequiredHeaders(Ns_Conn *conn, char *type, int length)
      * if AOLpress support is enabled.
      */
 
-    if (connPtr->servPtr->opts.aolpress) {
+    if (connPtr->servPtr->opts.flags & SERV_AOLPRESS) {
     	Ns_DStringAppend(&ds, "NaviServer/2.0 ");
     }
     Ns_DStringVarAppend(&ds, Ns_InfoServerName(), "/", Ns_InfoServerVersion(), NULL);
@@ -466,13 +415,17 @@ Ns_ConnSetRequiredHeaders(Ns_Conn *conn, char *type, int length)
      * that a valid length is required for connection keep-alive.
      */
 
-    if (type != NULL) {
-    	Ns_ConnSetTypeHeader(conn, type);
+    type = Ns_ConnGetType(conn);
+    if (type != newtype) {
+	Ns_ConnSetType(conn, newtype);
+	type = Ns_ConnGetType(conn);
+    }
+    if (type) {
+	Ns_ConnSetTypeHeader(conn, type);
     }
     if (length >= 0) {
 	Ns_ConnSetLengthHeader(conn, length);
     }
-
     Ns_DStringFree(&ds);
 }
 
@@ -621,13 +574,15 @@ Ns_ConnPrintfHeader(Ns_Conn *conn, char *fmt,...)
  *
  * Ns_ConnResetReturn --
  *
- *	Deprecated 
+ *	Reset the connection, clearing any queued headers, so a
+ *	basic result may be sent.
  *
  * Results:
- *	NS_OK 
+ *	NS_OK if connection could be cleared, NS_ERROR if data has
+ *	already been sent.
  *
  * Side effects:
- *	None. 
+ *	Will truncate queued headers.
  *
  *----------------------------------------------------------------------
  */
@@ -635,6 +590,19 @@ Ns_ConnPrintfHeader(Ns_Conn *conn, char *fmt,...)
 int
 Ns_ConnResetReturn(Ns_Conn *conn)
 {
+    Conn *connPtr = (Conn *) conn;
+
+    if (connPtr->nContentSent) {
+	return NS_ERROR;
+    }
+
+    /*
+     * Clear queued headers and reset status and type.
+     */
+
+    Ns_DStringFree(&connPtr->obuf);
+    Ns_ConnSetType(conn, NULL);
+    Ns_ConnSetStatus(conn, 0);
     return NS_OK;
 }
 
@@ -671,7 +639,7 @@ Ns_ConnReturnAdminNotice(Ns_Conn *conn, int status, char *title, char *notice)
  *	Return a short notice to a client. 
  *
  * Results:
- *	See Ns_ReturnHtml. 
+ *	See Ns_ConnReturnHtml. 
  *
  * Side effects:
  *	None.
@@ -704,7 +672,7 @@ Ns_ConnReturnNotice(Ns_Conn *conn, int status, char *title, char *notice)
     /*
      * Detailed server information at the bottom of the page.
      */
-    if (servPtr->opts.noticedetail) {
+    if (servPtr->opts.flags & SERV_NOTICEDETAIL) {
 	Ns_DStringVarAppend(&ds, "<P ALIGN=RIGHT><SMALL><I>",
 			    Ns_InfoServerName(), "/",
 			    Ns_InfoServerVersion(), " on ",
@@ -749,7 +717,7 @@ Ns_ConnReturnNotice(Ns_Conn *conn, int status, char *title, char *notice)
 int
 Ns_ConnReturnData(Ns_Conn *conn, int status, char *data, int len, char *type)
 {
-   return ReturnCharData(conn, status, data, len, type, NS_TRUE);
+   return ReturnData(conn, status, data, len, type, 1);
 }
 
 
@@ -772,90 +740,38 @@ Ns_ConnReturnData(Ns_Conn *conn, int status, char *data, int len, char *type)
 int
 Ns_ConnReturnCharData(Ns_Conn *conn, int status, char *data, int len, char *type)
 {
-   return ReturnCharData(conn, status, data, len, type, NS_FALSE);
+   return ReturnData(conn, status, data, len, type, 0);
 }
 
 
 /*
  *----------------------------------------------------------------------
  *
- * ReturnCharData --
+ * ReturnData --
  *
  *	Sets required headers, dumps them, and then writes your data. 
- *      If sendRaw is false, then translate the data from utf-8 to the
- *      correct character encoding if appropriate.
+ *      If direct is true, calls Ns_ConnFlushDirect which bypasses
+ *	character encoding and/or gzip compression.
  *
  * Results:
  *	NS_OK/NS_ERROR
  *
  * Side effects:
- *	May set numerous headers, will close connection. 
+ *	See Ns_ConnFlush.
  *
  *----------------------------------------------------------------------
  */
 
 static int
-ReturnCharData(Ns_Conn *conn, int status, char *data, int len, char *type,
-               int sendRaw)
+ReturnData(Ns_Conn *conn, int status, char *data, int len, char *type,
+	   int direct)
 {
-    int          result;
-    Conn        *connPtr;
-    Tcl_DString  ds;
-    Tcl_Encoding enc;
-    Tcl_DString  type_ds;
-    int          new_type = NS_FALSE;
-
-    connPtr = (Conn *)conn;
-
-    /* Make sure we know what output encoding (if any) to use. */
-    if (!sendRaw ) {
-
-        NsComputeEncodingFromType(type, &enc, &new_type, &type_ds);
-        if (new_type) {
-            type = Tcl_DStringValue(&type_ds);
-        }
-        if (enc != NULL) {
-            connPtr->encoding = enc;
-        } else if (connPtr->encoding == NULL) {
-            sendRaw = NS_TRUE;
-        }
+    Ns_ConnSetStatus(conn, status);
+    Ns_ConnSetType(conn, type);
+    if (direct) {
+    	return Ns_ConnFlushDirect(conn, data, len, 0);
     }
-
-    if (!sendRaw) {
-
-        Tcl_UtfToExternalDString(connPtr->encoding, data, len, &ds);
-
-	data = Tcl_DStringValue(&ds);
-	len = Tcl_DStringLength(&ds);
-
-    } else {
-
-	/* Send data unmodified. */
-	if (len == -1) {
-	    len = data ? strlen(data) : 0;
-	}
-
-    }
-
-    Ns_ConnSetRequiredHeaders(conn, type, len);
-    Ns_ConnQueueHeaders(conn, status);
-    if (conn->flags & NS_CONN_SKIPBODY) {
-        data = NULL;
-        len = 0;
-    }
-    result = Ns_WriteConn(conn, data, len);
-    if (result == NS_OK) {
-        result = Ns_ConnClose(conn);
-    }
-
-    if (!sendRaw && (connPtr->encoding != NULL)) {
-        Tcl_DStringFree(&ds);
-    }
-    if (new_type) {
-        Tcl_DStringFree(&type_ds);
-    }
-
-    return result;
+    return Ns_ConnFlush(conn, data, len, 0);
 }
 
 
@@ -878,7 +794,7 @@ ReturnCharData(Ns_Conn *conn, int status, char *data, int len, char *type,
 int
 Ns_ConnReturnHtml(Ns_Conn *conn, int status, char *html, int len)
 {
-    return Ns_ConnReturnData(conn, status, html, len, "text/html");
+    return Ns_ConnReturnCharData(conn, status, html, len, "text/html");
 }
 
 
@@ -1043,7 +959,7 @@ Ns_ConnReturnUnauthorized(Ns_Conn *conn)
     Ns_DStringFree(&ds);
 
     if (ReturnRedirect(conn, 401, &result)) {
-        return result;
+	return result;
     }
     return Ns_ConnReturnNotice(conn, 401, "Access Denied",
 	"The requested URL cannot be accessed because a "
@@ -1192,6 +1108,37 @@ Ns_ConnReturnInternalError(Ns_Conn *conn)
 	"due to a system error on this server.");
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Ns_ConnReturnServiceUnavailable --
+ *
+ *	Return a 503 Service Unavailable response. 
+ *
+ * Results:
+ *	NS_OK/NS_ERROR 
+ *
+ * Side effects:
+ *	Will close the connection. 
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+Ns_ConnReturnServiceUnavailable(Ns_Conn *conn)
+{
+    int result;
+
+    if (ReturnRedirect(conn, 503, &result)) {
+	return result;
+    }
+    return Ns_ConnReturnNotice(conn, 503, "Service Unavailable",
+	"The requested URL cannot be accessed "
+	"because the server is temporarily unable "
+        "to fulfill your request.  Please try again later.");
+}
+
 
 /*
  *----------------------------------------------------------------------
@@ -1243,7 +1190,7 @@ int
 Ns_ConnReturnOpenChannel(Ns_Conn *conn, int status, char *type,
 			 Tcl_Channel chan, int len)
 {
-    return ReturnOpen(conn, status, type, chan, NULL, -1, len);
+    return ReturnOpen(conn, status, type, chan, NULL, -1, -1, len);
 }
 
 
@@ -1266,7 +1213,7 @@ Ns_ConnReturnOpenChannel(Ns_Conn *conn, int status, char *type,
 int
 Ns_ConnReturnOpenFile(Ns_Conn *conn, int status, char *type, FILE *fp, int len)
 {
-    return ReturnOpen(conn, status, type, NULL, fp, -1, len);
+    return ReturnOpen(conn, status, type, NULL, fp, -1, -1, len);
 }
 
 
@@ -1289,7 +1236,32 @@ Ns_ConnReturnOpenFile(Ns_Conn *conn, int status, char *type, FILE *fp, int len)
 int
 Ns_ConnReturnOpenFd(Ns_Conn *conn, int status, char *type, int fd, int len)
 {
-    return ReturnOpen(conn, status, type, NULL, NULL, fd, len);
+    return ReturnOpen(conn, status, type, NULL, NULL, fd, -1, len);
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * Ns_ConnReturnOpenFdEx --
+ *
+ *	Send an open fd out the conn starting at given offset.  The
+ *	current file position is not changed.
+ *
+ * Results:
+ *	See ReturnOpen. 
+ *
+ * Side effects:
+ *	See ReturnOpen. 
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+Ns_ConnReturnOpenFdEx(Ns_Conn *conn, int status, char *type, int fd,
+		      off_t off, int len)
+{
+    return ReturnOpen(conn, status, type, NULL, NULL, fd, off, len);
 }
 
 
@@ -1320,13 +1292,8 @@ ReturnRedirect(Ns_Conn *conn, int status, int *resultPtr)
     servPtr = connPtr->servPtr;
     hPtr = Tcl_FindHashEntry(&servPtr->request.redirect, (char *) status);
     if (hPtr != NULL) {
-	if (++connPtr->recursionCount > MAX_RECURSION) {
-	    Ns_Log(Error, "return: failed to redirect '%d': "
-		   "exceeded recursion limit of %d", status, MAX_RECURSION);
-	} else {
-    	    *resultPtr = Ns_ConnRedirect(conn, Tcl_GetHashValue(hPtr));
-	    return 1;
-	}
+        *resultPtr = Ns_ConnRedirect(conn, Tcl_GetHashValue(hPtr));
+        return 1;
     }
     return 0;
 }
@@ -1350,7 +1317,7 @@ ReturnRedirect(Ns_Conn *conn, int status, int *resultPtr)
 
 static int
 ReturnOpen(Ns_Conn *conn, int status, char *type, Tcl_Channel chan,
-	FILE *fp, int fd, int len)
+	FILE *fp, int fd, off_t off, int len)
 {
     int result;
 
@@ -1360,11 +1327,92 @@ ReturnOpen(Ns_Conn *conn, int status, char *type, Tcl_Channel chan,
 	result = Ns_ConnSendChannel(conn, chan, len);
     } else if (fp != NULL) {
 	result = Ns_ConnSendFp(conn, fp, len);
-    } else {
+    } else if (off < 0) {
 	result = Ns_ConnSendFd(conn, fd, len);
+    } else {
+	result = Ns_ConnSendFdEx(conn, fd, off, len);
     }
     if (result == NS_OK) {
         result = Ns_ConnClose(conn);
     }
     return result;
 }
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * HdrEq --
+ *
+ *	Test if given set contains a key which matches given value.
+ *
+ * Results:
+ *	1 if there is a match, 0 otherwise.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static int
+HdrEq(Ns_Set *set, char *name, char *value)
+{
+    char *hdrvalue;
+
+    if (set != NULL
+	&& (hdrvalue = Ns_SetIGet(set, name)) != NULL
+	&& STRIEQ(hdrvalue, value)) {
+	return 1;
+    }
+    return 0;
+}
+
+
+static int
+CheckKeep(Ns_Conn *conn, int status)
+{
+    Conn *connPtr = (Conn *) conn;
+    char *hdr;
+
+    /*
+     * First, ensure the driver supports keep-alive, the request method
+     * was GET, and the client sent a connection: keep-alive header.
+     */
+
+    if (connPtr->drvPtr->keepwait > 0 &&
+	conn->request != NULL &&
+	STREQ(conn->request->method, "GET") &&
+	HdrEq(conn->headers, "connection", "keep-alive")) {
+
+	/*
+	 * Status 304, without any content, is ok.
+	 */
+
+    	if (status == 304) {
+	    return 1;
+    	}
+
+	/*
+	 * Status 200 requires either chunked encoding or a a valid
+	 * content-length header.
+	 */
+
+	if (status == 200) {
+    	    if (HdrEq(conn->outputheaders, "transfer-encoding", "chunked")) {
+		return 1;
+	    }
+    	    hdr = Ns_SetIGet(conn->outputheaders, "content-length");
+    	    if (hdr != NULL &&
+		((int) strtol(hdr, NULL, 10) == connPtr->responseLength)) {
+	    	return 1;
+	    }
+	}
+    }
+
+    /*
+     * Test for keep-alive failed.
+     */
+
+    return 0;
+}
