@@ -34,7 +34,7 @@
  *	Tcl commands that do stuff to the filesystem. 
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclfile.c,v 1.19 2003/08/21 20:11:14 vasiljevic Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclfile.c,v 1.24 2005/08/23 21:41:31 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 #ifdef _WIN32
@@ -43,6 +43,15 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclfi
 #include <utime.h>
 #endif
 
+/*
+ * Structure handling one registered channel for the [ns_chan] command
+ */
+
+typedef struct _NsRegChan {
+    char *name;
+    Tcl_Channel chan;
+} NsRegChan;
+
 static void SpliceChannel(Tcl_Interp *interp, Tcl_Channel chan);
 static void UnspliceChannel(Tcl_Interp *interp, Tcl_Channel chan);
 
@@ -634,7 +643,6 @@ NsTclUrl2FileObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST
     NsUrlToFile(&ds, itPtr->servPtr, Tcl_GetString(objv[1]));
     Tcl_SetResult(interp, ds.string, TCL_VOLATILE);
     Ns_DStringFree(&ds);
-    
     return TCL_OK;
 }
 
@@ -1016,7 +1024,8 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
     NsInterp *itPtr = arg;
     NsServer *servPtr = itPtr->servPtr;
     Tcl_Channel chan = NULL;
-    char *name;
+    char *name, *chanName;
+    NsRegChan *regChan = NULL;
     int new, shared;
     Tcl_HashTable *tabPtr;
     Tcl_HashEntry *hPtr;
@@ -1026,7 +1035,7 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
     };
     enum {
 	CCleanupIdx, CListIdx, CCreateIdx, CPutIdx, CGetIdx
-    } opt;
+    } _nsmayalias opt;
 
     if (objc < 2) {
 	Tcl_WrongNumArgs(interp, 1, objv, "command ?args?");
@@ -1040,28 +1049,29 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
     switch (opt) {
     case CCreateIdx:
 	if (objc != 4) {
-            Tcl_WrongNumArgs(interp, 1, objv, "create channel name");
+	    Tcl_WrongNumArgs(interp, 1, objv, "create channel name");
 	    return TCL_ERROR;
 	}
-	chan = Tcl_GetChannel(interp, Tcl_GetString(objv[2]), NULL);
+	chanName = Tcl_GetString(objv[2]);
+	chan = Tcl_GetChannel(interp, chanName, NULL);
 	if (chan == (Tcl_Channel)NULL) {
 	    return TCL_ERROR;
 	}
-	if (!Tcl_IsChannelRegistered(interp, chan)) {
-	    Tcl_SetResult(interp, "channel is not registered", TCL_STATIC);
-	    return TCL_ERROR;
-	}
 	if (Tcl_IsChannelShared(chan)) {
 	    Tcl_SetResult(interp, "channel is shared", TCL_STATIC);
 	    return TCL_ERROR;
 	}
-        name = Tcl_GetString(objv[3]);
-    	Ns_MutexLock(&servPtr->chans.lock);
+	name = Tcl_GetString(objv[3]);
+	Ns_MutexLock(&servPtr->chans.lock);
 	hPtr = Tcl_CreateHashEntry(&servPtr->chans.table, name, &new);
-    	if (new) {
-            Tcl_SetHashValue(hPtr, chan);
-    	}
-    	Ns_MutexUnlock(&servPtr->chans.lock);
+	if (new) {
+	    regChan = ns_malloc(sizeof(NsRegChan));
+	    regChan->name = ns_malloc(strlen(chanName)+1);
+	    regChan->chan = chan;
+	    strcpy(regChan->name, chanName);
+	    Tcl_SetHashValue(hPtr, regChan);
+	}
+	Ns_MutexUnlock(&servPtr->chans.lock);
 	if (!new) {
 	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
 		    "channel with name \"", Tcl_GetString(objv[3]),
@@ -1069,7 +1079,7 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	    return TCL_ERROR;
 	}
 	UnspliceChannel(interp, chan);
-        break;
+	break;
 
     case CGetIdx:
 	if (objc != 3) {
@@ -1077,22 +1087,22 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	    return TCL_ERROR;
 	}
 	name = Tcl_GetString(objv[2]);
-    	Ns_MutexLock(&servPtr->chans.lock);
-    	hPtr = Tcl_FindHashEntry(&servPtr->chans.table, name);
+	Ns_MutexLock(&servPtr->chans.lock);
+	hPtr = Tcl_FindHashEntry(&servPtr->chans.table, name);
 	if (hPtr != NULL) {
-	    chan = (Tcl_Channel)Tcl_GetHashValue(hPtr);
+	    regChan = (NsRegChan*)Tcl_GetHashValue(hPtr);
 	    Tcl_DeleteHashEntry(hPtr);
-        }
+	}
 	Ns_MutexUnlock(&servPtr->chans.lock);
 	if (hPtr == NULL) {
 	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
 		    "no such shared channel: ", name, NULL);
 	    return TCL_ERROR;
 	}
-	SpliceChannel(interp, chan);
-	Tcl_SetResult(interp, (char*)Tcl_GetChannelName(chan), TCL_VOLATILE);
+	SpliceChannel(interp, regChan->chan);
+	Tcl_SetResult(interp, regChan->name, TCL_VOLATILE);
 	hPtr = Tcl_CreateHashEntry(&itPtr->chans, name, &new);
-	Tcl_SetHashValue(hPtr, chan);
+	Tcl_SetHashValue(hPtr, regChan);
 	break;
 
     case CPutIdx:
@@ -1104,16 +1114,24 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	hPtr = Tcl_FindHashEntry(&itPtr->chans, name);
 	if (hPtr == NULL) {
 	    Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
-		    "no such shared channel: ", name, NULL);
+		   "no such shared channel: ", name, NULL);
 	    return TCL_ERROR;
 	}
-	chan = Tcl_GetHashValue(hPtr);
-	UnspliceChannel(interp, chan);
+	regChan = (NsRegChan*)Tcl_GetHashValue(hPtr);
+    chan = Tcl_GetChannel(interp, regChan->name, NULL);
+	if (chan == (Tcl_Channel)NULL || chan != regChan->chan) {
+	    Tcl_DeleteHashEntry(hPtr);
+        if (chan != regChan->chan) {
+            Tcl_SetResult(interp, "channel mismatch", TCL_STATIC);
+        }
+	    return TCL_ERROR;
+	}
+	UnspliceChannel(interp, regChan->chan);
 	Tcl_DeleteHashEntry(hPtr);
-    	Ns_MutexLock(&servPtr->chans.lock);
+	Ns_MutexLock(&servPtr->chans.lock);
 	hPtr = Tcl_CreateHashEntry(&servPtr->chans.table, name, &new);
-	Tcl_SetHashValue(hPtr, chan);
-    	Ns_MutexUnlock(&servPtr->chans.lock);
+	Tcl_SetHashValue(hPtr, regChan);
+	Ns_MutexUnlock(&servPtr->chans.lock);
 	break;
 
     case CListIdx:
@@ -1121,7 +1139,7 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	    Tcl_WrongNumArgs(interp, 1, objv, "list ?-shared?");
 	    return TCL_ERROR;
 	}
-        shared = (objc == 3);
+	shared = (objc == 3);
 	if (shared) {
 	    Ns_MutexLock(&servPtr->chans.lock);
 	    tabPtr = &servPtr->chans.table; 
@@ -1143,23 +1161,28 @@ NsTclChanObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj *CONST obj
 	    Tcl_WrongNumArgs(interp, 1, objv, "cleanup ?-shared?");
 	    return TCL_ERROR;
 	}
-        shared = (objc == 3);
+	shared = (objc == 3);
 	if (shared) {
 	    Ns_MutexLock(&servPtr->chans.lock);
-	    tabPtr = &servPtr->chans.table; 
+	    tabPtr = &servPtr->chans.table;
 	} else {
 	    tabPtr = &itPtr->chans;
 	}
 	hPtr = Tcl_FirstHashEntry(tabPtr, &search);
 	while (hPtr != NULL) {
-	    chan = Tcl_GetHashValue(hPtr);
+	    regChan = (NsRegChan*)Tcl_GetHashValue(hPtr);
 	    if (shared) {
-		Tcl_Close(NULL, chan);
+		Tcl_SpliceChannel(regChan->chan);
+		Tcl_UnregisterChannel((Tcl_Interp*)NULL, regChan->chan);
+	    } else {
+		Tcl_UnregisterChannel(interp, regChan->chan);
 	    }
+	    ns_free(regChan->name);
+	    ns_free(regChan);
 	    Tcl_DeleteHashEntry(hPtr);
 	    hPtr = Tcl_NextHashEntry(&search);
 	}
-	if (objc == 3) {
+	if (shared) {
 	    Ns_MutexUnlock(&servPtr->chans.lock);
 	}
 	break;
