@@ -33,7 +33,7 @@
  *	Implements the tcl ns_set commands 
  */
 
-static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclset.c,v 1.16 2003/08/06 20:30:12 elizthom Exp $, compiled: " __DATE__ " " __TIME__;
+static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclset.c,v 1.19 2005/08/01 20:27:35 jgdavidson Exp $, compiled: " __DATE__ " " __TIME__;
 
 #include "nsd.h"
 
@@ -56,8 +56,6 @@ static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/aolserver/nsd/tclse
  * Local functions defined in this file
  */
 
-static int NoServer(Tcl_Interp *interp);
-static int BadArgs(Tcl_Interp *interp, char **argv, char *args);
 static int LookupSet(NsInterp *itPtr, char *id, int delete, Ns_Set **setPtr);
 static int LookupObjSet(NsInterp *itPtr, Tcl_Obj *idPtr, int delete,
 			Ns_Set **setPtr);
@@ -86,7 +84,7 @@ static int EnterSet(NsInterp *itPtr, Ns_Set *set, int flags);
 int
 Ns_TclEnterSet(Tcl_Interp *interp, Ns_Set *set, int flags)
 {
-    NsInterp *itPtr = NsGetInterp(interp);
+    NsInterp *itPtr = NsGetInterpData(interp);
 
     if (itPtr == NULL) {
 	Tcl_SetResult(interp, "ns_set not supported", TCL_STATIC);
@@ -259,9 +257,6 @@ NsTclSetObjCmd(ClientData arg, Tcl_Interp *interp, int objc, Tcl_Obj **objv)
 	    tablePtr = &itPtr->sets;
     	    locked = 0;
 	} else if (STREQ(Tcl_GetString(objv[2]), "-shared")) {
-	    if (itPtr->servPtr == NULL) {
-		return NoServer(interp);
-	    }
 	    tablePtr = &itPtr->servPtr->sets.table;
 	    locked = 1;
 	    Ns_MutexLock(&itPtr->servPtr->sets.lock);
@@ -660,9 +655,6 @@ EnterSet(NsInterp *itPtr, Ns_Set *set, int flags)
 	 * Lock the global mutex and use the shared sets.
 	 */
 	
-	if (itPtr->servPtr == NULL) {
-	    return NoServer(itPtr->interp);
-	}
 	if (flags & NS_TCL_SET_DYNAMIC) {
 	    type = SET_SHARED_DYNAMIC;
 	} else {
@@ -731,7 +723,7 @@ LookupInterpSet(Tcl_Interp *interp, char *id, int delete, Ns_Set **setPtr)
 {
     NsInterp *itPtr;
 
-    itPtr = NsGetInterp(interp);
+    itPtr = NsGetInterpData(interp);
     if (itPtr == NULL) {
 	Tcl_SetResult(interp, "ns_set not supported", TCL_STATIC);
 	return TCL_ERROR;
@@ -753,9 +745,6 @@ LookupSet(NsInterp *itPtr, char *id, int delete, Ns_Set **setPtr)
     
     set = NULL;
     if (IS_SHARED(id)) {
-	if (itPtr->servPtr == NULL) {
-	    return NoServer(itPtr->interp);
-	}
     	tablePtr = &itPtr->servPtr->sets.table;
         Ns_MutexLock(&itPtr->servPtr->sets.lock);
     } else {
@@ -778,36 +767,3 @@ LookupSet(NsInterp *itPtr, char *id, int delete, Ns_Set **setPtr)
     *setPtr = set;
     return TCL_OK;
 }
-
-
-/*
- *----------------------------------------------------------------------
- *
- * BadArgs --
- *
- *	Complain that the wrong # args were recieved. 
- *
- * Results:
- *	TCL result. 
- *
- * Side effects:
- *	Error message appended to interp result.
- *
- *----------------------------------------------------------------------
- */
-
-static int
-BadArgs(Tcl_Interp *interp, char **argv, char *args)
-{
-    Tcl_AppendResult(interp, "wrong # of args: should be \"",
-        argv[0], " ", argv[1], " ", args, "\"", NULL);
-
-    return TCL_ERROR;
-}
-
-static int
-NoServer(Tcl_Interp *interp)
-{
-    Tcl_SetResult(interp, "no server for shared sets", TCL_STATIC);
-    return TCL_ERROR;
-}
