@@ -138,11 +138,7 @@ DllMain(HANDLE hModule, DWORD why, LPVOID lpReserved)
 
     switch (why) {
     case DLL_PROCESS_ATTACH:
-	tlskey = TlsAlloc();
-	if (tlskey == 0xFFFFFFFF) {
-	    return FALSE;
-	}
-	NsInitThreads();
+	NsThreads_LibInit();
 	/* FALLTHROUGH */
 
     case DLL_THREAD_ATTACH:
@@ -185,6 +181,55 @@ DllMain(HANDLE hModule, DWORD why, LPVOID lpReserved)
     return TRUE;
 }
 
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsInitThreads --
+ *
+ *	Core threads init.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+void
+NsInitThreads(void)
+{
+    tlskey = TlsAlloc();
+    if (tlskey == 0xFFFFFFFF) {
+	NsThreadFatal("NsInitThreads", "TlsAlloc", GetLastError());
+    }
+}
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * NsGetStack --
+ *
+ *	Return stack info.
+ *
+ * Results:
+ *	Unsupport, always returns 0.
+ *
+ * Side effects:
+ *	None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+int
+NsGetStack(void **addrPtr, size_t *sizePtr)
+{
+    return 0;
+}
+
 
 /*
  *----------------------------------------------------------------------
